Package: nhppp
Title: Simulating Nonhomogeneous Poisson Point Processes
Version: 1.0.5
Authors@R: 
    c(person(given = "Thomas", 
             family = "Trikalinos", 
             email = "thomas_trikalinos@brown.edu", 
             role = c("aut", "cre", "cph"),
             comment = c(ORCID = "0000-0002-3990-1848")), 
      person(given = "Yuliia", 
             family = "Sereda", 
             email = "sereda_yuliia@brown.edu", 
             role = c("aut"),
             comment = c(ORCID = "0000-0002-4017-4561"))
    )
Description: Simulates events from one dimensional nonhomogeneous Poisson point processes (NHPPPs) as per Trikalinos and Sereda (2024, <doi:10.48550/arXiv.2402.00358> and 2024, <doi:10.1371/journal.pone.0311311>). Functions are based on three algorithms that provably sample from a target NHPPP: the time-transformation of a homogeneous Poisson process (of intensity one) via the inverse of the integrated intensity function (Cinlar E, "Theory of stochastic processes" (1975, ISBN:0486497996)); the generation of a Poisson number of order statistics from a fixed density function; and the thinning of a majorizing NHPPP via an acceptance-rejection scheme (Lewis PAW, Shedler, GS (1979) <doi:10.1002/nav.3800260304>).
License: GPL (>= 3)
Imports: lifecycle, rstream, Rcpp (>= 1.0.12)
LinkingTo: Rcpp
Encoding: UTF-8
Language: en
RoxygenNote: 7.3.2
Suggests: data.table, ggplot2, knitr, rlecuyer, rmarkdown, testthat,
        tictoc, truncnorm, withr,
Config/Needs/website: rmarkdown
URL: https://bladder-ca.github.io/nhppp/,
        https://github.com/bladder-ca/nhppp
BugReports: https://github.com/bladder-ca/nhppp/issues
VignetteBuilder: knitr
Depends: R (>= 2.10)
LazyData: true
NeedsCompilation: yes
Packaged: 2026-04-21 23:00:51 UTC; tom
Author: Thomas Trikalinos [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0002-3990-1848>),
  Yuliia Sereda [aut] (ORCID: <https://orcid.org/0000-0002-4017-4561>)
Maintainer: Thomas Trikalinos <thomas_trikalinos@brown.edu>
Repository: CRAN
Date/Publication: 2026-04-21 23:22:09 UTC
Built: R 4.5.2; aarch64-apple-darwin20; 2026-04-22 00:08:08 UTC; unix
Archs: nhppp.so.dSYM

# Generated by roxygen2: do not edit by hand

export(draw)
export(draw_cumulative_intensity)
export(draw_intensity)
export(draw_sc_linear)
export(draw_sc_loglinear)
export(draw_sc_step)
export(draw_sc_step_regular)
export(get_step_majorizer)
export(inverse_with_uniroot)
export(inverse_with_uniroot_sorted)
export(ppp)
export(ppp_exactly_n)
export(ppp_n)
export(ppp_next_n)
export(ppp_orderstat)
export(ppp_sequential)
export(rng_stream_rexp)
export(rng_stream_rpois)
export(rng_stream_runif)
export(rng_stream_rztpois)
export(rztpois)
export(simpson_num_integr)
export(vdraw)
export(vdraw_cumulative_intensity)
export(vdraw_intensity)
export(vdraw_sc_step_regular)
export(vdraw_sc_step_regular_cpp)
export(ztdraw_cumulative_intensity)
export(ztdraw_sc_linear)
export(ztdraw_sc_loglinear)
export(ztppp)
importClassesFrom(rstream,rstream.mrg32k3a)
importFrom(Rcpp,sourceCpp)
importFrom(lifecycle,deprecated)
useDynLib(nhppp, .registration = TRUE)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# nhppp <img src="man/figures/logo.png" align="right" width="130">

<!-- badges: start -->

[![Codecov test
coverage](https://codecov.io/gh/bladder-ca/nhppp/branch/main/graph/badge.svg)](https://app.codecov.io/gh/bladder-ca/nhppp?branch=main)
[![R-CMD-check](https://github.com/bladder-ca/nhppp/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/bladder-ca/nhppp/actions/workflows/R-CMD-check.yaml)
<!-- badges: end -->

nhppp is a package for simulating events from one dimensional
nonhomogeneous Poisson point processes (NHPPPs). Its functions are based
on three algorithms that provably sample from a target NHPPP: the
time-transformation of a homogeneous Poisson process (of intensity one)
via the inverse of the integrated intensity function; the generation of
a Poisson number of order statistics from a fixed density function; and
the thinning of a majorizing NHPPP via an acceptance-rejection scheme.
It was developed to provide fast and memory efficient functions for
discrete event and statistical simulations. For a description of the
algorithms and a numerical comparison with other R packages, see
Trikalinos and Sereda (2024), accessible at
<https://arxiv.org/abs/2402.00358>.

## Installation

You can install the release version of nhppp from
[CRAN](https://cran.r-project.org) with:

``` r
install.packages("nhppp")
```

You can install the development version of nhppp from
[GitHub](https://github.com/) with:

``` r
# install.packages("devtools")
devtools::install_github("bladder-ca/nhppp")
```

## Example

These examples use the generic function `draw()`, which is a wrapper for
the packages specific functions. `draw()` is a non-vectorized function,
but `nhppp` includes vectorized functions that are fast and have small
memory footprint.

Consider the time varying intensity function
$\lambda(t) = e^{(0.2t)} (1 + \sin t)$, which is a sinusoidal intensity
function with an exponential amplitude. To draw samples over the
interval $(0, 6\pi]$ execute

``` r
l <- function(t) (1 + sin(t)) * exp(0.2 * t)
nhppp::draw(
  lambda = l,
  line_majorizer_intercept = l(6 * pi),
  line_majorizer_slope = 0,
  t_min = 0,
  t_max = 6 * pi
) |>
  head(n = 20)
#>  [1] 1.197587 1.238620 1.497499 1.713629 1.761914 2.256739 2.537528 3.622938
#>  [9] 5.822574 6.064265 6.645696 6.651551 6.684603 6.875765 6.891348 7.130680
#> [17] 7.446557 7.453139 7.545474 7.557381
```

where `line_majorizer_intercept` and `line_majorizer_slope` define a
majorizer constant.

When available, the integrated intensity function
$\Lambda(t) = \int_0^t \lambda(s) \ ds$ and its inverse
$\Lambda^{-1}(z)$ result in faster simulation times. For this example,
$\Lambda(t) = \frac{e^{0.2t}(0.2 \sin t - \cos t)+1}{1.04} + \frac{e^{0.2t} - 1}{0.2}$;
$\Lambda^{-1}(z)$ is constructed numerically upfront (or can be
calculated numerically by the function, at a computational cost).

``` r
L <- function(t) {
  exp(0.2 * t) * (0.2 * sin(t) - cos(t)) / 1.04 +
    exp(0.2 * t) / 0.2 - 4.038462
}
Li <- stats::approxfun(x = L(seq(0, 6 * pi, 10^-3)), y = seq(0, 6 * pi, 10^-3), rule = 2)

nhppp::draw(Lambda = L, Lambda_inv = Li, t_min = 0, t_max = 6 * pi) |>
  head(n = 20)
#>  [1] 0.01152846 0.23558627 0.32924742 0.49921843 0.63509297 1.36677413
#>  [7] 2.38941548 3.19511655 3.28049866 4.62140995 5.96916564 6.37504015
#> [13] 6.68283108 6.76577784 7.12919141 7.29249262 7.38665270 7.92953383
#> [19] 7.94791744 7.96591106
```

## Vectorized functions

See the vignette “Log-linear times”.

## ----include = FALSE----------------------------------------------------------
set.seed(20241018)
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
K <- 10^5

## ----setup--------------------------------------------------------------------
library(data.table)
library(nhppp)

## ----pop----------------------------------------------------------------------
pop <- data.table(
  id = 1:K,
  birth_cohort = 2015,
  spawn_age = 40,
  max_simulation_age = 110,
  sex = sample(c("male", "female"), K, replace = TRUE)
)

## It would make sense to execute the commented-out code now.
## It generates model parameters used in later stages.
## For expository clarity, we generate each parameter when it is introduced
# pop[, `:=`(
#    param_cancer_emergence_shape = runif(.N, 7, 9),
#    param_cancer_emergence_scale = rnorm(.N, 150, 20),
#    param_toxin_exposure_diff = pmax(0.005, rnorm(.N, 0.01, 0.005)),
#    param_cancer_death_intercept := rnorm(.N, -2, 0.5),
#    param_cancer_death_slope := runif(n= .N, min = 0, max = 0.003),
#    param_clinical_cancer_dx_rate := runif(.N, 0.20, 0.27)
#  )]

## ----rho----------------------------------------------------------------------
annual_mortality_rates_2015[
  sex %in% c("male", "female"),
  c(1:5, 111:113)
]

## ----death_other_causes_intensity_matrix--------------------------------------
rhos <- annual_mortality_rates_2015[
  pop,
  on = c("birth_cohort", "sex")
]
setindex(rhos, "id")
rho_matrix <- as.matrix(rhos[, c(paste0("age_", 0:109), "age_110+"),
  with = FALSE
])

rm(list = "rhos") # cleanup

## ----age_dead_other_causes----------------------------------------------------
pop[
  ,
  age_dead_from_other_causes :=
    nhppp::vdraw_sc_step_regular(
      lambda_matrix = rho_matrix,
      rate_matrix_t_min = 0,
      rate_matrix_t_max = 110,
      t_min = pop$spawn_age, # 40
      t_max = pop$max_simulation_age, # 110
      atmost1 = TRUE,
      atleast1 = TRUE
    )
]

## ----xi_params----------------------------------------------------------------
pop[, `:=`(
  exposure_start_age = max_simulation_age,
  exposure_stop_age = max_simulation_age,
  maximum_exposure = 0
)][
  ,
  will_start_exposure := runif(.N) < 0.20
][
  will_start_exposure == TRUE,
  will_stop_exposure := runif(.N) < 0.60
][
  will_start_exposure == TRUE,
  exposure_start_age := pmin(runif(.N, 12, 35), age_dead_from_other_causes)
][
  will_stop_exposure == TRUE,
  exposure_stop_age := pmin(
    exposure_start_age + runif(.N, 1, 35),
    age_dead_from_other_causes
  )
][
  will_start_exposure == TRUE,
  maximum_exposure := runif(.N, 1 / 5, 1)
]

# cleanup
pop[, will_start_exposure := NULL][, will_stop_exposure := NULL]

## ----xi-----------------------------------------------------------------------
xi <- toxin_exposure <- function(t, max_exposure, start_age, stop_age) {
  (start_age <= t) * (stop_age >= 1) * max_exposure * (1 / 2 + (cos(t / 2) + cos(0.9 * t / 2)) / 4)
}

## ----params_cancer_generation-------------------------------------------------
pop[, `:=`(
  param_cancer_emergence_shape = runif(.N, 7, 9),
  param_cancer_emergence_scale = rnorm(.N, 150, 20)
)]

## ----delta_k------------------------------------------------------------------
pop[, param_toxin_exposure_diff := pmax(0, rnorm(.N, 0.01, 0.005))]

## ----cancer_generation--------------------------------------------------------
lambda <- function(t, P = pop, ...) {
  # non-risk factor part: shape / scale * (t/scale)^(shape - 1)
  (P$param_cancer_emergence_shape / P$param_cancer_emergence_scale) *
    (t / P$param_cancer_emergence_scale)^(P$param_cancer_emergence_shape - 1) +
    # risk factor (toxin exposure) part: delta_k * xi(t)
    P$param_toxin_exposure_diff *
      xi(
        t = t,
        max_exposure = P$maximum_exposure,
        start_age = P$exposure_start_age,
        stop_age = P$exposure_stop_age
      )
}

## ----time breaks--------------------------------------------------------------
# define interval bounds for the step function, one row per person
M <- 10
time_breaks <- matrix(
  data = rep(x = seq(from = 40, to = 110, length.out = M + 1), each = K),
  byrow = FALSE,
  nrow = K
)

time_breaks[1:3, ]

## ----lambda star--------------------------------------------------------------
lambda_star <- nhppp::get_step_majorizer(
  fun = lambda,
  breaks = time_breaks,
  is_monotone = FALSE,
  K = 1.9 / 4 # This is the maximum slope of xi() -- which you get with some calculus
)

lambda_star[1:3, ]

## ----lambda-------------------------------------------------------------------
pop[
  ,
  age_cancer_emergence := nhppp::vdraw_intensity(
    lambda = lambda,
    lambda_maj_matrix = lambda_star,
    rate_matrix_t_min = 40,
    rate_matrix_t_max = 110,
    t_min = pop$spawn_age,
    t_max = pmin(pop$age_dead_from_other_causes, 110, na.rm = TRUE),
    atmost1 = TRUE
  )
][
  ,
  with_cancer := !is.na(age_cancer_emergence),
]

## ----params_cancer_death------------------------------------------------------
pop[, param_cancer_death_intercept := rnorm(.N, -3, 0.2)]
pop[, param_cancer_death_slope := runif(.N, 0, 0.003)]

## ----Nu-----------------------------------------------------------------------
Nu <- function(t, Lambda_args = list(population), ...) {
  P <- Lambda_args$population
  (
    exp(P$param_cancer_death_intercept + P$param_cancer_death_slope * t) -
      exp(P$param_cancer_death_intercept)
  ) / P$param_cancer_death_slope
}

## ----Nu_inv-------------------------------------------------------------------
Nu_inv <- function(z, Lambda_inv_args = list(population), ...) {
  P <- Lambda_inv_args$population
  (
    log(P$param_cancer_death_slope * z +
      exp(P$param_cancer_death_intercept)) -
      P$param_cancer_death_intercept
  ) / P$param_cancer_death_slope
}

## ----age_dead_cancer_causes---------------------------------------------------
args_list <- list(population = pop[!is.na(age_cancer_emergence), ])
pop[
  !is.na(age_cancer_emergence),
  age_dead_from_cancer_causes := nhppp::vdraw_cumulative_intensity(
    Lambda = Nu,
    Lambda_args = args_list,
    Lambda_inv = Nu_inv,
    Lambda_inv_args = args_list,
    t_min = pop[!is.na(age_cancer_emergence), age_cancer_emergence],
    t_max = pop[!is.na(age_cancer_emergence), age_dead_from_other_causes],
    atmost1 = TRUE
  )
]

rm(list = "args_list") # cleanup

## ----age_dead-----------------------------------------------------------------
pop[
  ,
  age_dead := pmin(age_dead_from_other_causes,
    age_dead_from_cancer_causes,
    na.rm = TRUE
  )
]

## ----param_cancer_dx_rates----------------------------------------------------
pop[
  !is.na(age_cancer_emergence),
  param_clinical_cancer_dx_rate := runif(.N, 0.20, 0.27)
]

## ----rexp---------------------------------------------------------------------
### Using rexp()
tictoc::tic()
pop[
  !is.na(age_cancer_emergence),
  age_clinical_cancer_dx :=
    age_cancer_emergence +
    rexp(.N, rate = param_clinical_cancer_dx_rate)
]
pop[
  age_clinical_cancer_dx >= age_dead,
  age_clinical_cancer_dx := NA
]
tictoc::toc()

## ----cancer_dx_rates----------------------------------------------------------
tictoc::tic()
mu_mat <- as.matrix(pop[
  !is.na(age_cancer_emergence),
  param_clinical_cancer_dx_rate
])

pop[
  !is.na(age_cancer_emergence),
  age_clinical_cancer_dx := nhppp::vdraw_sc_step_regular(
    lambda_matrix = mu_mat,
    rate_matrix_t_min = pop[!is.na(age_cancer_emergence), age_cancer_emergence],
    rate_matrix_t_max = pop[!is.na(age_cancer_emergence), age_dead],
    atmost1 = TRUE
  )
]
tictoc::toc()

## ----descriptives-------------------------------------------------------------
# pop$age_cancer_emergence |> summary()
summary(pop)

## ----include = FALSE----------------------------------------------------------
set.seed(20250820)
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
K <- 10^6

## ----setup--------------------------------------------------------------------
library(data.table)
library(nhppp)

## ----population---------------------------------------------------------------
pop <- setDT(
  list(
    id = 1:K,
    T0 = rep(20, K),
    T1 = stats::runif(n = K, min = 60, max = 100)
  )
)
setindex(pop, id)
pop

## -----------------------------------------------------------------------------
pop[, `:=`(
  alpha_weibull = stats::runif(n = K, min = 2.28, max = 3.28),
  sigma_weibull = stats::runif(n = K, min = 64, max = 84)
  )]

## ----function-lambda-weibull--------------------------------------------------
l_weibull <- function(t, alpha = pop$alpha_weibull, sigma = pop$sigma_weibull, ...) {
  alpha/sigma * (t/sigma)^(alpha-1)
}

L_weibull <- function(t, alpha = pop$alpha_weibull, sigma = pop$sigma_weibull, ...) {
  (t/sigma)^alpha
}

Li_weibull<- function(z, alpha = pop$alpha_weibull, sigma = pop$sigma_weibull, ...) {
  sigma * z^(1/alpha)
} 

## ----method-1-----------------------------------------------------------------
tictoc::tic("Method 1 (vectorized, thinning)")
M <- 5
break_points <- seq.int(from = 20, to = 100, length.out = M + 1)
breaks_mat <- matrix(rep(break_points, each = K), nrow = K)

lmaj_mat <- nhppp::get_step_majorizer(
  fun = l_weibull,
  breaks = breaks_mat,
  is_monotone = TRUE
)

pop[
  ,
  t_thinning := nhppp::vdraw_intensity(
    lambda = l_weibull,
    lambda_maj_matrix = lmaj_mat,
    rate_matrix_t_min = 20,
    rate_matrix_t_max = 100,
    t_min = pop$T0,
    t_max = pop$T1,
    atmost1 = TRUE,
    atmostB = 5
  )
]
tictoc::toc(log = TRUE) # timer end

## ----method-2-----------------------------------------------------------------
tictoc::tic("Method 2 (vectorized, inversion)")
pop[
  ,
  t_inversion := nhppp::vdraw_cumulative_intensity(
    Lambda = L_weibull,
    Lambda_inv = Li_weibull,
    t_min = pop$T0,
    t_max = pop$T1,
    atmost1 = TRUE
  )
]
tictoc::toc(log = TRUE) # timer end

## ----qq-plot, fig.alt="QQ plot comparing simulated times with the two methods. The QQ plot indicates excellent agreement."----
qqplot(pop$t_thinning, pop$t_inversion)

## ----histogram----------------------------------------------------------------
if(nchar(system.file(package = 'ggplot2'))>0) {
  pop2 <- setDT(list(id = 1:10000)) 
  pop2[, `:=`(
    T0 = 20, 
    T1 = 100, 
    a = 2.78, 
    s = 74
  )]
  setindex(pop2, id)
  pop2
  
  ## re-define functions to use `pop2` by default 
  ##    clunckiness of `nhppp` -- help us improve it!
  l  <- function(t, a = pop2$a, s = pop2$s, ...) a/s * (t/s)^(a-1)
  L  <- function(t, a = pop2$a, s = pop2$s, ...) (t/s)^a
  Li <- function(z, a = pop2$a, s = pop2$s, ...)  s * z^(1/a)
  
  ## get all times (full trajectories)
  Z <- nhppp::vdraw_cumulative_intensity(
      Lambda = L,
      Lambda_inv = Li,
      t_min = pop2$T0,
      t_max = pop2$T1,
      atmost1 = FALSE
    )
  
  ## ignore non-realized times 
  Z <- as.vector(Z)
  Z <- Z[!is.na(Z)]
  
  x <- seq(from = 20, to = 100, length.out = 100)
  ## normalize the hazard rate to have area 1 between 20 and 100
  y <- l(x, a = pop2$a[1], s = pop2$s[1]) / (
        L(100, a = pop2$a[1], s = pop2$s[1]) - 
        L(20, a = pop2$a[1], s = pop2$s[1])
       )
  
  
  library(ggplot2)
  ggplot(xlim = c(0, 110)) +
    geom_histogram(
      aes(x = Z, y = after_stat(density)), 
      bins = 100, 
      fill = "gray", 
      color = "white") +
    geom_line(
      aes(x = x, y = y), 
      color = "red"
    ) +
    labs(title = "Histogram of all simulated Weibull times", x = "Time", y = "Empirical and theoretical density")
}

## ----include = FALSE----------------------------------------------------------
set.seed(20250820)
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
K <- 10^5

## ----setup--------------------------------------------------------------------
library(data.table)
library(nhppp)

## ----population---------------------------------------------------------------
pop <- setDT(
  list(
    id = 1:K,
    T0 = rep(20, K),
    T1 = stats::runif(n = K, min = 60, max = 100)
  )
)
setindex(pop, id)
pop

## -----------------------------------------------------------------------------
pop[, `:=`(
  a_gompertz = stats::runif(n = K, min = 0.0045, max = 0.0055),
  b_gompertz = stats::runif(n = K, min = 0.085, max = 0.095)
  )]

## ----function-lambda-gompertz-------------------------------------------------
l_gompertz <- function(t, a = pop$a_gompertz, b = pop$b_gompertz, ...) {
  a * b * exp(b * t)
}

L_gompertz <- function(t, a = pop$a_gompertz, b = pop$b_gompertz, ...) {
  a * exp(b * t) - a
}

Li_gompertz <- function(z, a = pop$a_gompertz, b = pop$b_gompertz, ...) {
  log(z/a +1) / b
}
 

## ----method-1-----------------------------------------------------------------
tictoc::tic("Method 1 (vectorized, thinning)")
M <- 5
break_points <- seq.int(from = 20, to = 100, length.out = M + 1)
breaks_mat <- matrix(rep(break_points, each = K), nrow = K)

lmaj_mat <- nhppp::get_step_majorizer(
  fun = l_gompertz,
  breaks = breaks_mat,
  is_monotone = TRUE
)

pop[
  ,
  t_thinning := nhppp::vdraw_intensity(
    lambda = l_gompertz,
    lambda_maj_matrix = lmaj_mat,
    rate_matrix_t_min = 20,
    rate_matrix_t_max = 100,
    t_min = pop$T0,
    t_max = pop$T1,
    atmost1 = TRUE,
    atmostB = 10
  )
]
tictoc::toc(log = TRUE) # timer end

## ----method-2-----------------------------------------------------------------
tictoc::tic("Method 2 (vectorized, inversion)")
pop[
  ,
  t_inversion := nhppp::vdraw_cumulative_intensity(
    Lambda = L_gompertz,
    Lambda_inv = Li_gompertz,
    t_min = pop$T0,
    t_max = pop$T1,
    atmost1 = TRUE
  )
]
tictoc::toc(log = TRUE) # timer end

## ----qq-plot, fig.alt="QQ plot comparing simulated times with the two methods. The QQ plot indicates excellent agreement."----
qqplot(pop$t_thinning, pop$t_inversion)

## ----all_events---------------------------------------------------------------
Z <- nhppp::vdraw_cumulative_intensity(
    Lambda = L_gompertz,
    Lambda_inv = Li_gompertz,
    t_min = pop$T0,
    t_max = pop$T1,
    atmost1 = FALSE
  )

dim(Z) # a lot of events!

## ----histogram----------------------------------------------------------------
if(nchar(system.file(package = 'ggplot2'))>0) {
  pop2 <- setDT(list(id = 1:1000)) 
  pop2[, `:=`(
    T0 = 20, 
    T1 = 100, 
    a = 0.005, 
    b = 0.090
  )]
  setindex(pop2, id)
  pop2
  
  ## re-define functions to use `pop2` by default 
  ##    clunckiness of `nhppp` -- help us improve it!
  l  <- function(t, a = pop2$a, b = pop2$b, ...) a * b * exp(b * t)
  L  <- function(t, a = pop2$a, b = pop2$b, ...) a * exp(b * t) - a
  Li <- function(z, a = pop2$a, b = pop2$b, ...) log(z/a +1) / b
  
  ## get all times (full trajectories)
  Z <- nhppp::vdraw_cumulative_intensity(
      Lambda = L,
      Lambda_inv = Li,
      t_min = pop2$T0,
      t_max = pop2$T1,
      atmost1 = FALSE
    )
  
  ## ignore non-realized times 
  Z <- as.vector(Z)
  Z <- Z[!is.na(Z)]
  
  x <- seq(from = 20, to = 100, length.out = 100)
  ## normalize the hazard rate to have area 1 between 20 and 100
  y <- l(x, a = pop2$a[1], b = pop2$b[1]) / (
        L(100, a = pop2$a[1], b = pop2$b[1]) - 
        L(20, a = pop2$a[1], b = pop2$b[1])
       )
  
  
  library(ggplot2)
  ggplot(xlim = c(0, 110)) +
    geom_histogram(
      aes(x = Z, y = after_stat(density)), 
      bins = 100, 
      fill = "gray", 
      color = "white") +
    geom_line(
      aes(x = x, y = y), 
      color = "red"
    ) +
    labs(title = "Histogram of all simulated Gompertz times", x = "Time", y = "Empirical and theoretical density")
}

#' Vectorized sampling from NHPPPs with piecewise constant intensities
#' with same interval lengths (C++)
#'
#' @description
#' Simulate a piecewise constant-rate Poisson Point Process over `(t_min, t_max]` (inversion method)
#' where the intervals have the same length (are "regular").
#'
#' @param Lambda_matrix (matrix) integrated intensity rates at the end of each interval
#' @param lambda_matrix (matrix) intensity rates, one per interval
#' @param rate_matrix_t_min (scalar | vector | column matrix) is the lower bound
#'        of the time interval for each row of (Lambda|lambda)_maj_matrix.
#'        The length of this argument is the number of point processes that should be drawn.
#' @param rate_matrix_t_max (scalar | vector | column matrix) the upper bound
#'        of the time interval for each row of (Lambda|lambda)_maj_matrix.
#'        The length of this argument is the number of point processes that should be drawn.
#' @param t_min (scalar | vector | column matrix) is the lower bound
#'        of a subinterval of (rate_matrix_t_min, rate_matrix_t_max]. If set,
#'        times are sampled from the subinterval.
#'        If omitted, it is equivalent to `rate_matrix_t_min`.
#' @param t_max (scalar | vector | column matrix) is the upper bound
#'        of a subinterval of (rate_matrix_t_min, rate_matrix_t_max]. If set,
#'        times are sampled from the subinterval.
#'        If omitted, it is equivalent to `rate_matrix_t_max`.
#' @param tol (scalar, double) tolerance for the number of events
#' @param atmost1 boolean, draw at most 1 event time
#' @param atmostB If not NULL, draw at most B (B>0) event times. NULL means ignore.
#'
#' @return a matrix of event times t, with rows corresponding to the sampled point processes.
#'
#' @examples
#' x <- vdraw_sc_step_regular_cpp(
#'   Lambda_matrix = matrix(1:5, nrow = 1),
#'   rate_matrix_t_min = 100,
#'   rate_matrix_t_max = 110,
#'   atmost1 = FALSE
#' )
#' @export
#' @keywords internal
vdraw_sc_step_regular_cpp <- function(
    lambda_matrix = NULL,
    Lambda_matrix = NULL,
    rate_matrix_t_min = NULL,
    rate_matrix_t_max = NULL,
    t_min = NULL,
    t_max = NULL,
    tol = 10^-6,
    atmost1 = FALSE,
    atmostB = NULL) {
  if (is.null(atmostB)) {
    atmostB <- 0 # has to be <=0 in the C++ argument to be ignored
  }

  if (!is.null(lambda_matrix) && is.null(Lambda_matrix)) {
    rate <- lambda_matrix
    is_cumulative_rate <- FALSE
  } else if (is.null(lambda_matrix) && !is.null(Lambda_matrix)) {
    rate <- Lambda_matrix
    is_cumulative_rate <- TRUE
  } else {
    stop("lambda_matrix and Lambda_matrix cannot both be `NULL`")
  }
  mode(rate) <- "numeric"

  num_na <- sum(is.na(rate))
  if (num_na > 0) {
    rate_argument <- if (is_cumulative_rate) "Lambda_matrix" else "lambda_matrix"
    stop("The ", rate_argument, " contains ", num_na, " NA values")
  }

  range_t <- cbind(as.vector(rate_matrix_t_min), as.vector(rate_matrix_t_max))
  if (nrow(range_t) > 1 && nrow(range_t) != nrow(rate)) {
    stop("The (rows of) [Lambda|lambda]_matrix and (length of) [rate_matrix_t_min|rate_matrix_t_max] imply different numbers of point processes to be sampled.")
  }
  if (nrow(range_t) == 1 && nrow(rate) != 1) {
    range_t <- range_t[rep(1, nrow(rate)), ]
  }

  if (is.null(t_min) && is.null(t_max)) {
    return(
      .Call(
        `_nhppp_vdraw_sc_step_regular`,
        rate, is_cumulative_rate, range_t, tol, atmost1
      )
    )
  }

  # if here, at most one of t_min t_max is null
  if (is.null(t_min)) t_min <- range_t[, 1, drop = FALSE]
  if (is.null(t_max)) t_max <- range_t[, 2, drop = FALSE]

  subinterval <- cbind(as.vector(t_min), as.vector(t_max))
  if (nrow(subinterval) > 1 && nrow(subinterval) != nrow(rate)) {
    stop("The (rows of) [Lambda|lambda]_matrix and (length of) [t_min|t_max] imply different numbers of point processes to be sampled.")
  }
  if (nrow(subinterval) == 1 && nrow(rate) != 1) {
    subinterval <- subinterval[rep(1, nrow(rate)), ]
  }

  stopifnot(all(subinterval[, 1] >= range_t[, 1]), all(subinterval[, 2] <= range_t[, 2]))
  return(
    .Call(
      `_nhppp_vdraw_sc_step_regular2`,
      rate, is_cumulative_rate, range_t, subinterval, tol, atmost1, atmostB
    )
  )
}

test_that("vdraw_sc_step_regular_cpp() works", {
  set.seed(123)
  # 1 row matrix
  expect_no_error(Z0 <- vdraw_sc_step_regular_cpp(
    Lambda_matrix = matrix(1:5, nrow = 1),
    rate_matrix_t_min = 100,
    rate_matrix_t_max = 110,
    tol = 10^-6,
    atmost1 = FALSE
  ))
  check_ppp_sample_validity(Z0, t_min = 100, t_max = 110)


  l <- lref <- matrix(rep(1, 50), ncol = 5)
  L <- Lref <- mat_cumsum_columns(l)

  expect_no_error(Z <- vdraw_sc_step_regular_cpp(
    Lambda_matrix = L,
    rate_matrix_t_min = 100,
    rate_matrix_t_max = 110,
    tol = 10^-6,
    atmost1 = FALSE
  ))
  check_ppp_sample_validity(Z, t_min = 100, t_max = 110)

  expect_no_error(Z1 <- vdraw_sc_step_regular_cpp(
    Lambda_matrix = L,
    rate_matrix_t_min = 100,
    rate_matrix_t_max = 110,
    tol = 10^-6,
    atmost1 = TRUE
  ))
  check_ppp_sample_validity(Z1, t_min = 100, t_max = 110)

  expect_no_error(Z2 <- vdraw_sc_step_regular_cpp(
    lambda_matrix = l,
    rate_matrix_t_min = 100,
    rate_matrix_t_max = 110,
    tol = 10^-6,
    atmost1 = TRUE
  ))
  check_ppp_sample_validity(Z2, t_min = 100, t_max = 110)

  expect_equal(L, Lref) # no side effects on l, L
  expect_equal(l, lref)
})

test_that("vdraw_sc_step_regular_cpp() does not break with matrices whose mode is list", {
  set.seed(123)
  l <- matrix(rep(1, 50), ncol = 5)
  L <- mat_cumsum_columns(l)

  mode(L) <- "list"
  expect_no_error(Z <- vdraw_sc_step_regular_cpp(
    Lambda_matrix = L,
    rate_matrix_t_min = 100,
    rate_matrix_t_max = 110,
    tol = 10^-6,
    atmost1 = FALSE
  ))
})



test_that("vdraw_sc_step_regular_cpp() uses blocked random numbers", {
  set.seed(123)
  l <- matrix(rep(1, 500), ncol = 5)
  L <- mat_cumsum_columns(l)

  Z <- list()
  for (i in 1:2) {
    set.seed(123)
    expect_no_error(Z[[i]] <- vdraw_sc_step_regular_cpp(
      Lambda_matrix = L,
      rate_matrix_t_min = 100,
      rate_matrix_t_max = 110,
      tol = 10^-6,
      atmost1 = FALSE
    ))
    if (i > 1) {
      expect_true(identical(Z[[1]], Z[[i]]))
    }
  }
})


test_that("vdraw_sc_step_regular_cpp() works with subinterval", {
  set.seed(123)
  expect_no_error(Z0 <- vdraw_sc_step_regular_cpp(
    Lambda_matrix = matrix(1:5, nrow = 1),
    rate_matrix_t_min = 100,
    rate_matrix_t_max = 110,
    t_min = 100,
    t_max = 110,
    tol = 10^-6,
    atmost1 = FALSE
  ))
  check_ppp_sample_validity(Z0, t_min = 100, t_max = 110, atmost1 = FALSE)
  expect_no_error(Z0 <- vdraw_sc_step_regular_cpp(
    Lambda_matrix = matrix(1:5, nrow = 1),
    rate_matrix_t_min = 100,
    rate_matrix_t_max = 110,
    t_min = 101.01,
    t_max = 108.99,
    tol = 10^-6,
    atmost1 = FALSE
  ))
  check_ppp_sample_validity(Z0, t_min = 101.01, t_max = 108.99, atmost1 = FALSE)

  expect_no_error(Z0 <- vdraw_sc_step_regular_cpp(
    Lambda_matrix = matrix(1:5, nrow = 1) * 10,
    rate_matrix_t_min = 100,
    rate_matrix_t_max = 110,
    t_min = 105.01,
    t_max = 105.99,
    tol = 10^-6,
    atmost1 = FALSE
  ))
  check_ppp_sample_validity(Z0, t_min = 105.01, t_max = 105.99, atmost1 = FALSE)
})

test_that("vdraw_sc_step_regular_cpp() uses user-supplied RNGs", {
  l <- matrix(rep(1, 50), ncol = 5)
  L <- mat_cumsum_columns(l)

  seed <- rep(1, 6)
  rlecuyer::.lec.SetPackageSeed(seed)
  rlecuyer::.lec.CreateStream(c("a", "b"))

  res <- list()

  # seed the old.kind generator (default ) hereon "R"
  set.seed(123)
  res[["R0"]] <- vdraw_sc_step_regular_cpp(
    Lambda_matrix = L, rate_matrix_t_min = 100,
    rate_matrix_t_max = 110
  )
  set.seed(123)
  res[["R1"]] <- vdraw_sc_step_regular_cpp(
    Lambda_matrix = L, rate_matrix_t_min = 100,
    rate_matrix_t_max = 110
  )
  expect_equal(res[["R0"]], res[["R1"]])

  # activate "a" and reseed "R" -- If the function uses the "R", the test will fail
  set.seed(123)
  old.kind <- rlecuyer::.lec.CurrentStream("a")
  res[["a0"]] <- vdraw_sc_step_regular_cpp(
    Lambda_matrix = L, rate_matrix_t_min = 100,
    rate_matrix_t_max = 110
  )
  rlecuyer::.lec.CurrentStreamEnd(old.kind)
  expect_false(identical(res[["R0"]], res[["a0"]]))

  # activate "b", again, re-seed "R"
  set.seed(123)
  old.kind <- rlecuyer::.lec.CurrentStream("b")
  res[["b0"]] <- vdraw_sc_step_regular_cpp(
    Lambda_matrix = L, rate_matrix_t_min = 100,
    rate_matrix_t_max = 110
  )
  rlecuyer::.lec.CurrentStreamEnd(old.kind)
  expect_false(identical(res[["R0"]], res[["b0"]]))
  expect_false(identical(res[["b0"]], res[["a0"]]))

  # reset the RNGs -- but advance the "R"
  burn <- runif(10000)
  rm("burn")
  rlecuyer::.lec.ResetStartStream("a")
  rlecuyer::.lec.ResetStartStream("b")

  # activate "a" -- do not reseed "R"
  old.kind <- rlecuyer::.lec.CurrentStream("a")
  res[["a1"]] <- vdraw_sc_step_regular_cpp(
    Lambda_matrix = L, rate_matrix_t_min = 100,
    rate_matrix_t_max = 110
  )
  rlecuyer::.lec.CurrentStreamEnd(old.kind)
  expect_equal(res[["a0"]], res[["a1"]])
  expect_false(identical(res[["R0"]], res[["a1"]]))

  # activate "b" -- do not reseed "R"
  old.kind <- rlecuyer::.lec.CurrentStream("b")
  res[["b1"]] <- vdraw_sc_step_regular_cpp(
    Lambda_matrix = L, rate_matrix_t_min = 100,
    rate_matrix_t_max = 110
  )
  rlecuyer::.lec.CurrentStreamEnd(old.kind)
  expect_equal(res[["b0"]], res[["b1"]])
  expect_false(identical(res[["R0"]], res[["b1"]]))
  expect_false(identical(res[["a1"]], res[["b1"]]))

  rlecuyer::.lec.exit()
})



test_that("vdraw_sc_step_regular_cpp() handles NA values in rate matrices", {
  set.seed(123)
  l <- matrix(rep(1, 50), ncol = 5)
  L <- mat_cumsum_columns(l)

  l[1, 1] <- NA
  L[1, 2] <- NA

  expect_error(vdraw_sc_step_regular_cpp(
    lambda_matrix = l,
    rate_matrix_t_min = 100,
    rate_matrix_t_max = 110,
    tol = 10^-6,
    atmost1 = FALSE
  ))

  expect_error(vdraw_sc_step_regular_cpp(
    Lambda_matrix = L,
    rate_matrix_t_min = 100,
    rate_matrix_t_max = 110,
    tol = 10^-6,
    atmost1 = FALSE
  ))
})

#' Simulate from a non homogeneous Poisson Point Process (NHPPP) from
#'    (t0, t_max) (thinning method) with piecewise constant_majorizer
#'
#' @description Sample NHPPP times using the thinning method
#' @param lambda (function) the instantaneous rate of the NHPPP.
#' A continuous function of time.
#' @param majorizer_vector (scalar, double) `K` constant majorizing rates, one per interval
#' @param time_breaks (vector, double) `K+1` time points defining `K` intervals
#'        of constant rates:
#'             `[t_1 = range_t[1],       t_2)`: the first interval
#'             `[t_k,                t_{k+1})`: the `k`-th interval
#'             `[t_{K}, t_{K+1} = range_t[2])`: the `K`-th (last) interval
#' @param atmost1 boolean, draw at most 1 event time
#'
#' @return a vector of event times (t_); if no events realize,
#'         a vector of length 0
#' @keywords internal
draw_intensity_step <- function(lambda,
                                majorizer_vector,
                                time_breaks,
                                atmost1 = FALSE) {
  len_lambda <- length(majorizer_vector)
  candidate_times <- draw_sc_step(lambda_vector = majorizer_vector, time_breaks = time_breaks, atmost1 = FALSE)
  num_candidates <- length(candidate_times)
  if (num_candidates == 0) {
    return(candidate_times)
  }
  u <- stats::runif(n = num_candidates, min = 0, max = 1)
  acceptance_prob <- lambda(candidate_times) /
    stats::approx(
      x = time_breaks[1:len_lambda],
      y = majorizer_vector,
      xout = candidate_times, method = "constant", rule = 2, f = 0
    )$y
  if (!all(acceptance_prob <= 1 + 10^-6)) stop("lambda > lambda_maj\n")
  if (atmost1) {
    t <- candidate_times[u < acceptance_prob][1]
    if (is.na(t)) {
      t <- numeric(0)
    }
  } else {
    t <- candidate_times[u < acceptance_prob]
  }
  return(t)
}

#' Uniform random samples from `rstream` objects
#'
#' @description Sample from `rstream` objects
#' @param size Integer, number of samples
#' @param minimum Lower bound
#' @param maximum Upper bound
#' @param rng_stream (`rstream`) an `rstream` object or `NULL`
#'
#' @return a vector of uniform variates of size `size`
#' @keywords internal
#' @export
#' @importClassesFrom rstream rstream.mrg32k3a
#' @examples
#' rng_stream_runif(10)
rng_stream_runif <- function(size = 1, minimum = 0, maximum = 1, rng_stream = NULL) {
  if (!is.null(rng_stream)) {
    if (class(rng_stream)[1] == "RNGClass") {
      rng_stream <- rng_stream$.getPointer()
    }
    u <- rstream::r(rng_stream, size) * (maximum - minimum) + minimum
  } else {
    u <- stats::runif(n = size, min = minimum, max = maximum)
  }
  return(u)
}


#' Exponential random samples from `rstream` objects
#'
#' @description Sample from `rstream` objects

#' @param size Integer, number of samples
#' @param rate Positive number, the rate (i.e., 1/mean)
#' @param rng_stream (`rstream`) an `rstream` object or `NULL`
#'
#' @return a vector of exponential variates of size `size`
#' @keywords internal
#' @export
#' @importClassesFrom rstream rstream.mrg32k3a
#' @examples
#' rng_stream_rexp(10)
rng_stream_rexp <- function(size = 1, rate = 1, rng_stream = NULL) {
  if (!is.null(rng_stream)) {
    if (class(rng_stream)[1] == "RNGClass") {
      rng_stream <- rng_stream$.getPointer()
    }
    p <- rng_stream_runif(size = size, rng_stream = rng_stream)
    dt_ <- stats::qexp(p = p, rate = rate)
  } else {
    dt_ <- stats::rexp(n = size, rate = rate)
  }
  return(dt_)
}

#' Poisson random samples from `rstream` objects
#'
#' @description Sample from `rstream` objects
#' @param size Integer, number of samples
#' @param lambda Positive number, the mean
#' @param rng_stream (`rstream`) an `rstream` object or `NULL`
#'
#' @return a vector of counts of size `size`
#' @keywords internal
#' @export
#' @importClassesFrom rstream rstream.mrg32k3a
#' @examples
#' rng_stream_rpois(10)
rng_stream_rpois <- function(size = 1, lambda = 1, rng_stream = NULL) {
  if (!is.null(rng_stream)) {
    if (class(rng_stream)[1] == "RNGClass") {
      rng_stream <- rng_stream$.getPointer()
    }
    p <- rstream::r(rng_stream, size)
    x <- stats::qpois(p = p, lambda = lambda)
  } else {
    x <- stats::rpois(n = size, lambda = lambda)
  }
  return(x)
}



#' Zero-truncated Poisson random samples from `rstream` objects
#'
#' @description Sample from `rstream` objects
#' @param size Integer, number of samples
#' @param lambda Positive number, the mean of the original
#'        (untruncated) Poisson distribution
#' @param rng_stream (`rstream`) an `rstream` object or `NULL`
#'
#' @return a vector of non zero counts of size `size`
#' @keywords internal
#' @export
#' @importClassesFrom rstream rstream.mrg32k3a
#' @examples
#' rng_stream_rztpois(10)
rng_stream_rztpois <- function(size = 1, lambda = 1, rng_stream = NULL) {
  exp_minus_lambda <- exp(-lambda)
  if (!is.null(rng_stream)) {
    if (class(rng_stream)[1] == "RNGClass") {
      rng_stream <- rng_stream$.getPointer()
    }
    p <- rstream::r(rng_stream, size) * (1 - exp_minus_lambda) + exp_minus_lambda
  } else {
    p <- stats::runif(n = size, min = exp_minus_lambda, max = 1)
  }
  x <- stats::qpois(p = p, lambda = lambda)
  return(x)
}


#' Zero-truncated Poisson random samples (basic R)
#'
#' @description Sample zero-truncated Poisson random samples (basic R)
#' @param n Integer, number of samples
#' @param lambda Positive number, the mean of the original
#'        (untruncated) Poisson distribution
#'
#' @return a vector of non zero counts of size `n`
#' @export
#' @keywords internal
#' @examples
#' rztpois(10, 1)
#' rztpois(10, 1:10)
rztpois <- function(n, lambda) {
  p <- stats::runif(n = n, min = exp(-lambda), max = 1)
  x <- stats::qpois(p = p, lambda = lambda)
  return(x)
}

test_that("rng_stream_runif() works without rstream object", {
  expect_no_error(withr::with_seed(12345, df1 <- rng_stream_runif(size = 10)))
  expect_identical(withr::with_seed(12345, rng_stream_runif(size = 10)), df1)
  expect_equal(length(df1), 10)
})

test_that("rng_stream_runif() works with rstream object", {
  # check rstream
  S <- methods::new("rstream.mrg32k3a")
  expect_no_error(df2 <- rng_stream_runif(size = 10, rng_stream = S))
  expect_equal(length(df2), 10)

  # Check that unpacking and resetting the rng_stream returns correct sequence
  rstream::rstream.packed(S) <- FALSE
  rstream::rstream.reset(S)
  expect_identical(rng_stream_runif(size = 10, rng_stream = S), df2)
})

test_that("rng_stream_rexp() works ", {
  expect_no_error(rng_stream_rexp(size = 10, rate = 1, rng_stream = NULL))
  expect_no_error(S1 <- methods::new("rstream.mrg32k3a"))
  expect_no_error(rng_stream_rexp(size = 10, rate = 1, rng_stream = S1))
})


test_that("rng_stream_rztpois() works ", {
  expect_no_error(rng_stream_rztpois(size = 10, lambda = 1, rng_stream = NULL))
  expect_no_error(S1 <- methods::new("rstream.mrg32k3a"))
  expect_no_error(rng_stream_rztpois(size = 10, lambda = 1, rng_stream = S1))
})


test_that("rztpois() works", {
  expect_no_error(rztpois(n = 10, lambda = 1))
  # vectorized for lambda
  expect_no_error(rztpois(n = 10, lambda = rep(1, 10)))
  expect_no_error(rztpois(n = 10, lambda = 1:10))

  expect_equal(sum(rztpois(4, c(rep(.0001, 3), 10000)) < 2), 3)
})

#' Simpson's method to integrate a univariate function.
#'
#' @description Simpson's method to integrate a univariate continuous function.
#' Faster that R's `integrate()` and precise enough, but does not do any checks.
#' The error is at most `M (b-a)^5/(180 n^4)` where `M` is the maximum of
#' the fourth derivative of the integrand in the interval `[a, b]`.
#' @param f function that takes a single argument
#' @param a the lower limit of integration
#' @param b the upper limit of integration
#' @param n integer, number of integration points with a and b
#' @return numeric, the integration value
#' examples
#' #expect 1
#' simpson_num_integr(sin, 0, pi/2, 100)
#' #max error for simpson_num_integr(sin, 0, pi/2, 100) is 5.312842e-10
#' 1 * (pi/2 - 0)^5/(180 * 100^4)
#' @export
#' @keywords internal
simpson_num_integr <- function(f, a, b, n) {
  if (is.function(f) == FALSE) {
    stop("f must be a function with one parameter (variable)")
  }
  h <- (b - a) / n
  xj <- seq.int(a, b, length.out = n + 1)
  xj <- xj[-1]
  xj <- xj[-length(xj)]
  integral <- (h / 3) *
    (f(a) +
      2 * sum(f(xj[seq.int(2, length(xj), 2)])) +
      4 * sum(f(xj[seq.int(1, length(xj), 2)])) +
      f(b))
  return(integral)
}

#' Simulate a piecewise constant-rate Poisson Point Process over `(t_min, t_max]` (inversion method)
#' The intervals need not have the same length.
#'
#' @param lambda_vector (scalar, double) `K` constant rates, one per interval
#' @param time_breaks (vector, double) `K+1` time points defining `K` intervals
#'        of constant rates:
#'             `[t_1 = range_t[1],       t_2)`: the first interval
#'             `[t_k,                t_{k+1})`: the `k`-th interval
#'             `[t_{K}, t_{K+1} = range_t[2])`: the `K`-th (last) interval
#' @param atmost1 boolean, draw at most 1 event time
#' @param atleast1 boolean, draw at least 1 event time
#'
#' @return a vector of event times t
#'         if no events realize, it will have 0 length
#' @export
#'
#' @examples
#' x <- draw_sc_step(lambda_vector = rep(1, 5), time_breaks = c(0:5))
#' @export
draw_sc_step <- function(lambda_vector,
                         time_breaks,
                         # rng_stream = NULL,
                         atmost1 = FALSE,
                         atleast1 = FALSE) {
  len_time_breaks <- length(time_breaks)
  if (len_time_breaks != (length(lambda_vector) + 1)) stop("incompatible `time_breaks` and `lambda_vector` lengths")

  if (atleast1 == FALSE) {
    ppp_t_fun <- ppp2
  } else {
    ppp_t_fun <- ztppp
  }

  if (len_time_breaks == 2) {
    return(ppp_t_fun(t_min = time_breaks[1], t_max = time_breaks[2], rate = lambda_vector, atmost1 = atmost1))
  }

  time_H <- time_breaks
  lambda <- c(0, lambda_vector)
  time_L <- c(NA, time_H[1:(len_time_breaks - 1)]) # data.table::shift(time_H)
  dtime <- time_H - time_L
  Lambda <- lambda * dtime
  Lambda[1] <- 0
  time_warped_H <- cumsum(Lambda)
  time_warped_L <- c(0, time_warped_H[1:(len_time_breaks - 1)]) # data.table::shift(time_warped_H, n=1, type = "lag", fill = 0)

  times_warped <- ppp_t_fun(rate = 1, t_min = 0, t_max = time_warped_H[len_time_breaks], atmost1 = atmost1)

  num_times <- length(times_warped)
  if (num_times == 0 || any(is.na(times_warped))) {
    return(numeric(0))
  }

  x_index <- rep(1:len_time_breaks, num_times)
  t_index <- rep(1:num_times, each = len_time_breaks)
  to_keep <- times_warped[t_index] >= time_warped_L[x_index] &
    times_warped[t_index] < time_warped_H[x_index]
  x_to_keep <- x_index[to_keep]
  t_to_keep <- t_index[to_keep]
  t_ <- ((times_warped[t_to_keep] - time_warped_L[x_to_keep]) /
    Lambda[x_to_keep]) * dtime[x_to_keep] + time_L[x_to_keep]

  return(t_[!is.na(t_)])
}

test_that("draw_sc_step() arguments work", {
  set.seed(123)
  expect_no_error(df0 <- draw_sc_step(time_breaks = c(0, 10), lambda_vector = 1, atmost1 = TRUE))
  check_ppp_sample_validity(times = df0, t_min = 0, t_max = 10, atmost1 = TRUE)

  expect_no_error(df0 <- draw_sc_step(time_breaks = c(0:10), lambda_vector = runif(10), atmost1 = TRUE))
  check_ppp_sample_validity(times = df0, t_min = 0, t_max = 10, atmost1 = TRUE)

  expect_no_error(df1 <- draw_sc_step(time_breaks = c(0, 10), lambda_vector = 1))
  check_ppp_sample_validity(times = df1, t_min = 0, t_max = 10)

  expect_no_error(df1 <- draw_sc_step(time_breaks = c(0:10), lambda_vector = runif(10)))
  check_ppp_sample_validity(times = df1, t_min = 0, t_max = 10)

  expect_no_error(df1 <- draw_sc_step(time_breaks = c(0:10), lambda_vector = runif(10) / 1000, atleast1 = TRUE))
  check_ppp_sample_validity(times = df1, t_min = 0, t_max = 10, atleast1 = TRUE)

  expect_no_error(df1 <- draw_sc_step(time_breaks = c(0:10), lambda_vector = runif(10) / 1000, atmost1 = TRUE, atleast1 = TRUE))
  check_ppp_sample_validity(times = df1, t_min = 0, t_max = 10, atmost1 = TRUE, atleast1 = TRUE)
})

#' Simulate a homogeneous Poisson Point Process over (t_min, t_max]
#' @description
#'   `r lifecycle::badge("deprecated")`
#' Use `ppp` instead.
#' @param range_t (vector, double) min and max of the time interval
#' @param rate (scalar, double) constant instantaneous rate
#' @param tol the probability that we will have more than
#'        the drawn events in (t_min, t_max]
#' @param rng_stream an `rstream` object
#' @param atmost1 boolean, draw at most 1 event time
#'
#' @return a vector of event times t
#'         if no events realize, it will have 0 length
#' @export
#'
#' @keywords internal
#' @examples
#' x <- ppp_sequential(range_t = c(0, 10), rate = 1, tol = 10^-6)
ppp_sequential <- function(range_t = c(0, 10), rate = 1, tol = 10^-6, rng_stream = NULL, atmost1 = FALSE) {
  # we expect lambda = t_max*rate events
  # we will draw n so that the probability that t>t_max is 1-tol
  lifecycle::deprecate_warn(when = "0.5.0", what = "ppp_sequential()", with = "ppp()")
  if (isTRUE(atmost1)) {
    n <- 1
  } else {
    n <- stats::qpois(p = 1 - tol, lambda = rate * (range_t[2] - range_t[1]))
  }
  dt_ <- rng_stream_rexp(size = n, rate = rate, rng_stream = rng_stream)
  t_ <- cumsum(dt_) + range_t[1]
  return(t_[t_ <= range_t[2]])
}

#' Vectorized sampling from a non homogeneous Poisson Point Process (NHPPP) from
#'    an interval (thinning method) with piecewise constant_majorizers (R)
#'    -- but can be forced to sample from zero-truncated proposals.
#'
#'
#'
#' @description
#' Vectorized sampling from a non homogeneous Poisson Point Process (NHPPP) from
#'    an interval (thinning method) with piecewise constant_majorizers.
#'    The majorizers are step functions over equal-length time intevals.
#'    This function is used for obtainning proposals for `vztdraw_intensity_step_regular()`
#'
#' @param lambda (function) intensity function, vectorized
#' @param lambda_args (list) optional named list of arguments to pass to `lambda`.
#'        If you have arguments for `lambda` that vary by draw, they should be passed as
#'        a data.table named `vector_arguments`.
#' @param Lambda_maj_matrix (matrix) integrated intensity rates at the end of each interval
#' @param lambda_maj_matrix (matrix) intensity rates, one per interval
#' @param rate_matrix_t_min (scalar | vector | column matrix) is the lower bound
#'        of the time interval for each row of (Lambda|lambda)_maj_matrix.
#'        The length of this argument is the number of point processes that should be drawn.
#' @param rate_matrix_t_max (scalar | vector | column matrix) the upper bound
#'        of the time interval for each row of (Lambda|lambda)_maj_matrix.
#'        The length of this argument is the number of point processes that should be drawn.
#' @param t_min (scalar | vector | column matrix) is the lower bound
#'        of a subinterval of (rate_matrix_t_min, rate_matrix_t_max]. If set,
#'        times are sampled from the subinterval.
#'        If omitted, it is equivalent to `rate_matrix_t_min`.
#' @param t_max (scalar | vector | column matrix) is the upper bound
#'        of a subinterval of (rate_matrix_t_min, rate_matrix_t_max]. If set,
#'        times are sampled from the subinterval.
#'        If omitted, it is equivalent to `rate_matrix_t_max`.
#' @param tol (scalar, double) tolerance for the number of events
#' @param atmost1 boolean, draw at most 1 event time
#' @param force_zt_majorizer boolean, force the majorizer to be zero-truncated.
#'        This option is used when the function is called to make proposals for
#'        `vztdraw_intensity_step_regular()`. In general, do not set this option to `TRUE`.
#'
#' @keywords internal
vdraw_intensity_step_regular_forcezt <- function(
    lambda = NULL,
    lambda_args = NULL,
    Lambda_maj_matrix = NULL,
    lambda_maj_matrix = NULL,
    rate_matrix_t_min = NULL,
    rate_matrix_t_max = NULL,
    t_min = NULL,
    t_max = NULL,
    tol = 10^-6,
    atmost1 = FALSE,
    force_zt_majorizer = FALSE,
    ...) {
  #browser()
  if (!is.null(lambda_maj_matrix) && is.null(Lambda_maj_matrix)) {
    mode(lambda_maj_matrix) <- "numeric"
    rate <- lambda_maj_matrix
    is_cumulative_rate <- FALSE
  } else if (is.null(lambda_maj_matrix) && !is.null(Lambda_maj_matrix)) {
    mode(Lambda_maj_matrix) <- "numeric"
    rate <- Lambda_maj_matrix
    is_cumulative_rate <- TRUE
  } else {
    stop("lambda_maj_matrix and Lambda_maj_matrix cannot both be `NULL`")
  }

  num_na <- sum(is.na(rate))
  if (num_na > 0) {
    rate_argument <- if (is_cumulative_rate) "Lambda_maj_matrix" else "lambda_maj_matrix"
    stop("The ", rate_argument, " contains ", num_na, " NA values")
  }

  n_intervals <- ncol(rate)
  n_draws <- nrow(rate)


  range_t <- cbind(as.vector(rate_matrix_t_min), as.vector(rate_matrix_t_max))
  if (nrow(range_t) > 1 && nrow(range_t) != nrow(rate)) {
    stop("The (rows of) [Lambda|lambda]_maj_matrix and (length of) [rate_matrix_t_min|rate_matrix_t_max] imply different numbers of point processes to be sampled.")
  }
  if (nrow(range_t) == 1 && nrow(rate) != 1) {
    range_t <- range_t[rep(1, nrow(rate)), ]
  }
  interval_duration <- (range_t[, 2] - range_t[, 1]) / n_intervals

  if (is_cumulative_rate) {
    lambda_maj_matrix <- matrix(0, ncol = n_intervals, nrow = n_draws)
    lambda_maj_matrix[, 1] <- Lambda_maj_matrix[, 1] / interval_duration
    for (col in 2:n_intervals) {
      lambda_maj_matrix[, col] <- (Lambda_maj_matrix[, col] - Lambda_maj_matrix[, col - 1]) / interval_duration
    }
  }

  if (force_zt_majorizer) {
    Z_star <- vztdraw_sc_step_regular_cpp(
      lambda_matrix = lambda_maj_matrix,
      rate_matrix_t_min = rate_matrix_t_min,
      rate_matrix_t_max = rate_matrix_t_max,
      t_min = t_min,
      t_max = t_max,
      atmost1 = FALSE
    )
  } else {
    Z_star <- vdraw_sc_step_regular_cpp(
      lambda_matrix = lambda_maj_matrix,
      rate_matrix_t_min = rate_matrix_t_min,
      rate_matrix_t_max = rate_matrix_t_max,
      t_min = t_min,
      t_max = t_max,
      tol = tol,
      atmost1 = FALSE
    )
  }

  n_max_events <- ncol(Z_star)
  U <- matrix(stats::runif(length(Z_star)), ncol = n_max_events)

  # Extracts correct majorising values
  lambda_maj_idx <- ceiling((Z_star - range_t[, rep(1, n_max_events)]) / interval_duration)

  # browser()
  idx_adjusted <- as.vector(lambda_maj_idx + (1:nrow(lambda_maj_matrix) - 1) * ncol(lambda_maj_matrix))
  lambda_maj <- matrix(
    t(lambda_maj_matrix)[idx_adjusted],
    nrow = nrow(lambda_maj_idx), ncol = ncol(lambda_maj_idx), byrow = FALSE
  )

  #browser()

  accept <- ifelse(lambda(Z_star, lambda_args) / lambda_maj > U, TRUE, NA)
  Z <- Z_star * accept
  Z_sorted <- matrix(
    Z[order(row(Z), is.na(Z), method = "radix")], # shifts non-NAs to the left
    nrow = nrow(Z), ncol = ncol(Z), byrow = TRUE
  )

  if (ncol(Z_sorted) > 1) {
    Z_sorted <- Z_sorted[, colSums(!is.na(Z_sorted)) > 0, drop = FALSE] # removes empty columns after the shift
  }

  if (atmost1) {
    return(Z_sorted[, 1, drop = FALSE])
  } else {
    return(Z_sorted)
  }
}

## add tests that have lambda_maj < lambda
## add tests that have lamda_args

test_that("vdraw_intensity_step_regular_forcezt() works", {
  set.seed(123)
  lfun <- function(x, lambda_args, ...) .2 * x^lambda_args$exponent
  l_args <- list(exponent = 1L)
  lmaj <- matrix(rep(1, 50), ncol = 5)
  Lmaj <- mat_cumsum_columns(lmaj)

  expect_no_error(Z <- vdraw_intensity_step_regular_forcezt(
    lambda = lfun,
    lambda_args = l_args,
    lambda_maj_matrix = lmaj,
    rate_matrix_t_min = 1, rate_matrix_t_max = 5,
    tol = 10^-6,
    atmost1 = FALSE
  ))
  check_ppp_sample_validity(Z, t_min = 1, t_max = 5)

  expect_no_error(Z <- vdraw_intensity_step_regular_forcezt(
    lambda = lfun,
    lambda_args = l_args,
    Lambda_maj_matrix = Lmaj,
    rate_matrix_t_min = 1, rate_matrix_t_max = 5,
    tol = 10^-6,
    atmost1 = FALSE
  ))
  check_ppp_sample_validity(Z, t_min = 1, t_max = 5)

  expect_no_error(Z <- vdraw_intensity_step_regular_forcezt(
    lambda = lfun,
    lambda_args = l_args,
    lambda_maj_matrix = lmaj,
    rate_matrix_t_min = 1, rate_matrix_t_max = 5,
    tol = 10^-6,
    atmost1 = TRUE
  ))
  check_ppp_sample_validity(Z, t_min = 1, t_max = 5, atmost1 = TRUE)

  expect_no_error(Z <- vdraw_intensity_step_regular_forcezt(
    lambda = lfun,
    lambda_args = l_args,
    Lambda_maj_matrix = Lmaj,
    rate_matrix_t_min = 1, rate_matrix_t_max = 5,
    tol = 10^-6,
    atmost1 = TRUE
  ))
  check_ppp_sample_validity(Z, t_min = 1, t_max = 5, atmost1 = TRUE)
})



test_that("vdraw_intensity_step_regular_forcezt() does not break with matrices whose mode is list", {
  set.seed(123)
  lfun <- function(x, ...) .2 * x
  lmaj <- matrix(rep(1, 50), ncol = 5)
  Lmaj <- mat_cumsum_columns(lmaj)

  mode(Lmaj) <- "list"
  expect_no_error(Z <- vdraw_intensity_step_regular_forcezt(
    lambda = lfun,
    Lambda_maj_matrix = Lmaj,
    rate_matrix_t_min = 1, rate_matrix_t_max = 5,
    tol = 10^-6,
    atmost1 = FALSE
  ))
  check_ppp_sample_validity(Z, t_min = 1, t_max = 5, atmost1 = FALSE)
  mode(lmaj) <- "list"
  expect_no_error(Z <- vdraw_intensity_step_regular_forcezt(
    lambda = lfun,
    lambda_maj_matrix = lmaj,
    rate_matrix_t_min = 1, rate_matrix_t_max = 5,
    tol = 10^-6,
    atmost1 = FALSE
  ))
  check_ppp_sample_validity(Z, t_min = 1, t_max = 5, atmost1 = FALSE)
})


test_that("vdraw_intensity_step_regular_forcezt() works with different majorizers", {
  set.seed(123)
  lfun <- function(x, lambda_args, ...) .2 * x^lambda_args$exponent
  l_args <- list(exponent = 1L)
  lmaj <- matrix(rep(1, 1000), ncol = 5)

  expect_no_error(Z1 <- vdraw_intensity_step_regular_forcezt(
    lambda = lfun,
    lambda_args = l_args,
    lambda_maj_matrix = lmaj,
    rate_matrix_t_min = 1, rate_matrix_t_max = 5,
    tol = 10^-6,
    atmost1 = FALSE
  ))
  check_ppp_sample_validity(Z1, t_min = 1, t_max = 5)

  expect_no_error(Z2 <- vdraw_intensity_step_regular_forcezt(
    lambda = lfun,
    lambda_args = l_args,
    lambda_maj_matrix = lmaj + 10,
    rate_matrix_t_min = 1, rate_matrix_t_max = 5,
    tol = 10^-6,
    atmost1 = FALSE
  ))
  check_ppp_sample_validity(Z2, t_min = 1, t_max = 5)

  compare_ppp_vectors(ppp1 = Z1, ppp2 = Z2, threshold = 0.1, showQQ = TRUE)
})

#' Definite integral of `l = intercept + slope*t` at time `t`
#' with `L(t0) = 0`
#'
#' @description Definite integral of `l = intercept + slope*t` starting at
#'              `t0` -- only for `l+`.
#' @param t (double) the time point
#' @param intercept (double) the intercept
#' @param slope (double) the slope
#' @param t0 (double) the starting time
#' @keywords internal
Lambda_linear_form <- function(t, intercept, slope, t0) {
  if (min(t) < t0) stop()
  if (slope < 0) {
    t[t > -intercept / slope] <- -intercept / slope
  }
  return(intercept * (t - t0) + (slope / 2) * (t^2 - t0^2))
}

#' Inverse of the definite integral of `l = intercept + slope*t` at time `t`
#'
#' @description Inverse of the definite integral of `l = intercept + slope*t` only for `l+`.
#' @param z (double) the value of integrated rate for which you want to find the time
#' @param intercept (double) the intercept
#' @param slope (double) the slope
#' @param t0 (double) the starting time
#' @keywords internal
Lambda_inv_linear_form <- function(z, intercept, slope, t0) {
  if (slope == 0 & intercept == 0) stop()
  if (slope != 0) {
    L0 <- -intercept * t0 - slope / 2 * t0^2
    Delta <- intercept^2 - 2 * slope * (L0 - z)
    if (!all(Delta >= 0)) stop()
    t_ <- (-intercept + sqrt(Delta)) / slope
  } else if (slope == 0) {
    t_ <- z / intercept + t0
  }
  return(t_)
}

#' Definite integral of `l = exp(intercept + slope*t)` at time `t`
#' with `L(t0) = 0`
#'
#' @description Definite integral of `l = exp(intercept + slope*t)` starting at
#'              `t0` -- only for `l+`.
#' @param t (double) the time point
#' @param intercept (double) the intercept
#' @param slope (double) the slope
#' @param t0 (double) the starting time
#' @keywords internal
Lambda_exp_form <- function(t, intercept, slope, t0) {
  if (min(t) < t0) stop()
  return((exp(slope * t + intercept) - exp(slope * t0 + intercept)) / slope)
}

#' Inverse of the definite integral of `l = exp(intercept + slope*t)` at time `t`
#'
#' @description Inverse of the definite integral of `l = exp(intercept + slope*t)` only for `l+`.
#' @param z (double) the value of integrated rate for which you want to find the time
#' @param intercept (double) the intercept
#' @param slope (double) the slope
#' @param t0 (double) the starting time
#' @keywords internal
Lambda_inv_exp_form <- function(z, intercept, slope, t0) {
  tmp <- exp(slope * t0 + intercept)
  return((log(tmp + z * slope) - intercept) / slope)
}

test_that("ztppp agrees with ppp for high intensities", {
  set.seed(123)
  # High intensity, there are practically no zero event trajectories
  r_ppp_t <- unlist(lapply(integer(10000), function(x) ppp(t_min = 1, t_max = 3, rate = 10, atmost1 = TRUE)))
  r_ztppp <- unlist(lapply(integer(10000), function(x) ztppp(t_min = 1, t_max = 3, rate = 10, atmost1 = TRUE)))
  compare_ppp_vectors(ppp1 = r_ppp_t, ppp2 = r_ztppp, threshold = 0.1, showQQ = TRUE)
})

test_that("ztppp agrees with ppp for low intensities", {
  set.seed(123)
  # Low intensity, there are many zero event trajectories -- condition on having >=1 event
  r_ppp_t_low <- unlist(lapply(integer(10000), function(x) ppp(t_min = 1, t_max = 3, rate = 0.1, atmost1 = TRUE)))
  r_ztppp_low <- unlist(lapply(integer(length(r_ppp_t_low)), function(x) ztppp(t_min = 1, t_max = 3, rate = 0.1, atmost1 = TRUE)))
  compare_ppp_vectors(ppp1 = r_ppp_t_low, ppp2 = r_ztppp_low, threshold = 0.1, showQQ = TRUE)
})

test_that("ztNHPPP methods agree on the first time to event with constant rate", {
  set.seed(123)
  l <- function(t) rep(2, length(t))
  L <- function(t) 2 * t
  Li <- function(z) z / 2

  r_ztppp <- unlist(lapply(integer(10000), function(x) ztppp(t_min = 1, t_max = 13, rate = 2, atmost1 = TRUE)))
  r_ztnhppp_ci <- unlist(lapply(integer(10000), function(x) ztdraw_cumulative_intensity(Lambda = L, Lambda_inv = Li, t_min = 1, t_max = 13, atmost1 = TRUE)))
  compare_ppp_vectors(ppp1 = r_ztnhppp_ci, ppp2 = r_ztppp, threshold = 0.1, showQQ = TRUE)

  r_ztnhppp_intens_linear <- unlist(lapply(integer(10000), function(x) ztdraw_sc_linear(intercept = 2, slope = 0, t_min = 1, t_max = 13, atmost1 = TRUE)))
  compare_ppp_vectors(ppp1 = r_ztnhppp_intens_linear, ppp2 = r_ztppp, threshold = 0.1, showQQ = TRUE)

  r_ztnhppp_intens_exp <- unlist(lapply(integer(10000), function(x) ztdraw_sc_loglinear(intercept = log(2), slope = 0, t_min = 1, t_max = 13, atmost1 = TRUE)))
  compare_ppp_vectors(ppp1 = r_ztnhppp_intens_exp, ppp2 = r_ztppp, threshold = 0.1, showQQ = TRUE)

  r_ztnhppp_intens <- unlist(lapply(integer(10000), function(x) ztdraw_intensity(lambda = l, line_majorizer_intercept = 2.1, line_majorizer_slope = 0, t_min = 1, t_max = 13, atmost1 = TRUE)))
  compare_ppp_vectors(ppp1 = r_ztnhppp_intens, ppp2 = r_ztppp, threshold = 0.1, showQQ = TRUE)

  r_ztnhppp_intens_piecewise <- unlist(lapply(integer(10000), function(x) ztdraw_intensity_step(lambda = l, majorizer_vector = l(2:13), time_breaks = c(1:13), atmost1 = TRUE)))
  compare_ppp_vectors(ppp1 = r_ztnhppp_intens, ppp2 = r_ztppp, threshold = 0.1, showQQ = TRUE)
})

test_that("ztNHPPP linear intensity agrees with general methods", {
  set.seed(123)
  l <- function(t, intercept = 1, slope = 2) intercept + slope * t
  L <- function(t, intercept = 1, slope = 2, t0 = 1) Lambda_linear_form(t, intercept = intercept, slope = slope, t0 = t0)
  Li <- function(z, intercept = 1, slope = 2, t0 = 1) Lambda_inv_linear_form(z, intercept = intercept, slope = slope, t0 = t0)

  r_ztnhppp_intens_linear <- unlist(lapply(integer(10000), function(x) ztdraw_sc_linear(intercept = 1, slope = 2, t_min = 1, t_max = 13, atmost1 = TRUE)))
  r_ztnhppp_ci <- unlist(lapply(integer(10000), function(x) ztdraw_cumulative_intensity(Lambda = L, Lambda_inv = Li, t_min = 1, t_max = 13, atmost1 = TRUE)))
  compare_ppp_vectors(ppp1 = r_ztnhppp_intens_linear, ppp2 = r_ztnhppp_ci, threshold = 0.1, showQQ = TRUE)

  r_ztnhppp_intens <- unlist(lapply(integer(10000), function(x) ztdraw_intensity(lambda = l, line_majorizer_intercept = l(13), line_majorizer_slope = 0, t_min = 1, t_max = 13, atmost1 = TRUE)))
  compare_ppp_vectors(ppp1 = r_ztnhppp_intens_linear, ppp2 = r_ztnhppp_intens, threshold = 0.1, showQQ = TRUE)
})


test_that("ztNHPPP linear intensity agrees with linear intensity for high rates", {
  set.seed(123)
  r_ztnhppp_intens_linear <- unlist(lapply(integer(10000), function(x) ztdraw_sc_linear(intercept = 1, slope = 2, t_min = 1, t_max = 13, atmost1 = TRUE)))
  r_nhppp_intens_linear <- unlist(lapply(integer(10000), function(x) draw_sc_linear(intercept = 1, slope = 2, t_min = 1, t_max = 13, atmost1 = TRUE)))
  compare_ppp_vectors(ppp1 = r_ztnhppp_intens_linear, ppp2 = r_nhppp_intens_linear, threshold = 0.1, showQQ = TRUE)
})

test_that("ztNHPPP linear intensity agrees with linear intensity for low rates", {
  set.seed(123)
  r_nhppp_intens_linear <- unlist(lapply(integer(10000), function(x) draw_sc_linear(intercept = 1, slope = 2, t_min = 1, t_max = 1.1, atmost1 = TRUE)))
  r_ztnhppp_intens_linear <- unlist(lapply(integer(length(r_nhppp_intens_linear)), function(x) ztdraw_sc_linear(intercept = 1, slope = 2, t_min = 1, t_max = 1.1, atmost1 = TRUE)))
  compare_ppp_vectors(ppp1 = r_ztnhppp_intens_linear, ppp2 = r_nhppp_intens_linear, threshold = 0.15, showQQ = TRUE)
})


test_that("vztdraw_sc_step_regular() agrees with vdraw_sc_step_regular()", {
  set.seed(123)
  Lmat <- matrix(rep(c(1, 11, 14, 17), 10000), ncol = 4, byrow = TRUE)
  r_vdraw_sc_step_regular <- vdraw_sc_step_regular(Lambda_matrix = Lmat, rate_matrix_t_min = 1, rate_matrix_t_max = 5, atmost1 = FALSE)
  r_vdraw_sc_step_regular <- r_vdraw_sc_step_regular[!is.na(r_vdraw_sc_step_regular)]
  r_vztdraw_sc_step_regular <- vztdraw_sc_step_regular_cpp(Lambda_matrix = Lmat, rate_matrix_t_min = 1, rate_matrix_t_max = 5, atmost1 = FALSE)
  r_vztdraw_sc_step_regular <- r_vztdraw_sc_step_regular[!is.na(r_vztdraw_sc_step_regular)]
  compare_ppp_vectors(ppp1 = r_vdraw_sc_step_regular, ppp2 = r_vztdraw_sc_step_regular, threshold = 0.1, showQQ = TRUE)

  r_vdraw_sc_step_regular1 <- vdraw_sc_step_regular(Lambda_matrix = Lmat, rate_matrix_t_min = 1, rate_matrix_t_max = 5, atmost1 = TRUE)
  r_vdraw_sc_step_regular1 <- r_vdraw_sc_step_regular1[!is.na(r_vdraw_sc_step_regular1)]
  r_vztdraw_sc_step_regular1 <- vztdraw_sc_step_regular_cpp(Lambda_matrix = Lmat, rate_matrix_t_min = 1, rate_matrix_t_max = 5, atmost1 = TRUE)
  r_vztdraw_sc_step_regular1 <- r_vztdraw_sc_step_regular1[!is.na(r_vztdraw_sc_step_regular1)]
  compare_ppp_vectors(ppp1 = r_vdraw_sc_step_regular1, ppp2 = r_vztdraw_sc_step_regular1, threshold = 0.1, showQQ = TRUE)
})


test_that("vztdraw_sc_step_regular_cpp() agrees with vdraw_sc_step_regular()", {
  set.seed(123)
  Lmat <- matrix(rep(c(1, 11, 14, 17), 10000), ncol = 4, byrow = TRUE)
  r_vdraw_sc_step_regular <- vdraw_sc_step_regular(Lambda_matrix = Lmat, rate_matrix_t_min = 1, rate_matrix_t_max = 5, atmost1 = FALSE)
  r_vdraw_sc_step_regular <- r_vdraw_sc_step_regular[!is.na(r_vdraw_sc_step_regular)]
  r_vztdraw_sc_step_regular <- vztdraw_sc_step_regular_cpp(Lambda_matrix = Lmat, rate_matrix_t_min = 1, rate_matrix_t_max = 5, atmost1 = FALSE)
  r_vztdraw_sc_step_regular <- r_vztdraw_sc_step_regular[!is.na(r_vztdraw_sc_step_regular)]
  compare_ppp_vectors(ppp1 = r_vdraw_sc_step_regular, ppp2 = r_vztdraw_sc_step_regular, threshold = 0.1, showQQ = TRUE)

  r_vdraw_sc_step_regular1 <- vdraw_sc_step_regular(Lambda_matrix = Lmat, rate_matrix_t_min = 1, rate_matrix_t_max = 5, atmost1 = TRUE)
  r_vdraw_sc_step_regular1 <- r_vdraw_sc_step_regular1[!is.na(r_vdraw_sc_step_regular1)]
  r_vztdraw_sc_step_regular1 <- vztdraw_sc_step_regular_cpp(Lambda_matrix = Lmat, rate_matrix_t_min = 1, rate_matrix_t_max = 5, atmost1 = TRUE)
  r_vztdraw_sc_step_regular1 <- r_vztdraw_sc_step_regular1[!is.na(r_vztdraw_sc_step_regular1)]
  compare_ppp_vectors(ppp1 = r_vdraw_sc_step_regular1, ppp2 = r_vztdraw_sc_step_regular1, threshold = 0.1, showQQ = TRUE)
})



test_that("ztNHPPP exponential agrees with general methods", {
  set.seed(123)
  l <- function(t, intercept = .1, slope = .02) exp(intercept + slope * t)
  L <- function(t, intercept = .1, slope = .02, t0 = 1) Lambda_exp_form(t, intercept = intercept, slope = slope, t0 = t0)
  Li <- function(z, intercept = .1, slope = .02, t0 = 1) Lambda_inv_exp_form(z, intercept = intercept, slope = slope, t0 = t0)

  r_ztnhppp_intens_exp <- unlist(lapply(integer(10000), function(x) ztdraw_sc_loglinear(intercept = .1, slope = .02, t_min = 1, t_max = 13, atmost1 = TRUE)))
  r_ztnhppp_ci <- unlist(lapply(integer(10000), function(x) ztdraw_cumulative_intensity(Lambda = L, Lambda_inv = Li, t_min = 1, t_max = 13, atmost1 = TRUE)))
  compare_ppp_vectors(ppp1 = r_ztnhppp_intens_exp, ppp2 = r_ztnhppp_ci, threshold = 0.1, showQQ = TRUE)

  r_ztnhppp_intens <- unlist(lapply(integer(10000), function(x) ztdraw_intensity(lambda = l, line_majorizer_intercept = l(13), line_majorizer_slope = 0, t_min = 1, t_max = 13, atmost1 = TRUE)))
  compare_ppp_vectors(ppp1 = r_ztnhppp_intens_exp, ppp2 = r_ztnhppp_intens, threshold = 0.1, showQQ = TRUE)
})


test_that("ztNHPPP exponential agrees with NHPPP exponential for high rates", {
  set.seed(123)
  r_nhppp_intens_exp <- unlist(lapply(integer(10000), function(x) draw_sc_loglinear(intercept = .1, slope = .02, t_min = 1, t_max = 13, atmost1 = TRUE)))
  r_ztnhppp_intens_exp <- unlist(lapply(integer(10000), function(x) ztdraw_sc_loglinear(intercept = .1, slope = .02, t_min = 1, t_max = 13, atmost1 = TRUE)))
  compare_ppp_vectors(ppp1 = r_nhppp_intens_exp, ppp2 = r_ztnhppp_intens_exp, threshold = 0.1, showQQ = TRUE)
})


test_that("ztNHPPP exponential agrees with NHPPP exponential for low rates", {
  set.seed(123)
  r_nhppp_intens_exp <- unlist(lapply(integer(10000), function(x) draw_sc_loglinear(intercept = .1, slope = .02, t_min = 1, t_max = 1.1, atmost1 = TRUE)))
  r_ztnhppp_intens_exp <- unlist(lapply(integer(length(r_nhppp_intens_exp)), function(x) ztdraw_sc_loglinear(intercept = .1, slope = .02, t_min = 1, t_max = 1.1, atmost1 = TRUE)))
  compare_ppp_vectors(ppp1 = r_nhppp_intens_exp, ppp2 = r_ztnhppp_intens_exp, threshold = 0.1, showQQ = TRUE)
})

test_that("Lambda_linear_form works", {
  expect_no_error(L_t <- Lambda_linear_form(t = 1:10, intercept = 0, slope = 0, t0 = 1))
  expect_identical(L_t, rep(0, 10))

  expect_no_error(L_t <- Lambda_linear_form(t = 1:10, intercept = 1, slope = 0, t0 = 1))
  expect_identical(L_t, as.double(0:9))

  expect_no_error(L_t <- Lambda_linear_form(t = 1:10, intercept = 1, slope = 1, t0 = 1))
  expect_identical(L_t, 0:9 + (1:10)^2 / 2 - 1 / 2)
})


test_that("Lambda_inv_linear_form works", {
  L_t <- Lambda_linear_form(t = 1:10, intercept = 0, slope = 1, t0 = 1)
  expect_identical(
    as.double(1:10),
    Lambda_inv_linear_form(z = L_t, intercept = 0, slope = 1, t0 = 1)
  )

  L_t <- Lambda_linear_form(t = 1:10, intercept = 1, slope = 0, t0 = 1)
  expect_identical(
    as.double(1:10),
    Lambda_inv_linear_form(z = L_t, intercept = 1, slope = 0, t0 = 1)
  )

  L_t <- Lambda_linear_form(t = 1:10, intercept = 1, slope = 1, t0 = 1)
  expect_identical(
    as.double(1:10),
    Lambda_inv_linear_form(z = L_t, intercept = 1, slope = 1, t0 = 1)
  )

  expect_error(
    Lambda_inv_linear_form(z = rep(0, 10), intercept = 0, slope = 0, t0 = 1)
  )
})

test_that("draw_sc_linear() works", {
  withr::with_seed(12345, df1 <- draw_sc_linear(intercept = 2, slope = 0, t_min = 1, t_max = 10))
  check_ppp_sample_validity(times = df1, t_min = 1, t_max = 10)

  withr::with_seed(12345, df1 <- draw_sc_linear(intercept = 0, slope = 1, t_min = 1, t_max = 10))
  check_ppp_sample_validity(times = df1, t_min = 1, t_max = 10)

  withr::with_seed(12345, df1 <- draw_sc_linear(intercept = 10, slope = -2, t_min = 1, t_max = 10))
  check_ppp_sample_validity(times = df1, t_min = 1, t_max = 10)

  withr::with_seed(12345, df1 <- draw_sc_linear(intercept = 2, slope = 0, t_min = 1, t_max = 10))
  withr::with_seed(12345, df2 <- ppp2(rate = 2, t_min = 1, t_max = 10))
  expect_identical(df1, df2)

  withr::with_seed(12345, df1 <- draw_sc_linear(intercept = 0, slope = 1, t_min = 1, t_max = 10))
  withr::with_seed(12345, df2 <- draw_cumulative_intensity_inversion(
    Lambda = function(t) Lambda_linear_form(t, intercept = 0, slope = 1, t0 = 1),
    Lambda_inv = function(z) Lambda_inv_linear_form(z, intercept = 0, slope = 1, t0 = 1),
    t_min = 1, t_max = 10
  ))
  expect_identical(round(df1, 3), round(df2, 3))

  withr::with_seed(12345, df1 <- draw_sc_linear(intercept = 10, slope = -2, t_min = 1, t_max = 10))
  withr::with_seed(12345, df2 <- draw_cumulative_intensity_inversion(
    Lambda = function(t) Lambda_linear_form(t, intercept = 10, slope = -2, t0 = 1),
    Lambda_inv = function(z) Lambda_inv_linear_form(z, intercept = 10, slope = -2, t0 = 1),
    t_min = 1, t_max = 5
  ))
  expect_identical(round(df1, 3), round(df2, 3))
})

test_that("PPP methods agree on the first time to event", {
  set.seed(123)
  r_ppp_next_n <- unlist(lapply(integer(10000), function(x) ppp_next_n(n = 1, rate = 10, t_min = 1)))
  r_ppp <- unlist(lapply(integer(10000), function(x) ppp(t_min = 1, t_max = 3, rate = 10, atmost1 = TRUE)))
  compare_ppp_vectors(ppp1 = r_ppp_next_n, ppp2 = r_ppp, threshold = 0.1, showQQ = TRUE)

  r_ppp2 <- unlist(lapply(integer(10000), function(x) ppp2(t_min = 1, t_max = 3, rate = 10, atmost1 = TRUE)))
  compare_ppp_vectors(ppp1 = r_ppp_next_n, ppp2 = r_ppp2, threshold = 0.1, showQQ = TRUE)

  r_draw_sc_step <- unlist(lapply(integer(10000), function(x) draw_sc_step(time_breaks = c(1, 2, 3), lambda_vector = rep(10, 2), atmost1 = TRUE)))
  compare_ppp_vectors(ppp1 = r_ppp_next_n, ppp2 = r_draw_sc_step, threshold = 0.1, showQQ = TRUE)
})


test_that("draw_sc_step() agrees with strung together constant rates", {
  set.seed(123)
  r_ppp <- unlist(lapply(
    integer(10000),
    function(x) {
      c(
        ppp(t_min = 1, t_max = 2, rate = 1, atmost1 = FALSE),
        ppp(t_min = 2, t_max = 2.2, rate = 10, atmost1 = FALSE),
        ppp(t_min = 2.2, t_max = 3, rate = 3, atmost1 = FALSE)
      )
    }
  ))

  r_draw_sc_step <- unlist(lapply(integer(10000), function(x) draw_sc_step(time_breaks = c(1, 2, 2.2, 3), lambda_vector = c(1, 10, 3), atmost1 = FALSE)))
  compare_ppp_vectors(ppp1 = r_ppp, ppp2 = r_draw_sc_step, threshold = 0.1, showQQ = TRUE)


  r_ppp1 <- unlist(lapply(
    integer(10000),
    function(x) {
      c(
        ppp(t_min = 1, t_max = 2, rate = 1, atmost1 = TRUE),
        ppp(t_min = 2, t_max = 2.2, rate = 10, atmost1 = TRUE),
        ppp(t_min = 2.2, t_max = 3, rate = 3, atmost1 = TRUE)
      )[1]
    }
  ))
  r_draw_sc_step1 <- unlist(lapply(integer(10000), function(x) draw_sc_step(time_breaks = c(1, 2, 2.2, 3), lambda_vector = c(1, 10, 3), atmost1 = TRUE)))
  compare_ppp_vectors(ppp1 = r_ppp1, ppp2 = r_draw_sc_step1, threshold = 0.1, showQQ = TRUE)
})


test_that("draw_sc_step_regular() agrees with strung together constant rates", {
  set.seed(123)
  r_ppp <- unlist(lapply(
    integer(10000),
    function(x) {
      c(
        ppp(t_min = 1, t_max = 2, rate = 1, atmost1 = FALSE),
        ppp(t_min = 2, t_max = 3, rate = 10, atmost1 = FALSE),
        ppp(t_min = 3, t_max = 5, rate = 3, atmost1 = FALSE)
      )
    }
  ))

  L <- c(1, 11, 14, 17)

  r_ppp_regularstep <- unlist(lapply(integer(10000), function(x) draw_sc_step_regular(Lambda_vector = L, t_min = 1, t_max = 5, atmost1 = FALSE)))
  compare_ppp_vectors(ppp1 = r_ppp, ppp2 = r_ppp_regularstep, threshold = 0.1, showQQ = TRUE)


  r_ppp1 <- unlist(lapply(
    integer(10000),
    function(x) {
      c(
        ppp(t_min = 1, t_max = 2, rate = 1, atmost1 = TRUE),
        ppp(t_min = 2, t_max = 3, rate = 10, atmost1 = TRUE),
        ppp(t_min = 3, t_max = 5, rate = 3, atmost1 = TRUE)
      )[1]
    }
  ))
  r_ppp_regularstep1 <- unlist(lapply(integer(10000), function(x) draw_sc_step_regular(Lambda_vector = L, t_min = 1, t_max = 5, atmost1 = TRUE)))
  compare_ppp_vectors(ppp1 = r_ppp1, ppp2 = r_ppp_regularstep1, threshold = 0.1, showQQ = TRUE)
})

test_that("vdraw_sc_step_regular() agrees with strung together constant rates", {
  set.seed(123)
  r_ppp <- unlist(lapply(
    integer(10000),
    function(x) {
      c(
        ppp(t_min = 1, t_max = 2, rate = 1, atmost1 = FALSE),
        ppp(t_min = 2, t_max = 3, rate = 10, atmost1 = FALSE),
        ppp(t_min = 3, t_max = 5, rate = 3, atmost1 = FALSE)
      )
    }
  ))

  Lmat <- matrix(rep(c(1, 11, 14, 17), 10000), ncol = 4, byrow = TRUE)

  r_vdraw_sc_step_regular <- vdraw_sc_step_regular(Lambda_matrix = Lmat, rate_matrix_t_min = 1, rate_matrix_t_max = 5, atmost1 = FALSE)
  r_vdraw_sc_step_regular <- r_vdraw_sc_step_regular[!is.na(r_vdraw_sc_step_regular)]
  compare_ppp_vectors(ppp1 = r_ppp, ppp2 = r_vdraw_sc_step_regular, threshold = 0.1, showQQ = TRUE)

  r_ppp1 <- unlist(lapply(
    integer(10000),
    function(x) {
      c(
        ppp(t_min = 1, t_max = 2, rate = 1, atmost1 = TRUE),
        ppp(t_min = 2, t_max = 3, rate = 10, atmost1 = TRUE),
        ppp(t_min = 3, t_max = 5, rate = 3, atmost1 = TRUE)
      )[1]
    }
  ))
  r_vdraw_sc_step_regular1 <- vdraw_sc_step_regular(Lambda_matrix = Lmat, rate_matrix_t_min = 1, rate_matrix_t_max = 5, atmost1 = TRUE)
  r_vdraw_sc_step_regular1 <- r_vdraw_sc_step_regular1[!is.na(r_vdraw_sc_step_regular1)]
  compare_ppp_vectors(ppp1 = r_ppp1, ppp2 = r_vdraw_sc_step_regular1, threshold = 0.1, showQQ = TRUE)
})



test_that("vdraw_sc_step_regular_cpp() agrees with strung together constant rates", {
  set.seed(123)
  r_ppp <- unlist(lapply(
    integer(10000),
    function(x) {
      c(
        ppp(t_min = 1, t_max = 2, rate = 1, atmost1 = FALSE),
        ppp(t_min = 2, t_max = 3, rate = 10, atmost1 = FALSE),
        ppp(t_min = 3, t_max = 5, rate = 3, atmost1 = FALSE)
      )
    }
  ))

  Lmat <- matrix(rep(c(1, 11, 14, 17), 10000), ncol = 4, byrow = TRUE)

  r_vdraw_sc_step_regular <- vdraw_sc_step_regular_cpp(Lambda_matrix = Lmat, rate_matrix_t_min = 1, rate_matrix_t_max = 5, atmost1 = FALSE)
  r_vdraw_sc_step_regular <- r_vdraw_sc_step_regular[!is.na(r_vdraw_sc_step_regular)]
  compare_ppp_vectors(ppp1 = r_ppp, ppp2 = r_vdraw_sc_step_regular, threshold = 0.1, showQQ = TRUE)

  r_ppp1 <- unlist(lapply(
    integer(10000),
    function(x) {
      c(
        ppp(t_min = 1, t_max = 2, rate = 1, atmost1 = TRUE),
        ppp(t_min = 2, t_max = 3, rate = 10, atmost1 = TRUE),
        ppp(t_min = 3, t_max = 5, rate = 3, atmost1 = TRUE)
      )[1]
    }
  ))
  r_vdraw_sc_step_regular1 <- vdraw_sc_step_regular_cpp(Lambda_matrix = Lmat, rate_matrix_t_min = 1, rate_matrix_t_max = 5, atmost1 = TRUE)
  r_vdraw_sc_step_regular1 <- r_vdraw_sc_step_regular1[!is.na(r_vdraw_sc_step_regular1)]
  compare_ppp_vectors(ppp1 = r_ppp1, ppp2 = r_vdraw_sc_step_regular1, threshold = 0.1, showQQ = TRUE)
})


test_that("NHPPP methods agree on the first time to event with constant rate", {
  set.seed(123)
  l <- function(t) 2
  L <- function(t) 2 * t
  Li <- function(z) z / 2

  r_ppp <- unlist(lapply(integer(10000), function(x) ppp(t_min = 1, t_max = 13, rate = 2, atmost1 = TRUE)))
  r_nhppp_ci_inv <- unlist(lapply(integer(10000), function(x) draw_cumulative_intensity_inversion(Lambda = L, Lambda_inv = Li, t_min = 1, t_max = 13, atmost1 = TRUE)))
  compare_ppp_vectors(ppp1 = r_nhppp_ci_inv, ppp2 = r_ppp, threshold = 0.1, showQQ = TRUE)

  r_nhppp_ci_orderstats <- unlist(lapply(integer(10000), function(x) draw_cumulative_intensity_orderstats(Lambda = L, Lambda_inv = Li, t_min = 1, t_max = 13, atmost1 = TRUE)))
  compare_ppp_vectors(ppp1 = r_nhppp_ci_orderstats, ppp2 = r_ppp, threshold = 0.1, showQQ = TRUE)

  r_nhppp_intens_linear <- unlist(lapply(integer(10000), function(x) draw_sc_linear(intercept = 2, slope = 0, t_min = 1, t_max = 13, atmost1 = TRUE)))
  compare_ppp_vectors(ppp1 = r_nhppp_intens_linear, ppp2 = r_ppp, threshold = 0.1, showQQ = TRUE)

  r_nhppp_intens_exp <- unlist(lapply(integer(10000), function(x) draw_sc_loglinear(intercept = log(2), slope = 0, t_min = 1, t_max = 13, atmost1 = TRUE)))
  compare_ppp_vectors(ppp1 = r_nhppp_intens_exp, ppp2 = r_ppp, threshold = 0.1, showQQ = TRUE)

  r_nhppp_intens <- unlist(lapply(integer(10000), function(x) draw_intensity_line(lambda = l, majorizer_intercept = 2.1, majorizer_slope = 0, t_min = 1, t_max = 13, atmost1 = TRUE)))
  compare_ppp_vectors(ppp1 = r_nhppp_intens, ppp2 = r_ppp, threshold = 0.1, showQQ = TRUE)

  r_nhppp_intens_piecewise <- unlist(lapply(integer(10000), function(x) draw_intensity_step(lambda = l, majorizer_vector = rep(2.1, 4), time_breaks = c(1, 2, pi, 2 * pi, 13), atmost1 = TRUE)))
  compare_ppp_vectors(ppp1 = r_nhppp_intens_piecewise, ppp2 = r_ppp, threshold = 0.1, showQQ = TRUE)
})

test_that("NHPPP linear intensity agrees with general methods", {
  set.seed(123)
  l <- function(t, intercept = 1, slope = 2) intercept + slope * t
  L <- function(t, intercept = 1, slope = 2, t0 = 1) Lambda_linear_form(t, intercept = intercept, slope = slope, t0 = t0)
  Li <- function(z, intercept = 1, slope = 2, t0 = 1) Lambda_inv_linear_form(z, intercept = intercept, slope = slope, t0 = t0)

  r_nhppp_intens_linear <- unlist(lapply(integer(10000), function(x) draw_sc_linear(intercept = 1, slope = 2, t_min = 1, t_max = 13, atmost1 = TRUE)))
  r_nhppp_ci_inv <- unlist(lapply(integer(10000), function(x) draw_cumulative_intensity_inversion(Lambda = L, Lambda_inv = Li, t_min = 1, t_max = 13, atmost1 = TRUE)))
  compare_ppp_vectors(ppp1 = r_nhppp_intens_linear, ppp2 = r_nhppp_ci_inv, threshold = 0.1, showQQ = TRUE)

  r_nhppp_ci_os <- unlist(lapply(integer(10000), function(x) draw_cumulative_intensity_orderstats(Lambda = L, Lambda_inv = Li, t_min = 1, t_max = 13, atmost1 = TRUE)))
  compare_ppp_vectors(ppp1 = r_nhppp_intens_linear, ppp2 = r_nhppp_ci_os, threshold = 0.1, showQQ = TRUE)

  r_nhppp_intens <- unlist(lapply(integer(10000), function(x) draw_intensity_line(lambda = l, majorizer_intercept = l(13), majorizer_slope = 0, t_min = 1, t_max = 13, atmost1 = TRUE)))
  compare_ppp_vectors(ppp1 = r_nhppp_intens_linear, ppp2 = r_nhppp_intens, threshold = 0.1, showQQ = TRUE)

  r_nhppp_intens_piecewise <- unlist(lapply(integer(10000), function(x) draw_intensity_step(lambda = l, majorizer_vector = l(2:13), time_breaks = c(1:13), atmost1 = TRUE)))
  compare_ppp_vectors(ppp1 = r_nhppp_intens_linear, ppp2 = r_nhppp_intens_piecewise, threshold = 0.1, showQQ = TRUE)
})

test_that("NHPPP loglinear agrees with general methods", {
  set.seed(123)
  l <- function(t, intercept = .1, slope = .02) exp(intercept + slope * t)
  L <- function(t, intercept = .1, slope = .02, t0 = 1) Lambda_exp_form(t, intercept = intercept, slope = slope, t0 = t0)
  Li <- function(z, intercept = .1, slope = .02, t0 = 1) Lambda_inv_exp_form(z, intercept = intercept, slope = slope, t0 = t0)

  r_nhppp_intens_exp <- unlist(lapply(integer(10000), function(x) draw_sc_loglinear(intercept = .1, slope = .02, t_min = 1, t_max = 13, atmost1 = TRUE)))
  r_nhppp_ci_inv <- unlist(lapply(integer(10000), function(x) draw_cumulative_intensity_inversion(Lambda = L, Lambda_inv = Li, t_min = 1, t_max = 13, atmost1 = TRUE)))
  compare_ppp_vectors(ppp1 = r_nhppp_intens_exp, ppp2 = r_nhppp_ci_inv, threshold = 0.1, showQQ = TRUE)

  r_nhppp_ci_os <- unlist(lapply(integer(10000), function(x) draw_cumulative_intensity_orderstats(Lambda = L, Lambda_inv = Li, t_min = 1, t_max = 13, atmost1 = TRUE)))
  compare_ppp_vectors(ppp1 = r_nhppp_intens_exp, ppp2 = r_nhppp_ci_os, threshold = 0.1, showQQ = TRUE)

  r_nhppp_intens <- unlist(lapply(integer(10000), function(x) draw_intensity_line(lambda = l, majorizer_intercept = l(13), majorizer_slope = 0, t_min = 1, t_max = 13, atmost1 = TRUE)))
  compare_ppp_vectors(ppp1 = r_nhppp_intens_exp, ppp2 = r_nhppp_intens, threshold = 0.1, showQQ = TRUE)

  r_nhppp_intens_piecewise <- unlist(lapply(integer(10000), function(x) draw_intensity_step(lambda = l, majorizer_vector = l(2:13), time_breaks = c(1:13), atmost1 = TRUE)))
  compare_ppp_vectors(ppp1 = r_nhppp_intens_exp, ppp2 = r_nhppp_intens_piecewise, threshold = 0.1, showQQ = TRUE)
})

#' @keywords internal
"_PACKAGE"

## usethis namespace: start
#' @importFrom lifecycle deprecated
#' @importFrom Rcpp sourceCpp
#' @useDynLib nhppp, .registration = TRUE
## usethis namespace: end
NULL

#' Helper function for the vectorized versions of sampling functions.
#' Takes the usual ways that `range_t` is specified
#' (a 2-vector, a 1 x 2 or an r x 2 matrix) and
#' returns a r x 2 matrix.
#'
#' @param range_t a 2-vector, a 1 x 2 or an r x 2 matrix
#' @param n_rows the number of rows in the fully expanded matrix (`r`)
#' @return A matrix (r x 2), row-expanded if needed
#' @keywords internal
make_range_t_matrix <- function(range_t, n_rows) {
  if (is.matrix(range_t) && nrow(range_t) == n_rows && ncol(range_t) == 2) {
    return(range_t)
  }
  if (is.matrix(range_t) && nrow(range_t) == 1 && ncol(range_t) == 2) {
    return(range_t[rep(1, n_rows), ])
  }
  if (is.vector(range_t) && length(range_t) == 2) {
    return(matrix(rep(range_t, each = n_rows), ncol = 2))
  }
  stop("`range_t` is not a 2-vector, a 1 x 2 or an `r x 2` matrix")
}


#' Helper function for the vectorized versions of sampling functions.
#' Takes the usual ways that `lambda_mat` and `Lambda_mat` are specified
#' and returns `lambda_mat`.
#'
#' @param lambda_mat a matrix of intensities or missing
#' @param Lambda_mat a matrix of cumulative intensities or missing
#' @param interval_duration a vector with the same number of elements as the rows of `Lambda_mat`
#' @return A matrix (r x 2), row-expanded if needed
#' @keywords internal
make_lambda_matrix <- function(lambda_mat = NULL, Lambda_mat = NULL, interval_duration = NULL) {
  if (!is.null(lambda_mat)) {
    mode(lambda_mat) <- "numeric"
    return(lambda_mat)
  }
  if (!is.null(Lambda_mat) && !is.null(interval_duration)) {
    mode(Lambda_mat) <- "numeric"
    lambda_mat <- mat_diff_columns(Lambda_mat) / interval_duration
    return(lambda_mat)
  }
  stop("missing argument (`interval_duration` is needed with `Lambda_mat`")
}


#' Helper function for the vectorized versions of sampling functions.
#' Takes the usual ways that `lambda_mat` and `Lambda_mat` are specified
#' and returns `Lambda_mat`.
#'
#' @param lambda_mat a matrix of intensities or missing
#' @param Lambda_mat a matrix of cumulative intensities or missing
#' @param interval_duration a vector with the same number of elements as the rows of `Lambda_mat`
#' @return A matrix (r x 2), row-expanded if needed
#' @keywords internal
make_cumulative_Lambda_matrix <- function(Lambda_mat = NULL, lambda_mat = NULL, interval_duration = NULL) {
  if (!is.null(Lambda_mat)) {
    mode(Lambda_mat) <- "numeric"
    return(Lambda_mat)
  }
  if (!is.null(lambda_mat) && !is.null(interval_duration)) {
    mode(lambda_mat) <- "numeric"
    Lambda_mat <- mat_cumsum_columns(lambda_mat) * interval_duration
    return(Lambda_mat)
  }
  stop("missing argument (`interval_duration` is needed with `lambda_mat`")
}


#' Return matrix with column-wise differencing.
#' No checks for arguments is done.
#'
#' @param X (matrix)
#' @return matrix
#' @keywords internal
mat_diff_columns <- function(X) {
  if (ncol(X) > 1) {
    X <- cbind(rep(0, nrow(X)), X)
    return(t(apply(X, 1, diff)))
  } else {
    return(X)
  }
}


#' Return matrix with column-wise cumulative sum
#' No checks for arguments is done.
#'
#' @param X (matrix)
#' @return matrix
#' @keywords internal
mat_cumsum_columns <- function(X) {
  if (ncol(X) > 1) {
    return(t(apply(X, 1, cumsum)))
  } else {
    return(X)
  }
}

#' Return matrix with column-wise cumulative sum
#' replacing cells larger than `ceil` with `NA`.
#' No checks for arguments is done.
#'
#' @param X (matrix)
#' @param ceil (double or Inf) the ceiling to be applied
#' @return matrix
#' @keywords internal
mat_cumsum_columns_with_scalar_ceiling <- function(X, ceil = Inf) {
  X <- mat_cumsum_columns(X)

  if (is.infinite(ceil)) {
    return(X)
  }

  X[which(X > ceil)] <- NA
  for (col in 1:ncol(X)) {
    if (all(is.na(X[, col]))) {
      if (col > 2) {
        return(X[, 1:(col - 1), drop = FALSE])
      } else {
        return(X[, 1, drop = FALSE])
      }
    }
  }
  return(X)
}


#' Return matrix with column-wise cumulative sum
#' replacing cells larger than `ceil` with `NA`.
#' No checks for arguments is done.
#'
#' @param X (matrix)
#' @param ceil (vector or Inf) the set of ceilings to be applied, per row of `X`
#' @return matrix
#' @keywords internal
mat_cumsum_columns_with_vector_ceiling <- function(X, ceil = Inf) {
  # browser()
  X <- mat_cumsum_columns(X)
  if (all(is.infinite(ceil))) {
    return(X)
  }
  for (col in 1:ncol(X)) {
    X[, col] <- pmin(X[, col], ceil)
    X[X[, col] == ceil, col] <- NA
    if (all(is.na(X[, col]))) {
      if (col > 2) {
        return(X[, 1:(col - 1), drop = FALSE])
      } else {
        return(X[, 1, drop = FALSE])
      }
    }
  }
  return(X)
}

## add tests that have lambda_maj < lambda
## add tests that have lamda_args

test_that("vdraw_intensity_step_regular_cpp() works", {
  set.seed(123)
  lfun <- function(x, lambda_args, ...) .2 * x^lambda_args$exponent
  l_args <- list(exponent = 1L)
  lmaj <- matrix(rep(1, 50), ncol = 5)
  Lmaj <- mat_cumsum_columns(lmaj)

  expect_no_error(Z <- vdraw_intensity_step_regular_cpp(
    lambda = lfun,
    lambda_args = l_args,
    lambda_maj_matrix = lmaj,
    rate_matrix_t_min = 1,
    rate_matrix_t_max = 5,
    tol = 10^-6,
    atmost1 = FALSE
  ))
  check_ppp_sample_validity(Z, t_min = 1, t_max = 5)

  expect_no_error(Z <- vdraw_intensity_step_regular_cpp(
    lambda = lfun,
    lambda_args = l_args,
    Lambda_maj_matrix = Lmaj,
    rate_matrix_t_min = 1,
    rate_matrix_t_max = 5,
    tol = 10^-6,
    atmost1 = FALSE
  ))
  check_ppp_sample_validity(Z, t_min = 1, t_max = 5)

  expect_no_error(Z <- vdraw_intensity_step_regular_cpp(
    lambda = lfun,
    lambda_args = l_args,
    lambda_maj_matrix = lmaj,
    rate_matrix_t_min = 1,
    rate_matrix_t_max = 5,
    tol = 10^-6,
    atmost1 = TRUE
  ))
  check_ppp_sample_validity(Z, t_min = 1, t_max = 5, atmost1 = TRUE)

  expect_no_error(Z <- vdraw_intensity_step_regular_cpp(
    lambda = lfun,
    lambda_args = l_args,
    Lambda_maj_matrix = Lmaj,
    rate_matrix_t_min = 1,
    rate_matrix_t_max = 5,
    tol = 10^-6,
    atmost1 = TRUE
  ))
  check_ppp_sample_validity(Z, t_min = 1, t_max = 5, atmost1 = TRUE)


  #### Also works for 1 row and range_t being matrix or vector

  # range_t vector
  expect_no_error(Z <- vdraw_intensity_step_regular_cpp(
    lambda = lfun,
    lambda_args = l_args,
    Lambda_maj_matrix = Lmaj[1, , drop = FALSE],
    rate_matrix_t_min = 1,
    rate_matrix_t_max = 5,
    tol = 10^-6,
    atmost1 = FALSE
  ))
  check_ppp_sample_validity(Z, t_min = 1, t_max = 5)

  # range_t matrix
  expect_no_error(Z <- vdraw_intensity_step_regular_cpp(
    lambda = lfun,
    lambda_args = l_args,
    Lambda_maj_matrix = Lmaj[1, , drop = FALSE],
    rate_matrix_t_min = 1,
    rate_matrix_t_max = 5,
    tol = 10^-6,
    atmost1 = FALSE
  ))
  check_ppp_sample_validity(Z, t_min = 1, t_max = 5)
})



test_that("vdraw_intensity_step_regular_cpp() does not break with matrices whose mode is list", {
  set.seed(123)
  lfun <- function(x, ...) .2 * x
  lmaj <- matrix(rep(1, 50), ncol = 5)
  Lmaj <- mat_cumsum_columns(lmaj)

  mode(Lmaj) <- "list"
  expect_no_error(Z <- vdraw_intensity_step_regular_cpp(
    lambda = lfun,
    Lambda_maj_matrix = Lmaj,
    rate_matrix_t_min = 1,
    rate_matrix_t_max = 5,
    tol = 10^-6,
    atmost1 = FALSE
  ))
  mode(lmaj) <- "list"
  expect_no_error(Z <- vdraw_intensity_step_regular_cpp(
    lambda = lfun,
    lambda_maj_matrix = lmaj,
    rate_matrix_t_min = 1,
    rate_matrix_t_max = 5,
    tol = 10^-6,
    atmost1 = FALSE
  ))
})



test_that("vdraw_intensity_step_regular_cpp() works with subinterval", {
  set.seed(123)
  lfun <- function(x, ...) .2 * x
  lmaj <- matrix(rep(1, 50), ncol = 5)
  Lmaj <- mat_cumsum_columns(lmaj)

  expect_no_error(Z <- vdraw_intensity_step_regular_cpp(
    lambda = lfun,
    Lambda_maj_matrix = Lmaj,
    rate_matrix_t_min = 1,
    rate_matrix_t_max = 5,
    t_min = 2.3,
    t_max = 4.8,
    tol = 10^-6,
    atmost1 = FALSE
  ))

  check_ppp_sample_validity(Z, t_min = 2.3, t_max = 4.8, atmost1 = FALSE)

  expect_no_error(Z <- vdraw_intensity_step_regular_cpp(
    lambda = lfun,
    lambda_maj_matrix = lmaj,
    rate_matrix_t_min = 1,
    rate_matrix_t_max = 5,
    t_min = 2.3,
    t_max = 4.8,
    tol = 10^-6,
    atmost1 = FALSE
  ))
  check_ppp_sample_validity(Z, t_min = 2.3, t_max = 4.8, atmost1 = FALSE)


  #### Also works for 1 row and range_t, subinterval being matrix or vector

  # range_t, subinterval vectors
  expect_no_error(Z <- vdraw_intensity_step_regular_cpp(
    lambda = lfun,
    lambda_maj_matrix = lmaj[1, , drop = FALSE],
    rate_matrix_t_min = 1,
    rate_matrix_t_max = 5,
    t_min = 2.3,
    t_max = 4.8,
    tol = 10^-6,
    atmost1 = FALSE
  ))
  check_ppp_sample_validity(Z, t_min = 2.3, t_max = 4.8, atmost1 = FALSE)

  # range_t,subinterval matrices
  expect_no_error(Z <- vdraw_intensity_step_regular_cpp(
    lambda = lfun,
    lambda_maj_matrix = lmaj[1, , drop = FALSE],
    rate_matrix_t_min = 1,
    rate_matrix_t_max = 5,
    t_min = 2.3,
    t_max = 4.8,
    tol = 10^-6,
    atmost1 = FALSE
  ))
  check_ppp_sample_validity(Z, t_min = 2.3, t_max = 4.8, atmost1 = FALSE)
})


test_that("vdraw_intensity_step_regular_cpp() works with different majorizers", {
  set.seed(123)
  lfun <- function(x, lambda_args, ...) .2 * x^lambda_args$exponent
  l_args <- list(exponent = 1L)
  lmaj <- matrix(rep(1, 1000), ncol = 5)

  expect_no_error(Z1 <- vdraw_intensity_step_regular_cpp(
    lambda = lfun,
    lambda_args = l_args,
    lambda_maj_matrix = lmaj,
    rate_matrix_t_min = 1,
    rate_matrix_t_max = 5,
    tol = 10^-6,
    atmost1 = FALSE
  ))
  check_ppp_sample_validity(Z1, t_min = 1, t_max = 5)

  expect_no_error(Z2 <- vdraw_intensity_step_regular_cpp(
    lambda = lfun,
    lambda_args = l_args,
    lambda_maj_matrix = lmaj + 10,
    rate_matrix_t_min = 1,
    rate_matrix_t_max = 5,
    tol = 10^-6,
    atmost1 = FALSE
  ))
  check_ppp_sample_validity(Z2, t_min = 1, t_max = 5)

  compare_ppp_vectors(ppp1 = Z1, ppp2 = Z2, threshold = 0.1, showQQ = TRUE)
})





test_that("vdraw_intensity_step_regular_cpp() uses blocked random numbers", {
  set.seed(123)
  lfun <- function(x, lambda_args, ...) .2 * x^lambda_args$exponent
  l_args <- list(exponent = 1L)
  l_ <- function(x) lfun(x, lambda_args = l_args)
  N <- 1000
  lmaj0 <- get_step_majorizer(fun = l_, breaks = matrix(rep(1:11, each = N), nrow = N), is_monotone = FALSE, K = 0)
  lmaj1 <- get_step_majorizer(fun = l_, breaks = matrix(rep(1:11, each = N), nrow = N), is_monotone = FALSE, K = 1)
  lmaj10 <- get_step_majorizer(fun = l_, breaks = matrix(rep(1:11, each = N), nrow = N), is_monotone = FALSE, K = 10)

  Z0 <- list()
  for (i in 1:2) {
    set.seed(123)
    expect_no_error(Z0[[i]] <- vdraw_intensity_step_regular_cpp(
      lambda = lfun,
      lambda_args = l_args,
      lambda_maj_matrix = lmaj0,
      rate_matrix_t_min = 1,
      rate_matrix_t_max = 5,
      tol = 10^-6,
      atmost1 = FALSE
    ))
    if (i > 1) {
      expect_true(identical(Z0[[1]], Z0[[i]]))
    }
  }
  check_ppp_sample_validity(Z0[[1]], t_min = 1, t_max = 5)
})

test_that("vdraw_sc_step_regular() works", {
  set.seed(123)
  # 1 row matrix
  expect_no_error(Z0 <- vdraw_sc_step_regular(
    Lambda_matrix = matrix(1:5, nrow = 1),
    rate_matrix_t_min = 100,
    rate_matrix_t_max = 110,
    atmost1 = FALSE
  ))
  check_ppp_sample_validity(Z0, t_min = 100, t_max = 110)

  l <- matrix(rep(1, 50), ncol = 5)
  L <- mat_cumsum_columns(l)

  expect_no_error(Z <- vdraw_sc_step_regular(
    Lambda_matrix = L,
    rate_matrix_t_min = 100,
    rate_matrix_t_max = 110,
    atmost1 = FALSE
  ))
  check_ppp_sample_validity(Z, t_min = 100, t_max = 110)

  expect_no_error(Z1 <- vdraw_sc_step_regular(
    Lambda_matrix = L,
    rate_matrix_t_min = 100,
    rate_matrix_t_max = 110,
    atmost1 = TRUE
  ))
  check_ppp_sample_validity(Z1, t_min = 100, t_max = 110, atmost1 = TRUE)

  expect_no_error(Z2 <- vdraw_sc_step_regular(
    lambda_matrix = l,
    rate_matrix_t_min = 100,
    rate_matrix_t_max = 110,
    atmost1 = TRUE
  ))
  check_ppp_sample_validity(Z2, t_min = 100, t_max = 110, atmost1 = TRUE)
})


test_that("vdraw_sc_step_regular() works with t_min and t_max", {
  set.seed(123)
  l <- matrix(rep(1, 50), ncol = 5)
  L <- mat_cumsum_columns(l)

  t0_a <- 101
  t1_a <- 109
  t0_b <- 100 + runif(10)
  t1_b <- 110 - runif(10)

  expect_no_error(Z <- vdraw_sc_step_regular(
    Lambda_matrix = L,
    rate_matrix_t_min = 100,
    rate_matrix_t_max = 110,
    t_min = t0_a
  ))
  check_ppp_sample_validity(Z, t_min = t0_a, t_max = 110)

  expect_no_error(Z1 <- vdraw_sc_step_regular(
    Lambda_matrix = L,
    rate_matrix_t_min = 100,
    rate_matrix_t_max = 110,
    t_max = t1_a
  ))
  check_ppp_sample_validity(Z1, t_min = 100, t_max = t1_a)

  expect_no_error(Z2 <- vdraw_sc_step_regular(
    Lambda_matrix = L,
    rate_matrix_t_min = 100,
    rate_matrix_t_max = 110,
    t_min = t0_a,
    t_max = t1_a
  ))
  check_ppp_sample_validity(Z2, t_min = t0_a, t_max = t1_a)

  expect_no_error(Z3 <- vdraw_sc_step_regular(
    Lambda_matrix = L,
    rate_matrix_t_min = 100,
    rate_matrix_t_max = 110,
    t_min = t0_b
  ))
  check_ppp_sample_validity(Z3, t_min = t0_b, t_max = 110)

  expect_no_error(Z4 <- vdraw_sc_step_regular(
    Lambda_matrix = L,
    rate_matrix_t_min = 100,
    rate_matrix_t_max = 110,
    t_max = t1_b
  ))
  check_ppp_sample_validity(Z4, t_min = 100, t_max = t1_b)


  expect_no_error(Z5 <- vdraw_sc_step_regular(
    Lambda_matrix = L,
    rate_matrix_t_min = 100,
    rate_matrix_t_max = 110,
    t_min = t0_b,
    t_max = t1_b
  ))
  check_ppp_sample_validity(Z5, t_min = t0_b, t_max = t1_b)
})

test_that("vdraw_sc_step_regular() does not break with matrices whose mode is list", {
  set.seed(123)
  l <- matrix(rep(1, 50), ncol = 5)
  L <- mat_cumsum_columns(l)

  mode(L) <- "list"
  expect_no_error(Z <- vdraw_sc_step_regular(
    Lambda_matrix = L,
    rate_matrix_t_min = 100,
    rate_matrix_t_max = 110,
    atmost1 = FALSE
  ))
})



test_that("vdraw_sc_step_regular() uses blocked random numbers", {
  set.seed(123)
  l <- matrix(rep(1, 500), ncol = 5)
  L <- mat_cumsum_columns(l)

  Z <- list()
  for (i in 1:2) {
    set.seed(123)
    expect_no_error(Z[[i]] <- vdraw_sc_step_regular(
      Lambda_matrix = L,
      rate_matrix_t_min = 100,
      rate_matrix_t_max = 110,
      atmost1 = FALSE
    ))
    if (i > 1) {
      expect_true(identical(Z[[1]], Z[[i]]))
    }
  }
})


test_that("vdraw_sc_step_regular() handles NA values in rate matrices", {
  set.seed(123)
  l <- matrix(rep(1, 50), ncol = 5)
  L <- mat_cumsum_columns(l)

  l[1, 1] <- NA
  L[1, 2] <- NA

  expect_error(vdraw_sc_step_regular(
    lambda_matrix = l,
    rate_matrix_t_min = 100,
    rate_matrix_t_max = 110,
    atmost1 = FALSE
  ))

  expect_error(vdraw_sc_step_regular(
    Lambda_matrix = L,
    rate_matrix_t_min = 100,
    rate_matrix_t_max = 110,
    atmost1 = FALSE
  ))
})

test_that("mat_cumsum_columns() works", {
  X <- matrix(rep(1, 10), ncol = 1)
  cumul_tX <- matrix(1:10, nrow = 1)

  expect_equal(X, mat_cumsum_columns(X))
  expect_equal(matrix(1:10, nrow = 1), mat_cumsum_columns(t(X)))

  X2 <- matrix(rep(1, 100), ncol = 10)
  cumul_X2 <- matrix(rep(cumul_tX, 10), byrow = TRUE, ncol = 10)
  expect_equal(cumul_X2, mat_cumsum_columns(X2))
})


test_that("mat_cumsum_columns_with_scalar_ceiling() works", {
  X <- matrix(rep(1, 10), ncol = 1)
  cumul_tX <- matrix(1:10, nrow = 1)

  # test one-row matrix
  expect_equal(
    matrix(1:10, nrow = 1),
    mat_cumsum_columns_with_scalar_ceiling(t(X), ceil = Inf)
  )
  expect_equal(
    matrix(1:8, nrow = 1),
    mat_cumsum_columns_with_scalar_ceiling(t(X), ceil = 8.2)
  )

  # test one-column matrix
  expect_equal(
    X,
    mat_cumsum_columns_with_scalar_ceiling(X, ceil = Inf)
  )
  expect_equivalent(
    matrix(rep(NA, 10), ncol = 1),
    mat_cumsum_columns_with_scalar_ceiling(X, ceil = 0.5)
  )

  X2 <- X3 <- matrix(rep(1, 100), ncol = 10)
  cumul_X2 <- cumul_X3 <- matrix(rep(cumul_tX, 10), byrow = TRUE, ncol = 10)
  expect_equal(cumul_X2, mat_cumsum_columns_with_scalar_ceiling(X2, ceil = Inf))

  X3[1, ] <- 9
  cumul_X3[1, ] <- NA
  expect_equal(cumul_X3[, 1:8], mat_cumsum_columns_with_scalar_ceiling(X3, ceil = 8.2))
})


test_that("mat_cumsum_columns_with_vector_ceiling() works", {
  X <- matrix(rep(1, 10), ncol = 1)
  cumul_tX <- matrix(1:10, nrow = 1)

  # test one-row matrix
  expect_equal(
    matrix(1:10, nrow = 1),
    mat_cumsum_columns_with_vector_ceiling(t(X), ceil = Inf)
  )
  expect_equal(
    matrix(1:8, nrow = 1),
    mat_cumsum_columns_with_vector_ceiling(t(X), ceil = 8.2)
  )

  # test one-column matrix
  expect_equal(
    X,
    mat_cumsum_columns_with_vector_ceiling(X, ceil = Inf)
  )
  expect_equal(
    matrix(c(NA, X[2:10, 1]), ncol = 1),
    mat_cumsum_columns_with_vector_ceiling(X, ceil = c(0.5, rep(8.2, 9)))
  )


  X2 <- X3 <- matrix(rep(1, 100), ncol = 10)
  cumul_X2 <- cumul_X3 <- matrix(rep(cumul_tX, 10), byrow = TRUE, ncol = 10)
  expect_equal(cumul_X2, mat_cumsum_columns_with_vector_ceiling(X2, ceil = Inf))

  X3[1, ] <- 9
  cumul_X3[1, ] <- NA
  expect_equal(cumul_X3[, 1:8], mat_cumsum_columns_with_vector_ceiling(X3, ceil = (802:811) / 100))
})

test_that("vztdraw_sc_step_regular_cpp() works", {
  set.seed(123)
  # 1 row matrix
  expect_no_error(Z0 <- vztdraw_sc_step_regular_cpp(
    Lambda_matrix = matrix(1:5, nrow = 1),
    rate_matrix_t_min = 100, rate_matrix_t_max = 110,
    atmost1 = FALSE
  ))
  check_ppp_sample_validity(Z0, t_min = 100, t_max = 110, atleast1 = TRUE)


  l <- lref <- matrix(rep(1, 50), ncol = 5)
  L <- Lref <- mat_cumsum_columns(l)

  expect_no_error(Z <- vztdraw_sc_step_regular_cpp(
    Lambda_matrix = L,
    rate_matrix_t_min = 100, rate_matrix_t_max = 110,
    atmost1 = FALSE
  ))
  check_ppp_sample_validity(Z, t_min = 100, t_max = 110, atleast1 = TRUE)


  expect_no_error(Z1 <- vztdraw_sc_step_regular_cpp(
    Lambda_matrix = L,
    rate_matrix_t_min = 100, rate_matrix_t_max = 110,
    atmost1 = TRUE
  ))
  check_ppp_sample_validity(Z1, t_min = 100, t_max = 110, atmost1 = TRUE, atleast1 = TRUE)

  expect_no_error(Z2 <- vztdraw_sc_step_regular_cpp(
    lambda_matrix = l,
    rate_matrix_t_min = 100, rate_matrix_t_max = 110,
    atmost1 = TRUE
  ))
  check_ppp_sample_validity(Z2, t_min = 100, t_max = 110, atmost1 = TRUE, atleast1 = TRUE)

  # very small rates
  expect_no_error(Z2 <- vztdraw_sc_step_regular_cpp(
    lambda_matrix = l * 0.001,
    rate_matrix_t_min = 100, rate_matrix_t_max = 110,
    atmost1 = TRUE
  ))
  check_ppp_sample_validity(Z2, t_min = 100, t_max = 110, atmost1 = TRUE, atleast1 = TRUE)


  expect_equal(l, lref)
  expect_equal(L, Lref)
})

test_that("vztdraw_sc_step_regular_cpp() does not break with matrices whose mode is list", {
  set.seed(123)
  l <- matrix(rep(1, 50), ncol = 5)
  L <- mat_cumsum_columns(l)

  mode(L) <- "list"
  expect_no_error(Z <- vztdraw_sc_step_regular_cpp(
    Lambda_matrix = L,
    rate_matrix_t_min = 100, rate_matrix_t_max = 110,
    atmost1 = FALSE
  ))
})



test_that("vztdraw_sc_step_regular_cpp() uses blocked random numbers", {
  set.seed(123)
  l <- matrix(rep(1, 500), ncol = 5)
  L <- mat_cumsum_columns(l)

  Z <- list()
  for (i in 1:2) {
    set.seed(123)
    expect_no_error(Z[[i]] <- vztdraw_sc_step_regular_cpp(
      Lambda_matrix = L,
      rate_matrix_t_min = 100, rate_matrix_t_max = 110,
      atmost1 = FALSE
    ))
    if (i > 1) {
      expect_true(identical(Z[[1]], Z[[i]]))
    }
  }
})


test_that("vztdraw_sc_step_regular_cpp() works with subinterval", {
  set.seed(123)
  expect_no_error(Z0 <- vztdraw_sc_step_regular_cpp(
    Lambda_matrix = matrix(1:5, nrow = 1),
    rate_matrix_t_min = 100, rate_matrix_t_max = 110,
    t_min = 100, t_max = 110, # matrix
    atmost1 = FALSE
  ))

  expect_no_error(Z0 <- vztdraw_sc_step_regular_cpp(
    Lambda_matrix = matrix(1:5, nrow = 1),
    rate_matrix_t_min = 100, rate_matrix_t_max = 110,
    t_min = 100, t_max = 110, # vector
    atmost1 = FALSE
  ))
  check_ppp_sample_validity(Z0, t_min = 100, t_max = 110, atmost1 = FALSE, atleast1 = TRUE)
  expect_no_error(Z0 <- vztdraw_sc_step_regular_cpp(
    Lambda_matrix = matrix(1:5, nrow = 1),
    rate_matrix_t_min = 100, rate_matrix_t_max = 110,
    t_min = 101.01, t_max = 108.99,
    atmost1 = FALSE
  ))
  check_ppp_sample_validity(Z0, t_min = 101.01, t_max = 108.99, atmost1 = FALSE, atleast1 = TRUE)

  expect_no_error(Z0 <- vztdraw_sc_step_regular_cpp(
    Lambda_matrix = matrix(1:5, nrow = 1) * 10,
    rate_matrix_t_min = 100, rate_matrix_t_max = 110,
    t_min = 105.01, t_max = 105.99,
    atmost1 = FALSE
  ))
  check_ppp_sample_validity(Z0, t_min = 105.01, t_max = 105.99, atmost1 = FALSE, atleast1 = TRUE)

  expect_no_error(Z0 <- vztdraw_sc_step_regular_cpp(
    Lambda_matrix = matrix(1:5, nrow = 1),
    rate_matrix_t_min = 100, rate_matrix_t_max = 110,
    t_min = 101.01, t_max = 108.99,
    atmost1 = TRUE
  ))
  check_ppp_sample_validity(Z0, t_min = 101.01, t_max = 108.99, atmost1 = TRUE, atleast1 = TRUE)

  expect_no_error(Z0 <- vztdraw_sc_step_regular_cpp(
    Lambda_matrix = matrix(1:500, nrow = 10),
    rate_matrix_t_min = 0, rate_matrix_t_max = 10,
    t_min = 0, t_max = 5,
    atmost1 = TRUE
  ))
  check_ppp_sample_validity(Z0, t_min = 0, t_max = 5, atmost1 = TRUE, atleast1 = TRUE)
})

#' Simulate from a zero-truncated non homogeneous Poisson Point Process (NHPPP) from
#'    (t0, t_max) (thinning method) with piecewise constant_majorizer
#'
#' @description Sample zero-truncated NHPPP times using the thinning method
#' @param lambda (function) the instantaneous rate of the NHPPP.
#' A continuous function of time.
#' @param majorizer_vector (scalar, double) `K` constant majorizing rates, one per interval
#' @param time_breaks (vector, double) `K+1` time points defining `K` intervals
#'        of constant rates:
#'             `[t_1 = range_t[1],       t_2)`: the first interval
#'             `[t_k,                t_{k+1})`: the `k`-th interval
#'             `[t_{K}, t_{K+1} = range_t[2])`: the `K`-th (last) interval
#' @param atmost1 boolean, draw at most 1 event time
#'
#' @return a vector of event times (t_) with at least one element
#' @keywords internal
ztdraw_intensity_step <- function(lambda,
                                  majorizer_vector,
                                  time_breaks,
                                  atmost1 = FALSE) {
  len_lambda <- length(majorizer_vector)

  lambda_maj_fun <- stats::approxfun(
    x = time_breaks[1:len_lambda],
    y = majorizer_vector, method = "constant", rule = 2, f = 0
  )

  while (TRUE) {
    candidate_times <- draw_sc_step(lambda_vector = majorizer_vector, time_breaks = time_breaks, atmost1 = FALSE, atleast1 = TRUE)
    u <- rng_stream_runif(size = length(candidate_times), minimum = 0, maximum = 1)
    acceptance_prob <- lambda(candidate_times) / lambda_maj_fun(candidate_times)
    if (!all(acceptance_prob <= 1 + 10^-6)) stop("lambda > lambda_maj\n")
    candidate_times <- candidate_times[u < acceptance_prob]
    if (length(candidate_times) > 0) {
      if (atmost1) {
        return(candidate_times[1])
      } else {
        return(candidate_times)
      }
    }
  }
}

#' Simulate `size` samples from a zero-truncated non homogeneous Poisson Point Process (zt-NHPPP) from
#'    (t0, t_max) (thinning method)
#'
#' @description Sample zero-truncated NHPPP intensity times using the thinning method
#'
#' @param lambda (function) the instantaneous rate of the NHPPP.
#' @param majorizer_intercept (double) the intercept (`alpha`) of the [log]linear majorizer function.
#' @param majorizer_slope (double) the slope (`beta') of the [log]linear majorizer function.
#' @param t_min (double) the lower bound of the time interval.
#' @param t_max (double) the upper bound of the time interval.
#' @param majorizer_is_loglinear (boolean) if `TRUE` the majorizer is loglinear `exp(alpha + beta * t)`
#' @param atmost1 boolean, draw at most 1 event time
#'
#' @return a vector of at least 1 event times
#' @keywords internal
ztdraw_intensity_line <- function(lambda,
                                  majorizer_intercept,
                                  majorizer_slope,
                                  t_min,
                                  t_max,
                                  majorizer_is_loglinear = FALSE,
                                  atmost1 = FALSE) {
  if (isTRUE(majorizer_is_loglinear)) {
    ztnhppp_t <- ztdraw_sc_loglinear
    link <- exp
  } else {
    ztnhppp_t <- ztdraw_sc_linear
    link <- identity
  }

  while (TRUE) {
    candidate_times <- ztnhppp_t(intercept = majorizer_intercept, slope = majorizer_slope, t_min = t_min, t_max = t_max, atmost1 = FALSE)
    u <- stats::runif(n = length(candidate_times), min = 0, max = 1)
    acceptance_prob <- lambda(candidate_times) / link(majorizer_intercept + majorizer_slope * candidate_times)
    if (!all(acceptance_prob <= 1 + 10^-6)) stop("lambda > lambda_maj\n")
    candidate_times <- candidate_times[u < acceptance_prob]
    if (length(candidate_times) > 0) {
      if (atmost1) {
        return(candidate_times[1])
      } else {
        return(candidate_times)
      }
    }
  }
}

test_that("ztdraw_intensity_line() works", {
  set.seed(123)
  l <- function(t) {
    return(2)
  }
  lmaj_intercept <- 2.2
  lmaj_slope <- 0

  expect_no_error(withr::with_preserve_seed(df <- ztdraw_intensity_line(lambda = l, majorizer_intercept = lmaj_intercept, majorizer_slope = lmaj_slope, t_min = 0, t_max = 10)))
  check_ppp_sample_validity(times = df, t_min = 0, t_max = 10, atleast1 = TRUE)

  # works when range_t[1]>0
  expect_no_error(withr::with_preserve_seed(df <- ztdraw_intensity_line(lambda = l, majorizer_intercept = lmaj_intercept, majorizer_slope = lmaj_slope, t_min = 5, t_max = 10)))
  check_ppp_sample_validity(times = df, t_min = 5, t_max = 10, atleast1 = TRUE)

  # works with atmost1
  expect_no_error(withr::with_preserve_seed(df2 <- ztdraw_intensity_line(lambda = l, majorizer_intercept = lmaj_intercept, majorizer_slope = lmaj_slope, t_min = 5, t_max = 10, atmost1 = TRUE)))
  check_ppp_sample_validity(times = df2, t_min = 5, t_max = 10, atleast1 = TRUE, atmost1 = TRUE)
})

test_that("ztdraw_intensity_line() matches draw_cumulative_intensity_inversion() (and linear intensity)", {
  l <- function(t) {
    return(2)
  }
  L <- function(t) {
    return(2 * t)
  }
  L_inv <- function(z) {
    return(z / 2)
  }

  r1 <- lapply(integer(10000), function(x) ztdraw_intensity_line(lambda = l, majorizer_intercept = 2.01, majorizer_slope = 0, t_min = 0, t_max = 2, atmost1 = TRUE))
  r2 <- lapply(integer(10000), function(x) draw_cumulative_intensity_inversion(Lambda = L, Lambda_inv = L_inv, t_min = 0, t_max = 2, atmost1 = TRUE))
  compare_ppp_vectors(ppp1 = unlist(r1), ppp2 = unlist(r2), threshold = 0.20, showQQ = TRUE)
})



test_that("ztdraw_intensity_line() works with linear majorization function", {
  l <- function(t) {
    return(2)
  }


  # works with negative slope majorization function
  expect_no_error(withr::with_preserve_seed(
    df <- ztdraw_intensity_line(
      lambda = l,
      majorizer_intercept = 50, majorizer_slope = -1,
      t_min = 5, t_max = 10
    )
  ))
  check_ppp_sample_validity(times = df, t_min = 5, t_max = 10, atleast1 = TRUE)

  # works with positive slope majorization function
  expect_no_error(withr::with_preserve_seed(
    df <- ztdraw_intensity_line(
      lambda = l,
      majorizer_intercept = 2, majorizer_slope = .1,
      t_min = 5, t_max = 10
    )
  ))
  check_ppp_sample_validity(times = df, t_min = 5, t_max = 10, atleast1 = TRUE)

  # fails when majorization function is below l
  # this is only when you sample at that point
  expect_error(withr::with_preserve_seed(
    df <- ztdraw_intensity_line(
      lambda = l,
      majorizer_intercept = 2, majorizer_slope = -.1,
      t_min = 5, t_max = 10
    )
  ))
  check_ppp_sample_validity(times = df, t_min = 5, t_max = 10, atleast1 = TRUE)
})

#' Simulate n events from a homogeneous Poisson Point Process.
#'
#' @param rate scalar instantaneous rate
#' @param n scalar number of samples
#' @param t_min scalar for the starting time value
#' @param rng_stream  `r lifecycle::badge("deprecated")` an `rstream` object
#'
#' @return a vector with event times t (starting from t_min)
#' @export
#'
#' @examples
#' x <- ppp_next_n(n = 10, rate = 1, t_min = 0)
#' @importClassesFrom rstream rstream.mrg32k3a
ppp_next_n <- function(n = 1, rate = 1, t_min = 0, rng_stream = deprecated()) {
  if (lifecycle::is_present(rng_stream)) {
    lifecycle::deprecate_warn(when = "0.5.0", what = "ppp_next_n(rng_stream)")
    dt_ <- rng_stream_rexp(size = n, rate = rate, rng_stream = rng_stream)
  } else {
    dt_ <- stats::rexp(n = n, rate = rate)
  }

  t_ <- cumsum(dt_) + t_min
  return(t_)
}

test_that("ppp_next_n() works", {
  expect_no_error(
    withr::with_seed(
      12345,
      df1 <- ppp_next_n(n = 10, rate = 1, t_min = 5)
    )
  )
  check_ppp_sample_validity(times = df1, t_min = 5, size = 10)
})

#' Simulate from a zero-truncated non homogeneous Poisson Point Process (zt-NHPPP) from
#'    (t_min, t_max) (order statistics method)
#'
#' @description  Sample zero-truncated NHPPP times using the order statistics method,
#' optionally using an `rstream` generator
#' @param Lambda (function, double vector) a continuous increasing R to R map
#'               which is the integrated rate of the NHPPP
#' @param Lambda_inv (function, double vector) the inverse of `Lambda()`
#' @param t_min (double) the lower bound of the time interval
#' @param t_max (double) the upper bound of the time interval
#' @param atmost1 (boolean) draw at most 1 event time
#'
#' @return a vector of at least 1 event times
#' @export
ztdraw_cumulative_intensity <- function(Lambda,
                                        Lambda_inv,
                                        t_min,
                                        t_max,
                                        atmost1 = FALSE) {
  tmp_u <- ztppp(rate = 1, t_min = Lambda(t_min), t_max = Lambda(t_max))
  if (atmost1 == TRUE) {
    tmp_u <- tmp_u[1]
  }
  return(Lambda_inv(tmp_u))
}

test_that("ztdraw_cumulative_intensity() works", {
  set.seed(123)
  L <- function(t) {
    return(2 * t)
  }
  Li <- function(z) {
    return(z / 2)
  }

  expect_no_error(withr::with_preserve_seed(df1 <- ztdraw_cumulative_intensity(Lambda = L, Lambda_inv = Li, t_min = 0, t_max = 10)))
  check_ppp_sample_validity(times = df1, t_min = 0, t_max = 10, atleast1 = TRUE)

  # works when range_t[1]>0
  expect_no_error(withr::with_preserve_seed(df2 <- ztdraw_cumulative_intensity(Lambda = L, Lambda_inv = Li, t_min = 5, t_max = 10)))
  check_ppp_sample_validity(times = df2, t_min = 5, t_max = 10, atleast1 = TRUE)

  # only 1 works
  expect_no_error(withr::with_preserve_seed(df4 <- ztdraw_cumulative_intensity(Lambda = L, Lambda_inv = Li, t_min = 5, t_max = 10, atmost1 = TRUE)))
  check_ppp_sample_validity(times = df4, t_min = 5, t_max = 10, atleast1 = TRUE, atmost1 = TRUE)

  # works with very small range
  expect_no_error(withr::with_preserve_seed(df4 <- ztdraw_cumulative_intensity(Lambda = L, Lambda_inv = Li, t_min = 5, t_max = 5 + 10^-8, atmost1 = TRUE)))
  check_ppp_sample_validity(times = df4, t_min = 5, t_max = 5 + 10^-8, atleast1 = TRUE, atmost1 = TRUE)
})

#' Simulate a homogeneous Poisson Point Process in (t_min, t_max]
#'
#' @param rate (scalar, double) constant instantaneous rate
#' @param t_min (scalar, double) the lower bound of the time interval
#' @param t_max (scalar, double) the upper bound of the time interval
#' @param atmost1 boolean, draw at most 1 event time
#' @param tol the probability that we will have more than
#'        the drawn events in (t_min, t_max]
#'
#' @return a vector of event times t
#'         if no events realize, it will have 0 length
#' @export
#'
#' @examples
#' x <- ppp(rate = 1, t_min = 0, t_max = 10, tol = 10^-6)
ppp <- function(rate, t_min, t_max, atmost1 = FALSE, tol = 10^-6) {
  if (isTRUE(atmost1)) {
    n <- 1
  } else {
    n <- stats::qpois(p = 1 - tol, lambda = rate * (t_max - t_min))
  }
  dt_ <- stats::rexp(n = n, rate = rate)
  t_ <- cumsum(dt_) + t_min
  return(t_[t_ <= t_max])
}

test_that("ppp() works", {
  expect_no_error(
    withr::with_seed(
      12345,
      df1 <- ppp(rate = 1, t_min = 0, t_max = 10)
    )
  )
  check_ppp_sample_validity(times = df1, t_min = 0, t_max = 10)

  # check atmost1 = TRUE
  withr::with_seed(12345, df2 <- ppp(rate = 1, t_min = 0, t_max = 10, atmost1 = TRUE))
  expect_identical(df2, df1[1])
})

#' Vectorized sampling from a non homogeneous Poisson Point Process (NHPPP) given
#'    the intensity function with piecewise constant_majorizers (C++)
#'
#' @description
#' Vectorized sampling from a non homogeneous Poisson Point Process (NHPPP) using the intensity
#' function and piecewise constant_majorizers.
#'    The majorizers are step functions over equal-length time intevals.
#'
#' @param lambda (function) intensity function, vectorized
#' @param lambda_args (list) optional named list of arguments to pass to `lambda`.
#'        If you have arguments for `lambda` that vary by draw, they should be passed as
#'        a data.table named `vector_arguments`.
#' @param Lambda_maj_matrix (matrix) integrated intensity rates at the end of each interval
#' @param lambda_maj_matrix (matrix) intensity rates, one per interval
#' @param rate_matrix_t_min (scalar | vector | column matrix) is the lower bound
#'        of the time interval for each row of (Lambda|lambda)_maj_matrix.
#'        The length of this argument is the number of point processes that should be drawn.
#' @param rate_matrix_t_max (scalar | vector | column matrix) the upper bound
#'        of the time interval for each row of (Lambda|lambda)_maj_matrix.
#'        The length of this argument is the number of point processes that should be drawn.
#' @param t_min (scalar | vector | column matrix) is the lower bound
#'        of a subinterval of (rate_matrix_t_min, rate_matrix_t_max]. If set,
#'        times are sampled from the subinterval.
#'        If omitted, it is equivalent to `rate_matrix_t_min`.
#' @param t_max (scalar | vector | column matrix) is the upper bound
#'        of a subinterval of (rate_matrix_t_min, rate_matrix_t_max]. If set,
#'        times are sampled from the subinterval.
#'        If omitted, it is equivalent to `rate_matrix_t_max`.
#' @param tol (scalar, double) tolerance for the number of events
#' @param atmost1 boolean, draw at most 1 event time
#' @param atmostB If not NULL, draw at most B (B>0) event times. NULL means ignore.
#'
#' @return a matrix of event times (columns) per draw (rows)
#'         NAs are structural empty spots
#' @keywords internal
vdraw_intensity_step_regular_cpp <- function(lambda = NULL,
                                             lambda_args = NULL,
                                             Lambda_maj_matrix = NULL,
                                             lambda_maj_matrix = NULL,
                                             rate_matrix_t_min = NULL,
                                             rate_matrix_t_max = NULL,
                                             t_min = NULL,
                                             t_max = NULL,
                                             tol = 10^-6,
                                             atmost1 = FALSE,
                                             atmostB = NULL) {
  if (is.null(atmostB)) {
    atmostB <- 0 # has to be <=0 to be ignored
  }

  if (!is.null(lambda_maj_matrix) && is.null(Lambda_maj_matrix)) {
    rate <- lambda_maj_matrix
    is_cumulative_rate <- FALSE
  } else if (is.null(lambda_maj_matrix) && !is.null(Lambda_maj_matrix)) {
    rate <- Lambda_maj_matrix
    is_cumulative_rate <- TRUE
  } else {
    stop("lambda_maj_matrix and Lambda_maj_matrix cannot both be `NULL`")
  }
  mode(rate) <- "numeric"

  num_na <- sum(is.na(rate))
  if (num_na > 0) {
    rate_argument <- if (is_cumulative_rate) "Lambda_maj_matrix" else "lambda_maj_matrix"
    stop("The ", rate_argument, " contains ", num_na, " NA values")
  }

  range_t <- cbind(as.vector(rate_matrix_t_min), as.vector(rate_matrix_t_max))
  if (nrow(range_t) > 1 && nrow(range_t) != nrow(rate)) {
    stop("The (rows of) [Lambda|lambda]_maj_matrix and (length of) [rate_matrix_t_min|rate_matrix_t_max] imply different numbers of point processes to be sampled.")
  }
  if (nrow(range_t) == 1 && nrow(rate) != 1) {
    range_t <- range_t[rep(1, nrow(rate)), ]
  }


  if (is.null(lambda_args)) {
    l_ <- lambda
  } else {
    l_ <- function(X, ...) {
      return(lambda(X, lambda_args))
    }
  }


  if (is.null(t_min) && is.null(t_max)) {
    use_subinterval <- FALSE
    subinterval <- range_t
  } else {
    use_subinterval <- TRUE
    if (is.null(t_min)) t_min <- range_t[, 1, drop = FALSE]
    if (is.null(t_max)) t_max <- range_t[, 2, drop = FALSE]
    subinterval <- cbind(as.vector(t_min), as.vector(t_max))
    if (nrow(subinterval) > 1 && nrow(subinterval) != nrow(rate)) {
      stop("The (rows of) [Lambda|lambda]_maj_matrix and (length of) [t_min|t_max] imply different numbers of point processes to be sampled.")
    }
    if (nrow(subinterval) == 1 && nrow(rate) != 1) {
      subinterval <- subinterval[rep(1, nrow(rate)), ]
    }
    stopifnot(all(subinterval[, 1] >= range_t[, 1]), all(subinterval[, 2] <= range_t[, 2]))
  }


  return(
    .Call(
      `_nhppp_vdraw_intensity_step_regular`, l_,
      rate, is_cumulative_rate, range_t, subinterval, use_subinterval, tol, atmost1, atmostB
    )
  )
}

#' Simulate from a non homogeneous Poisson Point Process (NHPPP)
#'  over an interval when you know the cumulative intensity and its inverse.
#'
#' @description  Sample NHPPP times using the inversion method
#' @param Lambda (function, double vector) a continuous increasing R to R map
#'               which is the integrated rate of the NHPPP
#' @param Lambda_inv (function, double vector) the inverse of `Lambda()`
#' @param t_min (double) the lower bound of the time interval
#' @param t_max (double) the upper bound of the time interval
#' @param atmost1 boolean, draw at most 1 event time
#'
#' @return a vector of event times (t_); if no events realize,
#'         a vector of length 0
#' @export
draw_cumulative_intensity <- function(Lambda,
                                      Lambda_inv,
                                      t_min,
                                      t_max,
                                      atmost1 = FALSE) {
  draw_cumulative_intensity_inversion(
    Lambda = Lambda,
    Lambda_inv = Lambda_inv,
    t_min = t_min,
    t_max = t_max,
    atmost1 = atmost1
  )
}

#' Piecewise constant (step) majorizer for K-Lipschitz functions over an interval
#' (vectorized over the `breaks` argument).
#'
#' @description Return a piecewise constant (step) majorizer for K-Lipschitz functions
#'              over an interval. The function is vectorized over the `breaks` argument.
#'              The returned object has the same dimensions as `breaks`.
#' @param fun A function object with a single argument `x`.
#'            If `x` is a matrix, `fun` should be vectorized over it.
#' @param breaks (vector or matrix) The set of `M+1` boundaries for the `M` subintervals in `x`.
#'               If breaks is a matrix, each row is treated as a separate set of breaks.
#' @param is_monotone (boolean) Is the function monotone? (Default is `TRUE`.)
#' @param K (double) A non-negative number for the Lipschitz cone. (Default is 0.)
#' @return A vector of length `M` with the values of the piecewise constant majorizer
#'
#' @export
#' @examples
#' get_step_majorizer(fun = abs, breaks = -5:5, is_monotone = FALSE, K = 1)
get_step_majorizer <- function(fun, breaks, is_monotone = TRUE, K = 0) {
  # browser()
  if (K < 0) stop()
  drop <- FALSE
  if (is.vector(breaks)) {
    breaks <- matrix(breaks, nrow = 1)
    drop <- TRUE
  }
  M <- ncol(breaks) - 1
  f_breaks <- fun(breaks)

  lambda_star <- pmax(f_breaks[, 1:M, drop = drop], f_breaks[, 2:(M + 1), drop = drop])
  if (isTRUE(is_monotone)) {
    return(lambda_star)
  } else {
    return(lambda_star + K * abs(breaks[, 1:M, drop = drop] - breaks[, 2:(M + 1), drop = drop]) / 2)
  }
}

test_that("get_step_majorizer() works", {
  # with breaks being a vector
  expect_no_error(l1 <- get_step_majorizer(
    fun = abs,
    breaks = -10:10, is_monotone = FALSE, K = 1
  ))
  expect_true(is.vector(l1))
  l1_fun <- stats::approxfun(y = l1, x = -10:9, method = "constant", rule = 2)
  x <- seq(-10, 10, length.out = 1000)
  expect_true(sum(l1_fun(x) < abs(x)) == 0)


  # with breaks being a row_matrix
  expect_no_error(l1 <- get_step_majorizer(
    fun = abs,
    breaks = matrix(-10:10, nrow = 1), is_monotone = FALSE, K = 1
  ))
  expect_true(is.matrix(l1))
  l1_fun <- stats::approxfun(y = l1, x = -10:9, method = "constant", rule = 2)
  x <- seq(-10, 10, length.out = 1000)
  expect_true(sum(l1_fun(x) < abs(x)) == 0)

  # with breaks being multi-row matrix
  breaks_mat <- rbind(-10:10, (-10:10) / 2, (-10:10) * 3, (-10:10) - 20)

  expect_no_error(l1 <- get_step_majorizer(
    fun = abs,
    breaks = breaks_mat, is_monotone = FALSE, K = 1
  ))
  expect_true(is.matrix(l1))
  for (r in 1:4) {
    l1_fun <- stats::approxfun(y = l1[r, ], x = breaks_mat[r, 1:20], method = "constant", rule = 2)
    x <- seq(breaks_mat[r, 1], breaks_mat[r, 21], length.out = 1000)
    expect_true(sum(l1_fun(x) < abs(x)) == 0)
  }

  # with breaks being multi-row matrix with all rows the same
  expect_no_error(l1 <- get_step_majorizer(
    fun = abs,
    breaks = breaks_mat[rep(1, 4), ], is_monotone = FALSE, K = 1
  ))
  expect_identical(l1[1, ], l1[2, ])
  expect_identical(l1[1, ], l1[3, ])
  expect_identical(l1[1, ], l1[4, ])

  l1_fun <- stats::approxfun(y = l1[1, ], x = breaks_mat[1, 1:20], method = "constant", rule = 2)
  x <- seq(breaks_mat[1, 1], breaks_mat[1, 21], length.out = 1000)
  expect_true(sum(l1_fun(x) < abs(x)) == 0)
})

#' Simulate from a non homogeneous Poisson Point Process (NHPPP) from
#'    (t_min, t_max) (inversion method)
#'
#' @description  Sample NHPPP times using the inversion method,
#' optionally using an `rstream` generator object
#' @param Lambda (function, double vector) a continuous increasing R to R map
#'               which is the integrated rate of the NHPPP
#' @param Lambda_inv (function, double vector) the inverse of `Lambda()`
#' @param t_min (double) the lower bound of the time interval
#' @param t_max (double) the upper bound of the time interval
#' @param atmost1 boolean, draw at most 1 event time
#'
#' @return a vector of event times (t_); if no events realize,
#'         a vector of length 0
#' @keywords internal
draw_cumulative_intensity_inversion <- function(Lambda,
                                                Lambda_inv,
                                                t_min,
                                                t_max,
                                                atmost1 = FALSE) {
  L_min <- Lambda(t_min)
  L_max <- Lambda(t_max)
  dat_warped_time <- ppp(
    rate = 1,
    t_min = L_min,
    t_max = L_max,
    atmost1 = atmost1
  )

  if (length(dat_warped_time) == 0) {
    return(dat_warped_time)
  }
  return(Lambda_inv(dat_warped_time))
}

test_that("draw_cumulative_intensity_inversion() works with the Lambda_inv option", {
  L <- function(t) {
    return(2 * t)
  }
  Li <- function(z) {
    return(z / 2)
  }
  withr::with_seed(12345, df1 <- draw_cumulative_intensity_inversion(Lambda = L, Lambda_inv = Li, t_min = 0, t_max = 1))
  check_ppp_sample_validity(times = df1, t_min = 0, t_max = 1)
})

#' Simulate a homogeneous Poisson Point Process over (t_min, t_max] (order statistics method)
#'
#' Internal function -- not to be exported.
#' Same as `ppp` but uses the Order Statistics algorithm.
#' @param rate (scalar, double) constant instantaneous rate
#' @param t_min (scalar, double) the lower bound of the time interval
#' @param t_max (scalar, double) the upper bound of the time interval
#' @param atmost1 boolean, draw at most 1 event time
#'
#' @return a vector of event times t
#'         if no events realize, it will have 0 length
#'
#' @keywords internal
#' @examples
#' x <- ppp(rate = 1, t_min = 0, t_max = 10, tol = 10^-6)
ppp2 <- function(rate, t_min, t_max, atmost1 = FALSE) {
  n <- stats::rpois(n = 1, lambda = rate * (t_max - t_min))
  tmp <- ppp_exactly_n(n = n, t_min = t_min, t_max = t_max)
  if (atmost1 == TRUE) {
    return(tmp[1])
  }
  return(tmp)
}

#' Helper functions
#'
#' Small utility functions. Not to be exported to the user.

#' @description helper function that augments
#' `test_that::expect_no_error()` to expect no error.
#' Copied from the `R6` source code.
#' @param expr Expression.
#' @return NULL
#' @keywords internal
expect_no_error <- function(expr) {
  err <- FALSE
  tryCatch(force(expr),
    error = function(e) {
      err <<- TRUE
    }
  )
  testthat::expect(!err, "Expected no error, but had error.")
  invisible(NULL)
}

#' Read code from text file as string
#'
#' @param codeFile Path to file
#' @return \code{codeFile} contents as a character string
#' @keywords internal
read_code <- function(codeFile) {
  paste(readLines(file.path(codeFile)), collapse = "\n")
}





#' Check the validity of a ppp vector.
#'
#' @description Standard checks for a vector of ordered times. Check
#' that the `times` vector is sorted, has unique values, has all values
#' in `[t_min, t_max]`, and has length `size` (if applicable).
#'
#' @param times (vector, double) the times to be checked
#' @param t_min (double) the start of the time nterval
#' @param t_max (double) optional: the end of the time interval
#' @param size (double) optional: the size of the vector
#' @param atmost1 (boolean) optional: at most one sample returned
#' @param atleast1 (boolean) optional: at least one sample returned
#' @return None
#' @keywords internal
check_ppp_vector_validity <- function(times, t_min, t_max = NULL, size = NULL, atmost1 = FALSE, atleast1 = FALSE) {
  times <- times[!is.na(times)]
  if (atleast1) {
    testthat::expect_true(length(times) >= 1)
  }
  if (length(times != 0)) {
    testthat::expect_identical(times, sort(times))
    testthat::expect_identical(times, unique(times))
    testthat::expect_true(min(times, Inf) >= t_min)
    if (!is.null(t_max)) {
      testthat::expect_true(max(times, -Inf) <= t_max)
    }
    if (!is.null(size)) {
      testthat::expect_equal(length(times), size)
    }
    if (atmost1) {
      testthat::expect_true(length(times) <= 1)
    }
  }
}



#' Check the validity of ppp samples arranged in matrix format
#'
#' @description Standard checks for a matrix of ordered times
#' (event series in rows, times in columns). Check
#' that the times in the columns are sorted, have unique values
#' in `[t_min, t_max]`, and has length `size` (if applicable).
#'
#' @param times (vector, double | matrix) the times to be checked as vectors or matrices (time-vectors in rows)
#' @param t_min (double | vector) the start of the time nterval
#' @param t_max (double| vector) optional: the end of the time interval; if a vector, its length should match the number of rows of `times`.
#' @param size (double) optional: the size of the vector
#' @param atmost1 (boolean) optional: at most one sample returned
#' @param atleast1 (boolean) optional: at least one sample returned
#' @return None
#' @keywords internal
check_ppp_sample_validity <- function(times, t_min, t_max = NULL, size = NULL, atmost1 = FALSE, atleast1 = FALSE) {
  if (!is.matrix(times)) {
    check_ppp_vector_validity(times = times, t_min = t_min, t_max = t_max, size = size, atmost1 = atmost1, atleast1 = atleast1)
  } else {
    if (length(t_min) == 1) {
      t_min <- rep(t_min, nrow(times))
    }
    stopifnot(nrow(times) == length(t_min))
    if (!is.null(t_max)) {
      if (length(t_max) == 1) {
        t_max <- rep(t_max, nrow(times))
      }
      stopifnot(nrow(times) == length(t_max))
    }
    for (i in 1:nrow(times)) {
      tmax_i <- if (!is.null(t_max)) t_max[i] else NULL
      testthat::expect_identical(times[i, !is.na(times[i, ])], sort(times[i, ], na.last = NA))
      check_ppp_vector_validity(times = times[i, ], t_min = t_min[i], t_max = tmax_i, size = size, atmost1 = atmost1, atleast1 = atleast1)
    }
  }
}

#' Check that two ppp vectors Q-Q agree
#'
#' @description Compare that the deciles of two vectors have absolute difference
#' over average ratios less than `threshold`
#'
#' @param ppp1 (vector, double) the first vector
#' @param ppp2 (vector, double) the second vector
#' @param threshold (double) optional: the cutoff for a large absolute threshold
#' @param showQQ (boolean) optional: show the QQ plot if the absolute value of the
#' Difference vs Average ratio in any decile is bigger than the `threshold`
#' @return None
#' @keywords internal
compare_ppp_vectors <- function(ppp1,
                                ppp2,
                                threshold = 0.15, showQQ = TRUE) {
  res <- stats::qqplot(ppp1, ppp2, plot.it = FALSE)
  r1 <- res[[1]]
  r2 <- res[[2]]
  step <- floor(length(r1) / 10)

  decile_check <- logical(0)
  for (i in 1:9) {
    indices <- ((i - 1) * step + 1):(i * step)
    DvsA <- 2 * (r1[indices] - r2[indices]) / (r1[indices] + r2[indices])
    tmp <- stats::t.test(x = DvsA)
    decile_check[i] <- abs(tmp$estimate) < threshold
  }
  indices <- ((9) * step + 1):length(r1)
  DvsA <- 2 * (r1[indices] - r2[indices]) / (r1[indices] + r2[indices])
  tmp <- stats::t.test(x = DvsA)
  decile_check[10] <- abs(tmp$estimate) < threshold

  if (!all(decile_check)) {
    stats::qqplot(r1, r2, plot.it = TRUE)
    graphics::lines(rep(min(c(r1, r2)), 2), rep(max(c(r1, r2)), 2))
  }
  testthat::expect_true(all(decile_check))
}

test_that("ppp2() works", {
  expect_no_error(
    withr::with_seed(
      12345,
      df1 <- ppp2(rate = 1, t_min = 0, t_max = 10)
    )
  )
  check_ppp_sample_validity(times = df1, t_min = 0, t_max = 10)
})

#' Numerically evaluate the inverse of a function at a specific point
#'
#' @param f (function) the function to be inverted; must be continuous
#'            and increasing
#' @param y (scalar, double) the f(x)=y value in which to evaluate the inverse
#' @param min_x (scalar, double) the min of the domain of f()
#' @param max_x (scalar, double) the max of the domain of f()
#' @param min_y (scalar, double) the min in the range of f()
#' @param max_y (scalar, double) the max in the range of f()
#'
#' @return (scalar, double) vector of x=f^(-1)(y): the inverted value
#' @export
#'
#' @keywords internal
#' @examples
#' inverse_with_uniroot(f = function(x) {
#'   2 * x
#' }, y = 0.5)
inverse_with_uniroot <- function(f = f, y,
                                 min_x = 0, max_x = 1,
                                 min_y = f(min_x), max_y = f(max_x)) {
  x_star <- stats::uniroot((function(x) f(x) - y),
    lower = min_x,
    upper = max_x,
    f.lower = min_y - y,
    f.upper = max_y - y
  )$root
  return(x_star)
}


#' Numerically evaluate the inverse of a monotonically increasing continuous
#' function from R to R at specific points.
#'
#' @param f (function) the function to be inverted; must be continuous
#'            and increasing
#' @param y (vector, double) the f(x)=y values in which to evaluate the inverse;
#'          must be in ascending order
#' @param range_x (vector, double) the min and max of the domain of f()
#' @param range_y (vector, double) the min and max in the range of f()
#'
#' @return (vector, double) vector of x=f^(-1)(y): the inverted values
#' @export
#'
#' @keywords internal
#' @examples
#' inverse_with_uniroot_sorted(f = function(x) {
#'   2 * x
#' }, y = c(0, 0.5))
inverse_with_uniroot_sorted <- function(f, y,
                                        range_x = c(0, 10),
                                        range_y = c(f(range_x[1]), f(range_x[2]))) {
  len <- length(y)
  if (y[1] < range_y[1] || y[len] > range_y[2]) stop()

  x <- rep(NA, len)
  x[1] <- inverse_with_uniroot(
    f = f, y[1],
    min_x = range_x[1], max_x = range_x[2],
    min_y = range_y[1], max_y = range_y[2]
  )

  if (len == 1) {
    return(x)
  }

  for (i in 2:len) {
    x[i] <- inverse_with_uniroot(
      f = f, y[i],
      min_x = x[i - 1], max_x = range_x[2],
      min_y = y[i - 1], max_y = range_y[2]
    )
  }
  return(x)
}

test_that("inverse_with_uniroot() works", {
  expect_equal(inverse_with_uniroot(f = function(x) {
    2 * x
  }, y = 0.5), 0.25)
})


test_that("vectorization inverse_with_uniroot_sorted() works", {
  expect_equal(inverse_with_uniroot_sorted(f = function(x) {
    2 * x
  }, y = c(0, 0.5)), c(0, 0.25))
})

### Add testst hat show how these functions break

#' Vectorized sampling from a zero-truncated non homogeneous Poisson Point Process (NHPPP) from
#'    an interval (thinning method) with piecewise constant_majorizers
#'
#' Vectorized sampling from a zero-truncated non homogeneous Poisson Point Process (NHPPP) from
#'    an interval (thinning method) with piecewise constant_majorizers.
#'    The majorizers are step functions over equal-length time intevals.
#'
#' @param lambda (function) a vectorized intensity function, with one or two arguments.
#'  The first is time. The optional second is a named list with additional arguments.
#' @param lambda_args (list) optional named list of arguments to pass to `lambda`.
#'        If you have arguments for `lambda` that vary by draw, they should be passed as
#'        a data.table named `vector_arguments`.
#' @param Lambda_maj_matrix (matrix) for the majorizeintegrated intensity rates at the end of each interval
#' @param lambda_maj_matrix (matrix) intensity rates, one per interval
#' @param range_t (vector, or matrix) `t_min` and `t_max`, possibly vectorized
#' @param tol (scalar, double) tolerance for the number of events
#' @param atmost1 boolean, draw at most 1 event time
#' @param ... (any) other arguments (ignored  -- used for flexibility in calling from other functions)
#'
#' @return a matrix of event times (columns) per draw (rows)
#'         NAs are structural empty spots
#' @keywords internal
vztdraw_intensity <- function(
    lambda = NULL,
    lambda_args = NULL,
    Lambda_maj_matrix = NULL,
    lambda_maj_matrix = NULL,
    range_t = NULL,
    tol = 10^-6,
    atmost1 = FALSE,
    ...) {
  return(
    vztdraw_intensity_step_regular(
      lambda = lambda,
      lambda_args = lambda_args,
      Lambda_maj_matrix = Lambda_maj_matrix,
      lambda_maj_matrix = lambda_maj_matrix,
      range_t = range_t,
      tol = tol,
      atmost1 = atmost1,
      ...
    )
  )
}

#' Vectorized sampling from NHPPPs with piecewise constant intensities
#' with same interval lengths
#'
#' @description
#' Simulate a piecewise constant-rate Poisson Point Process over `(t_min, t_max]` (inversion method)
#' where the intervals have the same length (are "regular").
#'
#' @param Lambda_matrix (matrix) integrated intensity rates at the end of each interval
#' @param lambda_matrix (matrix) intensity rates, one per interval
#' @param rate_matrix_t_min (scalar | vector | column matrix) is the lower bound
#'        of the time interval for each row of (Lambda|lambda)_maj_matrix.
#'        The length of this argument is the number of point processes that should be drawn.
#' @param rate_matrix_t_max (scalar | vector | column matrix) the upper bound
#'        of the time interval for each row of (Lambda|lambda)_maj_matrix.
#'        The length of this argument is the number of point processes that should be drawn.
#' @param t_min (scalar | vector | column matrix) is the lower bound
#'        of a subinterval of (rate_matrix_t_min, rate_matrix_t_max]. If set,
#'        times are sampled from the subinterval.
#'        If omitted, it is equivalent to `rate_matrix_t_min`.
#' @param t_max (scalar | vector | column matrix) is the upper bound
#'        of a subinterval of (rate_matrix_t_min, rate_matrix_t_max]. If set,
#'        times are sampled from the subinterval.
#'        If omitted, it is equivalent to `rate_matrix_t_max`.
#' @param tol (scalar, double) tolerance for the number of events
#' @param atmost1 boolean, draw at most 1 event time
#' @param atmostB If not NULL, draw at most B (B>0) event times. NULL means ignore.
#' @param atleast1 boolean, draw at least 1 event time
#'
#' @return a vector of event times t
#'         if no events realize, it will have 0 length
#'
#' @examples
#' x <- vdraw_sc_step_regular(
#'   Lambda_matrix = matrix(1:5, nrow = 1),
#'   rate_matrix_t_min = 100,
#'   rate_matrix_t_max = 110,
#'   atmost1 = FALSE
#' )
#' @export
vdraw_sc_step_regular <- function(
    lambda_matrix = NULL,
    Lambda_matrix = NULL,
    rate_matrix_t_min = NULL,
    rate_matrix_t_max = NULL,
    t_min = NULL,
    t_max = NULL,
    tol = 10^-6,
    atmost1 = FALSE,
    atmostB = NULL,
    atleast1 = FALSE) {
  if (isFALSE(atleast1)) {
    return(
      vdraw_sc_step_regular_cpp(
        lambda_matrix = lambda_matrix,
        Lambda_matrix = Lambda_matrix,
        rate_matrix_t_min = rate_matrix_t_min,
        rate_matrix_t_max = rate_matrix_t_max,
        t_min = t_min,
        t_max = t_max,
        tol = tol,
        atmost1 = atmost1,
        atmostB = atmostB
      )
    )
  }
  if (isTRUE(atleast1)) {
    return(
      vztdraw_sc_step_regular_cpp(
        lambda_matrix = lambda_matrix,
        Lambda_matrix = Lambda_matrix,
        rate_matrix_t_min = rate_matrix_t_min,
        rate_matrix_t_max = rate_matrix_t_max,
        t_min = t_min,
        t_max = t_max,
        atmost1 = atmost1
      )
    )
  }
  stop("never here")
}

#' Human mortality database age and sex specific rates for all cause deaths
#'
#' This is the 2015 annual death rates from the 2023 version of the
#' Human Mortality Database for the USA.
#'
#' @keywords internal
#' @format  ## `annual_mortality_rates_2015`
#' a data.table with 3 rows and 113 columns.
#' \describe{
#'    \item{birth_cohort}{Birth cohort as a calendar year}
#'    \item{sex}{sex: `female`, `male`, `total`.}
#'    \item{age_0, ..., age 110+}{age-specific death rates}
#' }
#' @source <https://www.mortality.org>
"annual_mortality_rates_2015"

#' Simulate a homogeneous Poisson Point Process over (t_min, t_max] (order statistics method)
#'
#' @description
#'   `r lifecycle::badge("deprecated")`
#' Use `ppp2` instead.
#' @param range_t (vector, double) min and max of the time interval
#' @param rate (scalar, double) constant instantaneous rate
#' @param rng_stream an `rstream` object
#' @param atmost1 boolean, draw at most 1 event time
#'
#' @return a vector of event times t
#'         if no events realize, it will have 0 length
#' @export
#'
#' @keywords internal
#' @examples
#' x <- ppp_orderstat(range_t = c(0, 10), rate = 1)
ppp_orderstat <- function(range_t = c(0, 10), rate = 1, rng_stream = NULL, atmost1 = FALSE) {
  # we expect lambda = t_max*rate events
  lifecycle::deprecate_warn(when = "0.5.0", what = "ppp_orderstat()", with = "ppp2()")
  n <- rng_stream_rpois(size = 1, lambda = rate * (range_t[2] - range_t[1]), rng_stream = rng_stream)
  tmp <- ppp_n(size = n, range_t = range_t, rng_stream = rng_stream)
  if (atmost1 == TRUE) {
    return(tmp[1])
  }
  return(tmp)
}

#' Special case: Simulate from a non homogeneous Poisson Point Process (NHPPP) from
#'    (t_min, t_max) with linear intensity function (inversion method)
#'
#' @description  Sample NHPPP times from a linear intensity function
#' using the inversion method, optionally using an `rstream`
#' generator
#'
#' @param intercept (double) the intercept
#' @param slope (double) the slope
#' @param t_min (double) lower bound of the time interval
#' @param t_max (double) upper bound of the time interval
#' @param atmost1 boolean, draw at most 1 event time
#'
#' @return a vector of event times (t_); if no events realize,
#'         a vector of length 0
#' @export
#'
#' @examples
#' x <- draw_sc_linear(intercept = 0, slope = 0.2, t_min = 0, t_max = 10)
#'
draw_sc_linear <- function(intercept,
                           slope,
                           t_min,
                           t_max,
                           atmost1 = FALSE) {
  if ((slope <= 0 && intercept <= 0) || (intercept + slope * t_min < 0)) {
    return(c())
  }
  if (slope == 0) {
    return(ppp2(rate = intercept, t_min = t_min, t_max = t_max, atmost1 = atmost1))
  }
  if (slope < 0) {
    t_upper <- -intercept / slope
    t_max <- min(t_upper, t_max)
  }
  return(
    draw_cumulative_intensity_inversion(
      Lambda = function(t) Lambda_linear_form(t, intercept = intercept, slope = slope, t0 = t_min),
      Lambda_inv = function(z) Lambda_inv_linear_form(z, intercept = intercept, slope = slope, t0 = t_min),
      t_min = t_min,
      t_max = t_max,
      atmost1 = atmost1
    )
  )
}

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

find_lower_bound_index <- function(L, start, tau) {
    .Call(`_nhppp_find_lower_bound_index`, L, start, tau)
}

find_upper_bound_index <- function(L, start, tau) {
    .Call(`_nhppp_find_upper_bound_index`, L, start, tau)
}

matrix_cumsum_columns <- function(X) {
    .Call(`_nhppp_matrix_cumsum_columns`, X)
}

matrix_cumsum_columns_inplace <- function(X) {
    invisible(.Call(`_nhppp_matrix_cumsum_columns_inplace`, X))
}

matrix_diff_columns <- function(X) {
    .Call(`_nhppp_matrix_diff_columns`, X)
}

matrix_diff_columns_inplace <- function(X) {
    invisible(.Call(`_nhppp_matrix_diff_columns_inplace`, X))
}

rztpois <- function(lambda) {
    .Call(`_nhppp_rztpois`, lambda)
}

rztpois_vec <- function(lambda) {
    .Call(`_nhppp_rztpois_vec`, lambda)
}

step_regular_inverse <- function(max_events, Lambda, Tau, range_t, atmost1) {
    .Call(`_nhppp_step_regular_inverse`, max_events, Lambda, Tau, range_t, atmost1)
}

vdraw_intensity_step_regular <- function(lambda, rate_maj, is_cumulative, range_t, subinterval, use_subinteval, tol, atmost1, atmostB) {
    .Call(`_nhppp_vdraw_intensity_step_regular`, lambda, rate_maj, is_cumulative, range_t, subinterval, use_subinteval, tol, atmost1, atmostB)
}

vdraw_sc_step_regular <- function(rate, is_cumulative, range_t, tol, atmost1) {
    .Call(`_nhppp_vdraw_sc_step_regular`, rate, is_cumulative, range_t, tol, atmost1)
}

vdraw_sc_step_regular2 <- function(rate, is_cumulative, range_t, subinterval, tol, atmost1, atmostB) {
    .Call(`_nhppp_vdraw_sc_step_regular2`, rate, is_cumulative, range_t, subinterval, tol, atmost1, atmostB)
}

vztdraw_sc_step_regular <- function(rate, is_cumulative, range_t, atmost1) {
    .Call(`_nhppp_vztdraw_sc_step_regular`, rate, is_cumulative, range_t, atmost1)
}

vztdraw_sc_step_regular2 <- function(rate, is_cumulative, range_t, subinterval, atmost1) {
    .Call(`_nhppp_vztdraw_sc_step_regular2`, rate, is_cumulative, range_t, subinterval, atmost1)
}

test_that("matrix_diff_columns[_inplace]() works", {
  set.seed(123)
  N <- 100
  J0 <- J <- matrix(rnorm(10 * N), ncol = 10)
  J_target <- J
  for (i in 10:2) {
    J_target[, i] <- J_target[, i] - J_target[, i - 1]
  }

  J_R <- mat_diff_columns(J)
  expect_equal(J_R, J_target) # The  R code version of the function
  expect_false(data.table::address(J_R) == data.table::address(J))
  expect_equal(J, J0)

  J_Cpp <- matrix_diff_columns(J)
  expect_equal(J_Cpp, J_target) # The C++ version of the function -- not inplace
  expect_false(data.table::address(J_Cpp) == data.table::address(J))
  expect_equal(J, J0)

  expect_no_error(matrix_diff_columns_inplace(J))
  expect_equal(J, J_target)

  # cumsum/diff cancel out
  expect_equal(J0, matrix_diff_columns(matrix_cumsum_columns(J0)))

  # works with 1 column:
  expect_equal(J0[, 1, drop = FALSE], mat_diff_columns(J0[, 1, drop = FALSE]))
  expect_equal(J0[, 1, drop = FALSE], matrix_diff_columns(J0[, 1, drop = FALSE]))
})

#' Vectorized generic function for simulating from NHPPPs given the intensity function
#' or the cumulative intensity function
#'
#' @description
#' This is a wrapper to the package's specific functions, and thus slightly slower.
#' For time-intensive simulations prefer one of the specific functions.
#'
#' @param lambda (function) intensity function, vectorized
#' @param lambda_args (list) optional named list of arguments to pass to `lambda`.
#'        If you have arguments for `lambda` that vary by draw, they should be passed as
#'        a data.table named `vector_arguments`.
#' @param Lambda_maj_matrix (matrix) integrated intensity rates at the end of each interval
#' @param lambda_maj_matrix (matrix) intensity rates, one per interval
#' @param Lambda (function, double vector) an increasing function
#'        which is the integrated rate of the NHPPP.
#'        It should take a vectorized argument t for times and an optional arguments list.
#' @param Lambda_inv (function, double vector) the inverse of `Lambda()`, also in vectorized form
#'        It should take a vectorized argument z and an optional arguments list.
#' @param Lambda_args (list) optional named list of arguments to pass to `Lambda`.
#'        If you have arguments for `Lambda` that vary by draw, they should be passed as
#'        a data.table named `vector_arguments`.
#' @param Lambda_inv_args (list) optional named list of arguments to pass to `Lambda_inv`.
#'        If you have arguments for `Lambda_inv` that vary by draw, they should be passed as
#'        a data.table named `vector_arguments`.
#' @param t_min (scalar | vector | column matrix) is the lower bound
#'        of a subinterval of (rate_matrix_t_min, rate_matrix_t_max]. If set,
#'        times are sampled from the subinterval.
#'        If omitted, it is equivalent to `rate_matrix_t_min`.
#' @param t_max (scalar | vector | column matrix) is the upper bound
#'        of a subinterval of (rate_matrix_t_min, rate_matrix_t_max]. If set,
#'        times are sampled from the subinterval.
#'        If omitted, it is equivalent to `rate_matrix_t_max`.
#' @param rate_matrix_t_min (scalar | vector | column matrix) is the lower bound
#'        of the time interval for each row of (Lambda|lambda)_maj_matrix.
#'        The length of this argument is the number of point processes that should be drawn.
#' @param rate_matrix_t_max (scalar | vector | column matrix) the upper bound
#'        of the time interval for each row of (Lambda|lambda)_maj_matrix.
#'        The length of this argument is the number of point processes that should be drawn.
#' @param tol (scalar, double) tolerance for the number of events
#' @param atmost1 boolean, draw at most 1 event time
#' @param atleast1 boolean, draw at least 1 event time
#' @param atmostB If not NULL, draw at most B (B>0) event times. NULL means ignore.

#'
#' @return a vector of event times
#' @export
vdraw <- function(
    lambda = NULL,
    lambda_args = NULL,
    Lambda_maj_matrix = NULL,
    lambda_maj_matrix = NULL,
    Lambda = NULL,
    Lambda_inv = NULL,
    Lambda_args = NULL,
    Lambda_inv_args = NULL,
    t_min = NULL,
    t_max = NULL,
    rate_matrix_t_min = NULL,
    rate_matrix_t_max = NULL,
    tol = 10^-6,
    atmost1 = FALSE,
    atleast1 = FALSE,
    atmostB = NULL) {
  if (!is.null(lambda) &&
    !(is.null(lambda_maj_matrix) & is.null(lambda_maj_matrix))) {
    return(
      vdraw_intensity(
        lambda = lambda,
        lambda_args = lambda_args,
        Lambda_maj_matrix = Lambda_maj_matrix,
        lambda_maj_matrix = lambda_maj_matrix,
        rate_matrix_t_min = rate_matrix_t_min,
        rate_matrix_t_max = rate_matrix_t_max,
        t_min = t_min,
        t_max = t_max,
        tol = tol,
        atmost1 = atmost1,
        atleast1 = atleast1,
        atmostB = atmostB
      )
    )
  }

  return(
    vdraw_cumulative_intensity(
      Lambda = Lambda,
      Lambda_inv = Lambda_inv,
      t_min = t_min,
      t_max = t_max,
      Lambda_args = Lambda_args,
      Lambda_inv_args = Lambda_inv_args,
      tol = tol,
      atmost1 = atmost1,
      atleast1 = atleast1
    )
  )
}

#' Simulate exactly `n` points from a homogeneous Poisson Point Process over (t_min, t_max]
#'
#' @param n (int) the number of points to be simulated
#' @param t_min (double) the lower bound of the time interval
#' @param t_max (double) the upper bound of the time interval
#'
#' @return a vector of event times of size `n`
#' @export
#'
#' @examples
#' x <- ppp_exactly_n(n = 10, t_min = 0, t_max = 10)
ppp_exactly_n <- function(n, t_min, t_max) {
  U <- stats::runif(n = n, min = t_min, max = t_max)
  if (n == 1) {
    return(U)
  }
  return(U[order(U, method = "shell")])
}

#' Sampling from NHPPPs with piecewise constant intensities
#' with same interval lengths (non-vectorized)
#'
#' @param Lambda_vector (scalar, double) `K` integrated intensity rates at the end of each interval
#' @param lambda_vector (scalar, double) `K` constant intensity rates, one per interval
#' @param t_min (scalar, double) lower bound of the time interval
#' @param t_max (scalar, double) upper bound of the time interval
#' @param atmost1 boolean, draw at most 1 event time
#' @param atleast1 boolean, draw at least 1 event time
#'
#' @return a vector of event times t
#'         if no events realize, it will have 0 length
#' @export
#'
#' @examples
#' x <- draw_sc_step_regular(Lambda_vector = 1:5, t_min = 0, t_max = 5)
#' @export
draw_sc_step_regular <- function(Lambda_vector = NULL,
                                 lambda_vector = NULL,
                                 t_min = NULL,
                                 t_max = NULL,
                                 atmost1 = FALSE,
                                 atleast1 = FALSE) {
  stopifnot(!is.null(t_min) && !is.null(t_max))
  if (is.null(Lambda_vector) && !is.null(lambda_vector)) {
    Lambda_vector <- cumsum(lambda_vector)
  }
  n_intervals <- length(Lambda_vector)
  interval_length <- (t_max - t_min) / n_intervals
  Lambda_vector <- c(0, Lambda_vector)

  if (atleast1 == FALSE) {
    ppp_t_fun <- ppp2
  } else {
    ppp_t_fun <- ztppp
  }

  if (n_intervals == 1) {
    tau <- ppp_t_fun(rate = (Lambda_vector[2] - Lambda_vector[1]) / interval_length, t_min = t_min, t_max = t_max, atmost1 = atmost1)
    return(tau[!is.na(tau)])
  }

  tau <- ppp_t_fun(rate = 1, t_min = 0, t_max = Lambda_vector[n_intervals + 1], atmost1 = atmost1)

  n_events <- length(tau)
  if (n_events == 0 || any(is.na(tau))) {
    return(numeric(0))
  }

  tau_indices <- rep(1:n_events, each = n_intervals)
  Lambda_indices_low <- rep(1:n_intervals, n_events)
  Lambda_indices_high <- rep(2:(n_intervals + 1), n_events)

  tau_in_interval <- (tau[tau_indices] >= Lambda_vector[Lambda_indices_low] &
    tau[tau_indices] < Lambda_vector[Lambda_indices_high])



  t_ <- t_min +
    interval_length * (
      (Lambda_indices_low[tau_in_interval] - 1) +
        (tau - Lambda_vector[Lambda_indices_low[tau_in_interval]]) /
          (
            Lambda_vector[Lambda_indices_high[tau_in_interval]] -
              Lambda_vector[Lambda_indices_low[tau_in_interval]]
          )
    )

  return(t_)
}

test_that("draw_sc_step_regular() arguments work", {
  set.seed(123)
  l <- stats::runif(10)
  L <- cumsum(l)

  expect_no_error(df0 <- draw_sc_step_regular(t_min = 0, t_max = 10, lambda_vector = 1, atmost1 = TRUE))
  check_ppp_sample_validity(times = df0, t_min = 0, t_max = 10, atmost1 = TRUE)

  expect_no_error(df0a <- draw_sc_step_regular(t_min = 0, t_max = 10, lambda_vector = l, atmost1 = TRUE))
  check_ppp_sample_validity(times = df0a, t_min = 0, t_max = 10, atmost1 = TRUE)

  expect_no_error(df0b <- draw_sc_step_regular(t_min = 0, t_max = 10, Lambda_vector = 1, atmost1 = TRUE))
  check_ppp_sample_validity(times = df0b, t_min = 0, t_max = 10, atmost1 = TRUE)

  expect_no_error(df0c <- draw_sc_step_regular(t_min = 0, t_max = 10, Lambda_vector = L, atmost1 = TRUE))
  check_ppp_sample_validity(times = df0c, t_min = 0, t_max = 10, atmost1 = TRUE)

  expect_no_error(df1 <- draw_sc_step_regular(t_min = 0, t_max = 10, Lambda_vector = 1))
  check_ppp_sample_validity(times = df1, t_min = 0, t_max = 10)

  expect_no_error(df1a <- draw_sc_step_regular(t_min = 0, t_max = 10, Lambda_vector = L))
  check_ppp_sample_validity(times = df1a, t_min = 0, t_max = 10)

  expect_no_error(df1b <- draw_sc_step_regular(t_min = 0, t_max = 10, Lambda_vector = L / 1000, atleast1 = TRUE))
  check_ppp_sample_validity(times = df1b, t_min = 0, t_max = 10, atleast1 = TRUE)

  expect_no_error(df1c <- draw_sc_step_regular(t_min = 0, t_max = 10, Lambda_vector = L / 1000, atmost1 = TRUE, atleast1 = TRUE))
  check_ppp_sample_validity(times = df1c, t_min = 0, t_max = 10, atmost1 = TRUE, atleast1 = TRUE)
})

#' Vectorized simulation from a non homogeneous Poisson Point Process (NHPPP) from
#'    (t_min, t_max) given the cumulative intensity function and its inverse
#'
#' @description  Sample NHPPP times using the cumulative intensity function and its inverse.
#' @param Lambda (function, double vector) an increasing function
#'        which is the integrated rate of the NHPPP.
#'        It should take a vectorized argument t for times and an optional arguments list.
#' @param Lambda_inv (function, double vector) the inverse of `Lambda()`, also in vectorized form
#'        It should take a vectorized argument z and an optional arguments list.
#' @param t_min (scalar | vector | column matrix) the lower bound of the interval for each sampled point process
#'        The length of this argument is the number of point processes that should be drawn.
#' @param t_max (scalar | vector | column matrix) the upper bound of the interval for each sampled point process
#'        The length of this argument is the number of point processes that should be drawn.
#' @param Lambda_args (list) optional named list of arguments to pass to `Lambda`.
#'        If you have arguments for `Lambda` that vary by draw, they should be passed as
#'        a data.table named `vector_arguments`.
#' @param Lambda_inv_args (list) optional named list of arguments to pass to `Lambda_inv`.
#'        If you have arguments for `Lambda_inv` that vary by draw, they should be passed as
#'        a data.table named `vector_arguments`.
#' @param tol the tolerange for the calulations.
#' @param atmost1 boolean, draw at most 1 event time per sampled point process.
#' @param atleast1 boolean, draw at least 1 event time
#'
#' @return a matrix of event times with one row per sampled point process.
#' @export
#'
vdraw_cumulative_intensity <- function(Lambda,
                                       Lambda_inv,
                                       t_min,
                                       t_max,
                                       Lambda_args = NULL,
                                       Lambda_inv_args = NULL,
                                       tol = 10^-6,
                                       atmost1 = FALSE,
                                       atleast1 = FALSE) {
  if (atleast1 == TRUE) {
    stop("Option `atleast1 = TRUE` has not been implemented yet for vectorized functions.")
  }
  range_t <- cbind(as.vector(t_min), as.vector(t_max))
  N_rows <- nrow(range_t)
  range_L <- Lambda(range_t, Lambda_args = Lambda_args)

  if (isTRUE(atmost1)) {
    N_cols <- 1
  } else {
    N_cols <- max(stats::qpois(p = 1 - tol, lambda = 1 * (range_L[, 2] - range_L[, 1])))
  }

  warped_t <- matrix(stats::rexp(n = N_cols * N_rows, rate = 1), ncol = N_cols)
  matrix_cumsum_columns_inplace(warped_t)
  warped_t <- warped_t + range_L[, 1]
  for (col in 1:N_cols) {
    in_range_L <- (warped_t[, col] <= range_L[, 2])
    if (col > 1 && all(!in_range_L)) {
      warped_t <- warped_t[, 1:(col - 1)]
      break
    }
    warped_t[!in_range_L, col] <- NA
  }
  return(Lambda_inv(warped_t, Lambda_inv_args = Lambda_inv_args))
}

test_that("vdraw_cumulative_intensity_inversion() works with minimal options", {
  L <- function(t, ...) {
    return(2 * t)
  }
  Li <- function(z, ...) {
    return(z / 2)
  }

  t0_a <- 0.5
  t1_a <- 1.5
  t0_b <- rep(0.5, 10) + runif(n = 10)
  t1_b <- rep(2, 10) + runif(n = 10)
  t0_c <- matrix(t0_b, nrow = 1)
  t1_c <- matrix(t1_b, nrow = 1)
  t0_d <- matrix(t0_b, ncol = 1)
  t1_d <- matrix(t1_b, ncol = 1)


  # scalars
  expect_no_error(df0 <- vdraw_cumulative_intensity(Lambda = L, Lambda_inv = Li, t_min = t0_a, t_max = t1_a, atmost1 = FALSE))
  check_ppp_sample_validity(df0, t_min = t0_a, t_max = t1_a)

  # vectors
  expect_no_error(df1 <- vdraw_cumulative_intensity(Lambda = L, Lambda_inv = Li, t_min = t0_b, t_max = t1_b, atmost1 = FALSE))
  check_ppp_sample_validity(df1, t_min = t0_b, t_max = t1_b)

  # row matrices
  expect_no_error(df10 <- vdraw_cumulative_intensity(Lambda = L, Lambda_inv = Li, t_min = t0_c, t_max = t1_c, atmost1 = FALSE))
  check_ppp_sample_validity(df10, t_min = t0_b, t_max = t1_b)

  # col matrices
  expect_no_error(df10 <- vdraw_cumulative_intensity(Lambda = L, Lambda_inv = Li, t_min = t0_d, t_max = t1_d, atmost1 = FALSE))
  check_ppp_sample_validity(df10, t_min = t0_b, t_max = t1_b)


  # mixed arguments and atmost1 = TRUE
  expect_no_error(df10.1 <- vdraw_cumulative_intensity(Lambda = L, Lambda_inv = Li, t_min = t0_a, t_max = t1_d, atmost1 = TRUE))
  check_ppp_sample_validity(df10.1, t_min = t0_a, t_max = t1_b, atmost1 = TRUE)
})

test_that("vdraw_cumulative_intensity_inversion() works when functions take arguments", {
  L <- function(t, Lambda_args) {
    return(Lambda_args$a * t)
  }
  Li <- function(z, Lambda_inv_args) {
    return(z / Lambda_inv_args$a)
  }

  args <- list(a = 2)

  t0_a <- 0.5
  t1_a <- 1.5
  t0_b <- rep(0.5, 10) + runif(n = 10)
  t1_b <- rep(2, 10) + runif(n = 10)

  # scalars
  expect_no_error(df0 <- vdraw_cumulative_intensity(Lambda = L, Lambda_inv = Li, t_min = t0_a, t_max = t1_a, Lambda_args = args, Lambda_inv_args = args, atmost1 = FALSE))
  check_ppp_sample_validity(df0, t_min = t0_a, t_max = t1_a)
  # vectors
  expect_no_error(df1 <- vdraw_cumulative_intensity(Lambda = L, Lambda_inv = Li, t_min = t0_b, t_max = t1_b, Lambda_args = args, Lambda_inv_args = args, atmost1 = FALSE))
  check_ppp_sample_validity(df1, t_min = t0_b, t_max = t1_b)
})

test_that("vdraw_cumulative_intensity_inversion() uses blocked random numbers", {
  set.seed(123)
  L <- function(t, ...) {
    return(2 * t)
  }
  Li <- function(z, ...) {
    return(z / 2)
  }

  t0_a <- 0.5
  t1_a <- 1.5
  t0_b <- rep(0.5, 10) + runif(n = 10)
  t1_b <- rep(2, 10) + runif(n = 10)


  Z0 <- list()
  for (i in 1:2) {
    set.seed(123)
    expect_no_error(Z0[[i]] <- vdraw_cumulative_intensity(Lambda = L, Lambda_inv = Li, t_min = t0_b, t_max = t1_b, atmost1 = FALSE))
    if (i > 1) {
      expect_true(identical(Z0[[1]], Z0[[i]]))
    }
  }
  check_ppp_sample_validity(Z0[[1]], t_min = t0_b, t_max = t1_b)
})

#' Generic function for simulating from zero-truncated NHPPPs given the intensity function.
#'
#' @description Sample from zero-truncated NHPPP given the intensity function
#' This is a wrapper to the package's specific functions, and thus somewhat slower.
#' For time-intensive simulations prefer one of the specific functions.
#'
#' @param lambda (function) the instantaneous rate
#' @param line_majorizer_intercept The intercept `alpha` of the [log]linear majorizer function: `alpha + beta * t` or `exp(alpha + beta * t)`
#' @param line_majorizer_slope The slope `beta` of the [log]linear majorizer function: `alpha + beta * t` or `exp(alpha + beta * t)`
#' @param line_majorizer_is_loglinear (boolean) if `TRUE` the majorizer is loglinear `exp(alpha + beta * t)`; if `FALSE` it is a linear function
#' @param step_majorizer_vector (vector, double) `K` constant majorizing rates, one per interval; all intervals are of equal length (regular)
#' @param t_min (double) the lower bound of the interval
#' @param t_max (double) the upper bound of the interval
#' @param atmost1 boolean, draw at most 1 event time
#'
#' @return a vector of at least 1 event times
#' @keywords internal
ztdraw_intensity <- function(
    lambda,
    line_majorizer_intercept = NULL,
    line_majorizer_slope = NULL,
    line_majorizer_is_loglinear = FALSE,
    step_majorizer_vector = NULL,
    t_min = NULL,
    t_max = NULL,
    atmost1 = FALSE) {
  if (!is.null(step_majorizer_vector)) {
    return(
      ztdraw_intensity_step(
        lambda = lambda,
        majorizer_vector = step_majorizer_vector,
        time_breaks = seq.int(from = t_min, to = t_max, length.out = length(step_majorizer_vector) + 1),
        atmost1 = atmost1
      )
    )
  }
  return(
    ztdraw_intensity_line(
      lambda = lambda,
      majorizer_intercept = line_majorizer_intercept,
      majorizer_slope = line_majorizer_slope,
      majorizer_is_loglinear = line_majorizer_is_loglinear,
      t_min = t_min,
      t_max = t_max,
      atmost1 = atmost1
    )
  )
}

test_that("ztdraw_intensity() works", {
  set.seed(123)
  l <- function(t) {
    return(2)
  }
  lmaj_intercept <- 2.2
  lmaj_slope <- 0

  expect_no_error(withr::with_preserve_seed(df <- ztdraw_intensity(lambda = l, line_majorizer_intercept = lmaj_intercept, line_majorizer_slope = lmaj_slope, t_min = 0, t_max = 10)))
  check_ppp_sample_validity(times = df, t_min = 0, t_max = 10, atleast1 = TRUE)

  # works when range_t[1]>0
  expect_no_error(withr::with_preserve_seed(df <- ztdraw_intensity(lambda = l, line_majorizer_intercept = lmaj_intercept, line_majorizer_slope = lmaj_slope, t_min = 5, t_max = 10)))
  check_ppp_sample_validity(times = df, t_min = 5, t_max = 10, atleast1 = TRUE)

  # works with atmost1
  expect_no_error(withr::with_preserve_seed(df2 <- ztdraw_intensity(lambda = l, line_majorizer_intercept = lmaj_intercept, line_majorizer_slope = lmaj_slope, t_min = 5, t_max = 10, atmost1 = TRUE)))
  check_ppp_sample_validity(times = df2, t_min = 5, t_max = 10, atleast1 = TRUE, atmost1 = TRUE)


  lambda_maj <- c(2, 2.1, 2.9, 4)

  expect_no_error(withr::with_preserve_seed(df <- ztdraw_intensity(lambda = l, t_min = 0, t_max = 10, step_majorizer_vector = lambda_maj)))
  check_ppp_sample_validity(times = df, t_min = 0, t_max = 10, atleast1 = TRUE)

  # works when range_t[1]>0
  expect_no_error(withr::with_preserve_seed(df <- ztdraw_intensity(lambda = l, t_min = 5, t_max = 10, step_majorizer_vector = lambda_maj)))
  check_ppp_sample_validity(times = df, t_min = 5, t_max = 10, atleast1 = TRUE)

  # works with atmost1 = TRUE
  expect_no_error(
    withr::with_preserve_seed(
      df1 <- ztdraw_intensity(
        lambda = l, t_min = 5, t_max = 10, step_majorizer_vector = lambda_maj, atmost1 = TRUE
      )
    )
  )
  check_ppp_sample_validity(times = df1, t_min = 5, t_max = 10, atmost1 = TRUE, atleast1 = TRUE)
})

#' Special case: Simulate from a non homogeneous Poisson Point Process (NHPPP) from
#'    (t_min, t_max) with log-linear intensity function (inversion method)
#'
#' @description  Sample NHPPP times from an log linear intensity function
#' using the inversion method, optionally using an `rstream`
#' generator
#'
#' @param intercept (double) the intercept in the exponent
#' @param slope (double) the slope in the exponent
#' @param t_min (double) lower bound of the time interval
#' @param t_max (double) upper bound of the time interval
#' @param atmost1 boolean, draw at most 1 event time
#'
#' @return a vector of event times (t_); if no events realize,
#'         a vector of length 0
#' @export
#'
#' @examples
#' x <- draw_sc_loglinear(intercept = 0, slope = 0.2, t_min = 0, t_max = 10)
#'
draw_sc_loglinear <- function(intercept,
                              slope,
                              t_min,
                              t_max,
                              atmost1 = FALSE) {
  if (slope == 0) {
    return(ppp2(rate = exp(intercept), t_min = t_min, t_max = t_max, atmost1 = atmost1))
  }
  return(
    draw_cumulative_intensity_inversion(
      Lambda = function(t) Lambda_exp_form(t, intercept = intercept, slope = slope, t0 = t_min),
      Lambda_inv = function(z) Lambda_inv_exp_form(z, intercept = intercept, slope = slope, t0 = t_min),
      t_min = t_min,
      t_max = t_max,
      atmost1 = atmost1
    )
  )
}

#' Simulate specific number of points from a homogeneous Poisson Point Process over (t_min, t_max]
#'
#' @description
#'   `r lifecycle::badge("deprecated")`
#' Use `ppp_exactly_n` instead.
#' @param size (int) the number of points to be simulated
#' @param range_t (vector, double) min and max of the time interval
#' @param rng_stream an `rstream` object
#'
#' @return a vector of event times of size `size`
#' @export
#' @keywords internal
#' @examples
#' x <- ppp_n(size = 10, range_t = c(0, 10))
ppp_n <- function(size, range_t = c(0, 10), rng_stream = NULL) {
  lifecycle::deprecate_warn(when = "0.5.0", what = "ppp_n()", with = "ppp_exactly_n()")
  U <- rng_stream_runif(size = size, minimum = range_t[1], maximum = range_t[2], rng_stream = rng_stream)
  if (size == 1) {
    return(U)
  }
  return(U[order(U, method = "shell")])
}

#' Generic function for simulating from NHPPPs given the intensity function
#' or the cumulative intensity function.
#'
#' @description
#' This is a wrapper to the package's specific functions, and thus somewhat slower.
#' For time-intensive simulations prefer one of the specific functions.
#'
#' @param Lambda (function, double vector) the integrated (cumulative) rate of the NHPPP
#' @param Lambda_inv (function, double vector) the inverse of `Lambda()'
#' @param lambda (function) the instantaneous rate
#' @param line_majorizer_intercept The intercept `alpha` of the [log]linear majorizer function: `alpha + beta * t` or `exp(alpha + beta * t)`
#' @param line_majorizer_slope The slope `beta` of the [log]linear majorizer function: `alpha + beta * t` or `exp(alpha + beta * t)`
#' @param line_majorizer_is_loglinear (boolean) if `TRUE` the majorizer is loglinear `exp(alpha + beta * t)`; if `FALSE` it is a linear function
#' @param step_majorizer_vector (vector, double) `K` constant majorizing rates, one per interval; all intervals are of equal length (regular)
#' @param t_min (double) the lower bound of the interval
#' @param t_max (double) the upper bound of the interval
#' @param atmost1 boolean, draw at most 1 event time
#' @param atleast1 boolean, draw at least 1 event time in interval
#'
#' @return a vector of event times
#' @export
draw <- function(
    Lambda = NULL,
    Lambda_inv = NULL,
    lambda = NULL,
    line_majorizer_intercept = NULL,
    line_majorizer_slope = NULL,
    line_majorizer_is_loglinear = FALSE,
    step_majorizer_vector = NULL,
    t_min = NULL,
    t_max = NULL,
    atmost1 = FALSE,
    atleast1 = FALSE) {
  if (!is.null(Lambda)) {
    if (atleast1) {
      func <- ztdraw_cumulative_intensity
    } else {
      func <- draw_cumulative_intensity
    }
    return(func(
      Lambda = Lambda,
      Lambda_inv = Lambda_inv,
      t_min = t_min,
      t_max = t_max,
      atmost1 = atmost1
    ))
  }

  if (!is.null(lambda)) {
    if (atleast1) {
      func <- ztdraw_intensity
    } else {
      func <- draw_intensity
    }
    return(func(
      lambda = lambda,
      line_majorizer_intercept = line_majorizer_intercept,
      line_majorizer_slope = line_majorizer_slope,
      line_majorizer_is_loglinear = line_majorizer_is_loglinear,
      step_majorizer_vector = step_majorizer_vector,
      t_min = t_min,
      t_max = t_max,
      atmost1 = atmost1
    ))
  }
}

test_that("draw() works with lambda option", {
  set.seed(123)
  l <- function(t) {
    return(2)
  }
  lmaj_intercept <- 2.2
  lmaj_slope <- 0

  expect_no_error(withr::with_preserve_seed(df <- draw(lambda = l, line_majorizer_intercept = lmaj_intercept, line_majorizer_slope = lmaj_slope, t_min = 0, t_max = 10)))
  check_ppp_sample_validity(times = df, t_min = 0, t_max = 10)

  expect_no_error(withr::with_preserve_seed(df <- draw(lambda = l, line_majorizer_intercept = lmaj_intercept, line_majorizer_slope = lmaj_slope, t_min = 0, t_max = 10, atleast1 = TRUE)))
  check_ppp_sample_validity(times = df, t_min = 0, t_max = 10, atleast1 = TRUE)
})



test_that("draw() works with Lambda option", {
  set.seed(123)
  L <- function(t) {
    return(2 * t)
  }
  Li <- function(z) {
    return(z / 2)
  }

  expect_no_error(withr::with_preserve_seed(df <- draw(Lambda = L, Lambda_inv = Li, t_min = 0, t_max = 10)))
  check_ppp_sample_validity(times = df, t_min = 0, t_max = 10)

  expect_no_error(withr::with_preserve_seed(df <- draw(Lambda = L, Lambda_inv = Li, t_min = 0, t_max = 10, atleast1 = TRUE)))
  check_ppp_sample_validity(times = df, t_min = 0, t_max = 10, atleast1 = TRUE)
})

#' Simulate from a non homogeneous Poisson Point Process (NHPPP) from
#'    (t_min, t_max) (order statistics method)
#'
#' @description  Sample NHPPP times using the order statistics method,
#' optionally using an `rstream` generator object.
#' @param Lambda (function, double vector) a continuous increasing R to R map
#'               which is the integrated rate of the NHPPP
#' @param Lambda_inv (function, double vector) the inverse of `Lambda()`
#' @param t_min (double) the lower bound of the time interval
#' @param t_max (double) the upper bound of the time interval
#' @param atmost1 boolean, draw at most 1 event time
#'
#' @return a vector of event times (t_); if no events realize,
#'         a vector of length 0
#' @keywords internal
draw_cumulative_intensity_orderstats <- function(Lambda,
                                                 Lambda_inv,
                                                 t_min,
                                                 t_max,
                                                 atmost1 = FALSE) {
  L_min <- Lambda(t_min)
  L_max <- Lambda(t_max)
  n <- stats::rpois(n = 1, lambda = L_max - L_min)
  tmp_u <- ppp_exactly_n(n = n, t_min = L_min, t_max = L_max)
  if (atmost1 == TRUE) {
    tmp_u <- tmp_u[1]
  }
  return(Lambda_inv(tmp_u))
}

#' Simulate from a non homogeneous Poisson Point Process (NHPPP) from
#'    (t0, t_max) (thinning method)
#'
#' @description Sample NHPPP times using the thinning method
#'
#' @param lambda (function) the instantaneous rate of the NHPPP.
#' @param majorizer_intercept (double) the intercept (`alpha`) of the [log]linear majorizer function.
#' @param majorizer_slope (double) the slope (`beta') of the [log]linear majorizer function.
#' @param t_min (double) the lower bound of the time interval.
#' @param t_max (double) the upper bound of the time interval.
#' @param majorizer_is_loglinear (boolean) if `TRUE` the majorizer is loglinear `exp(alpha + beta * t)`
#' @param atmost1 boolean, draw at most 1 event time
#'
#' @return a vector of event times (t_); if no events realize,
#'         a vector of length 0
#' @keywords internal
draw_intensity_line <- function(lambda,
                                majorizer_intercept,
                                majorizer_slope,
                                t_min,
                                t_max,
                                majorizer_is_loglinear = FALSE,
                                atmost1 = FALSE) {
  if (isTRUE(majorizer_is_loglinear)) {
    nhppp_t <- draw_sc_loglinear
    link <- exp
  } else {
    nhppp_t <- draw_sc_linear
    link <- identity
  }

  candidate_times <- nhppp_t(intercept = majorizer_intercept, slope = majorizer_slope, t_min = t_min, t_max = t_max, atmost1 = FALSE)
  num_candidates <- length(candidate_times)
  if (num_candidates == 0) {
    return(candidate_times)
  }
  u <- stats::runif(n = num_candidates, min = 0, max = 1)
  acceptance_prob <- lambda(candidate_times) / link(majorizer_intercept + majorizer_slope * candidate_times)
  if (!all(acceptance_prob <= 1 + 10^-6)) {
    stop("lambda > lambda_maj\n")
  }
  if (atmost1) {
    t <- candidate_times[u < acceptance_prob][1]
    if (is.na(t)) {
      t <- numeric(0)
    }
  } else {
    t <- candidate_times[u < acceptance_prob]
  }
  return(t)
}

#' Vectorized sampling from a zero-truncated non homogeneous Poisson Point Process (NHPPP) from
#'    an interval (thinning method) with piecewise constant_majorizers (R)
#'
#' @description
#' Vectorized sampling from a zero-truncated non homogeneous Poisson Point Process (NHPPP) from
#'    an interval (thinning method) with piecewise constant_majorizers.
#'    The majorizers are step functions over equal-length time intevals.
#'
#' @param lambda (function) intensity function, vectorized
#' @param lambda_args (list) optional named list of arguments to pass to `lambda`.
#'        If you have arguments for `lambda` that vary by draw, they should be passed as
#'        a data.table named `vector_arguments`.
#' @param Lambda_maj_matrix (matrix) integrated intensity rates at the end of each interval
#' @param lambda_maj_matrix (matrix) intensity rates, one per interval
#' @param rate_matrix_t_min (scalar | vector | column matrix) is the lower bound
#'        of the time interval for each row of (Lambda|lambda)_maj_matrix.
#'        The length of this argument is the number of point processes that should be drawn.
#' @param rate_matrix_t_max (scalar | vector | column matrix) the upper bound
#'        of the time interval for each row of (Lambda|lambda)_maj_matrix.
#'        The length of this argument is the number of point processes that should be drawn.
#' @param t_min (scalar | vector | column matrix) is the lower bound
#'        of a subinterval of (rate_matrix_t_min, rate_matrix_t_max]. If set,
#'        times are sampled from the subinterval.
#'        If omitted, it is equivalent to `rate_matrix_t_min`.
#' @param t_max (scalar | vector | column matrix) is the upper bound
#'        of a subinterval of (rate_matrix_t_min, rate_matrix_t_max]. If set,
#'        times are sampled from the subinterval.
#'        If omitted, it is equivalent to `rate_matrix_t_max`.
#' @param atmost1 boolean, draw at most 1 event time
#' @keywords internal
vztdraw_intensity_step_regular <- function(
    lambda = NULL,
    lambda_args = NULL,
    Lambda_maj_matrix = NULL,
    lambda_maj_matrix = NULL,
    rate_matrix_t_min = NULL,
    rate_matrix_t_max = NULL,
    t_min = NULL,
    t_max = NULL,
    atmost1 = FALSE,
    ...) {

  #browser()
  if (!is.null(lambda_maj_matrix) && is.null(Lambda_maj_matrix)) {
    rate <- lambda_maj_matrix
  } else if (is.null(lambda_maj_matrix) && !is.null(Lambda_maj_matrix)) {
    rate <- Lambda_maj_matrix
  } else {
    stop("lambda_maj_matrix and Lambda_maj_matrix cannot both be `NULL`")
  }

  range_t <- cbind(as.vector(rate_matrix_t_min), as.vector(rate_matrix_t_max))
  if (nrow(range_t) > 1 && nrow(range_t) != nrow(rate)) {
    stop("The (rows of) [Lambda|lambda]_maj_matrix and (length of) [rate_matrix_t_min|rate_matrix_t_max] imply different numbers of point processes to be sampled.")
  }
  if (nrow(range_t) == 1 && nrow(rate) != 1) {
    range_t <- range_t[rep(1, nrow(rate)), ]
  }

  if (is.null(t_min) && is.null(t_max)) {
    subinterval <- range_t
  } else {
    if (is.null(t_min)) t_min <- range_t[, 1, drop = FALSE]
    if (is.null(t_max)) t_max <- range_t[, 2, drop = FALSE]
    subinterval <- cbind(as.vector(t_min), as.vector(t_max))
    if (nrow(subinterval) > 1 && nrow(subinterval) != nrow(rate)) {
      stop("The (rows of) [Lambda|lambda]_maj_matrix and (length of) [t_min|t_max] imply different numbers of point processes to be sampled.")
    }
    if (nrow(subinterval) == 1 && nrow(rate) != 1) {
      subinterval <- subinterval[rep(1, nrow(rate)), ]
    }
    stopifnot(all(subinterval[, 1] >= range_t[, 1]), all(subinterval[, 2] <= range_t[, 2]))
  }

  Z <- vdraw_intensity_step_regular_forcezt(
    lambda = lambda,
    lambda_args = lambda_args,
    Lambda_maj_matrix = Lambda_maj_matrix,
    lambda_maj_matrix = lambda_maj_matrix,
    rate_matrix_t_min = rate_matrix_t_min,
    rate_matrix_t_max = rate_matrix_t_max,
    t_min = t_min,
    t_max = t_max,
    atmost1 = atmost1,
    force_zt_majorizer = TRUE
  )

  #browser()
  has_no_times <- is.na(Z[, 1])
  max_events <- ncol(Z)

  if (sum(has_no_times) > 0 && !is.null(lambda_args$vector_arguments)) {
    stopifnot(data.table::is.data.table(lambda_args$vector_arguments))
    original_vector_arguments <- lambda_args$vector_arguments
  }


  while (sum(has_no_times) > 0) {
    if(!is.null(lambda_args$vector_arguments)) {
      lambda_args$vector_arguments <- original_vector_arguments[which(has_no_times==TRUE), , drop = FALSE]
    }
    Z_add <- vdraw_intensity_step_regular_forcezt(
      lambda = lambda,
      lambda_args = lambda_args,
      Lambda_maj_matrix = Lambda_maj_matrix[has_no_times, , drop = FALSE],
      lambda_maj_matrix = lambda_maj_matrix[has_no_times, , drop = FALSE],
      rate_matrix_t_min = range_t[has_no_times, 1, drop = FALSE],
      rate_matrix_t_max = range_t[has_no_times, 2, drop = FALSE],
      t_min = subinterval[has_no_times, 1, drop = FALSE],
      t_max = subinterval[has_no_times, 2, drop = FALSE],
      atmost1 = atmost1,
      force_zt_majorizer = TRUE
    ) # force_zt_majorizer)

    diff_cols <- ncol(Z_add) - ncol(Z)
    if (diff_cols > 0) {
      Z <- matrix(
        c(as.vector(Z), rep(NA, times = abs(diff_cols) * nrow(Z))),
        ncol = ncol(Z_add), nrow = nrow(Z), byrow = FALSE
      )
    }
    if (diff_cols < 0) {
      Z_add <- matrix(
        c(as.vector(Z_add), rep(NA, times = abs(diff_cols) * nrow(Z_add))),
        ncol = ncol(Z), nrow = nrow(Z_add), byrow = FALSE
      )
    }

    Z[has_no_times, ] <- Z_add
    has_no_times <- is.na(Z[, 1])
  }
  return(Z)
}

#' Vectorized sampling from a non homogeneous Poisson Point Process (NHPPP) from
#'    an interval (thinning method) with piecewise constant_majorizers (C++)
#'
#' @description
#' Vectorized sampling from a non homogeneous Poisson Point Process (NHPPP) from
#'    an interval (thinning method) with piecewise constant_majorizers.
#'    The majorizers are step functions over equal-length time intevals.
#'
#' @param lambda (function) intensity function, vectorized
#' @param lambda_args (list) optional named list of arguments to pass to `lambda`.
#'        If you have arguments for `lambda` that vary by draw, they should be passed as
#'        a data.table named `vector_arguments`.
#' @param Lambda_maj_matrix (matrix) integrated intensity rates at the end of each interval
#' @param lambda_maj_matrix (matrix) intensity rates, one per interval
#' @param rate_matrix_t_min (scalar | vector | column matrix) is the lower bound
#'        of the time interval for each row of (Lambda|lambda)_maj_matrix.
#'        The length of this argument is the number of point processes that should be drawn.
#' @param rate_matrix_t_max (scalar | vector | column matrix) the upper bound
#'        of the time interval for each row of (Lambda|lambda)_maj_matrix.
#'        The length of this argument is the number of point processes that should be drawn.
#' @param t_min (scalar | vector | column matrix) is the lower bound
#'        of a subinterval of (rate_matrix_t_min, rate_matrix_t_max]. If set,
#'        times are sampled from the subinterval.
#'        If omitted, it is equivalent to `rate_matrix_t_min`.
#' @param t_max (scalar | vector | column matrix) is the upper bound
#'        of a subinterval of (rate_matrix_t_min, rate_matrix_t_max]. If set,
#'        times are sampled from the subinterval.
#'        If omitted, it is equivalent to `rate_matrix_t_max`.
#' @param tol (scalar, double) tolerance for the number of events
#' @param atmost1 boolean, draw at most 1 event time
#' @param atleast1 boolean, draw at least 1 event time
#' @param atmostB If not NULL, draw at most B (B>0) event times. NULL means ignore.
#'
#'
#' @return a matrix of event times (columns) per draw (rows)
#'         NAs are structural empty spots
#'
#' @examples
#' x <- vdraw_intensity(
#'   lambda = function(x, ...) 0.1 * x,
#'   lambda_maj_matrix = matrix(rep(1, 5), nrow = 1),
#'   rate_matrix_t_min = 1,
#'   rate_matrix_t_max = 5
#' )
#' @export


vdraw_intensity <- function(
    lambda = NULL,
    lambda_args = NULL,
    Lambda_maj_matrix = NULL,
    lambda_maj_matrix = NULL,
    rate_matrix_t_min = NULL,
    rate_matrix_t_max = NULL,
    t_min = NULL,
    t_max = NULL,
    tol = 10^-6,
    atmost1 = FALSE,
    atleast1 = FALSE,
    atmostB = NULL) {
  if (isFALSE(atleast1)) {
    return(
      vdraw_intensity_step_regular_cpp(
        lambda = lambda,
        lambda_args = lambda_args,
        Lambda_maj_matrix = Lambda_maj_matrix,
        lambda_maj_matrix = lambda_maj_matrix,
        rate_matrix_t_min = rate_matrix_t_min,
        rate_matrix_t_max = rate_matrix_t_max,
        t_min = t_min,
        t_max = t_max,
        tol = tol,
        atmost1 = atmost1,
        atmostB = atmostB
      )
    )
  }

  if (isTRUE(atleast1)) {
    return(
      vztdraw_intensity_step_regular(
        lambda = lambda,
        lambda_args = lambda_args,
        Lambda_maj_matrix = Lambda_maj_matrix,
        lambda_maj_matrix = lambda_maj_matrix,
        rate_matrix_t_min = rate_matrix_t_min,
        rate_matrix_t_max = rate_matrix_t_max,
        t_min = t_min,
        t_max = t_max,
        tol = tol,
        atmost1 = atmost1
      )
    )
  }
}

test_that("ztdraw_sc_linear() works", {
  withr::with_seed(12345, df1 <- ztdraw_sc_linear(intercept = 2, slope = 0, t_min = 1, t_max = 10))
  check_ppp_sample_validity(times = df1, t_min = 1, t_max = 10, atleast1 = TRUE)

  withr::with_seed(12345, df1 <- ztdraw_sc_linear(intercept = 0, slope = 1, t_min = 1, t_max = 10))
  check_ppp_sample_validity(times = df1, t_min = 1, t_max = 10, atleast1 = TRUE)

  withr::with_seed(12345, df1 <- ztdraw_sc_linear(intercept = 10, slope = -2, t_min = 1, t_max = 10))
  check_ppp_sample_validity(times = df1, t_min = 1, t_max = 10, atleast1 = TRUE)

  withr::with_seed(12345, df1 <- ztdraw_sc_linear(intercept = 10, slope = -2, t_min = 1, t_max = 10, atmost1 = TRUE))
  check_ppp_sample_validity(times = df1, t_min = 1, t_max = 10, atmost1 = TRUE, atleast1 = TRUE)
})

test_that("ztppp() arguments work", {
  set.seed(123)
  expect_no_error(df0 <- ztppp(rate = 1, t_min = 0, t_max = 10, atmost1 = TRUE))
  check_ppp_sample_validity(times = df0, t_min = 0, t_max = 10, atleast1 = TRUE, atmost1 = TRUE)

  expect_no_error(df1 <- ztppp(rate = 1, t_min = 0, t_max = 10))
  check_ppp_sample_validity(times = df1, t_min = 0, t_max = 10, atleast1 = TRUE)
})

test_that("draw_intensity() works", {
  set.seed(123)
  l <- function(t) {
    return(2)
  }

  expect_no_error(withr::with_preserve_seed(df <- draw_intensity(lambda = l, line_majorizer_intercept = 2.2, line_majorizer_slope = 0, t_min = 0, t_max = 10)))
  check_ppp_sample_validity(times = df, t_min = 0, t_max = 10)

  # works when range_t[1]>0
  expect_no_error(withr::with_preserve_seed(df <- draw_intensity(lambda = l, line_majorizer_intercept = 2.2, line_majorizer_slope = 0, t_min = 5, t_max = 10)))
  check_ppp_sample_validity(times = df, t_min = 5, t_max = 10)

  # works with atmost1 = TRUE
  expect_no_error(
    withr::with_preserve_seed(
      df1 <- draw_intensity(
        lambda = l,
        line_majorizer_intercept = 2.2,
        line_majorizer_slope = 0,
        t_min = 5,
        t_max = 10,
        atmost1 = TRUE
      )
    )
  )
  check_ppp_sample_validity(times = df1, t_min = 5, t_max = 10, atmost1 = TRUE)
})


test_that("draw_intensity() works with linear majorization function", {
  set.seed(123)
  l <- function(t) {
    return(2)
  }
  S <- methods::new("rstream.mrg32k3a")

  # works with zero slope majorization function
  expect_no_error(withr::with_preserve_seed(
    df <- draw_intensity(
      lambda = l,
      line_majorizer_intercept = 2.2,
      line_majorizer_slope = 0,
      t_min = 5,
      t_max = 10
    )
  ))
  check_ppp_sample_validity(times = df, t_min = 5, t_max = 10)


  # works with negative slope majorization function
  expect_no_error(withr::with_preserve_seed(
    df <- draw_intensity(
      lambda = l,
      line_majorizer_intercept = 50,
      line_majorizer_slope = -1,
      t_min = 5,
      t_max = 10
    )
  ))
  check_ppp_sample_validity(times = df, t_min = 5, t_max = 10)

  # works with positive slope majorization function
  expect_no_error(withr::with_preserve_seed(
    df <- draw_intensity(
      lambda = l,
      line_majorizer_intercept = 2,
      line_majorizer_slope = 0.1,
      t_min = 5,
      t_max = 10
    )
  ))
  check_ppp_sample_validity(times = df, t_min = 5, t_max = 10)

  # fails when majorization function is below l
  # this is only when you sample at that point
  expect_error(withr::with_preserve_seed(
    df <- draw_intensity(
      lambda = l,
      line_majorizer_intercept = 2.2,
      line_majorizer_slope = -0.1,
      t_min = 5,
      t_max = 10
    )
  ))
  check_ppp_sample_validity(times = df, t_min = 5, t_max = 10)
})

---
title: "Sampling from Gompertz processes"
author: "TA Trikalinos" 
date: "2025-08-20"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Sampling Gompertz times}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
  % \VignetteDepends{tictoc, truncnorm}
---

```{r, include = FALSE}
set.seed(20250820)
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
K <- 10^5
```

## Simulation description  

Assume a population of $K = `r K`$ individuals indexed by $k \in [K] := \{ 1, \dots, K\}$. We will sample events from Gompertz processes over the age (time) interval $[T_{k0}, T_{k1})$. For example, $T_{k0}$ may be the age in years when person $k$ enters the simulation, and $T_{k1}$ the age in years when that person exits the simulation (e.g., dies from an 'other cause'). (In practice, $T_{k1}$ would be obtained by separate point process.)  Only some people will develop clinical cancer over their simulated lifetime. 

To fix a simulation scenario, let $T_{k0} = 20$ for all $k$ and $T_{k1} \sim U(60 , 100)$, where $U()$ is the uniform distribution.  


### Setup 
We will use `data.table`s -- but the same analysis should be obvious in base `R`. 
We will use functions from three more packages, without loading them:
The `tictoc()` package will be used for a simple time comparison between
the different ways one can simulate this problem. The `stats` package is used
to generate normally and uniformly distributed samples. 

```{r setup}
library(data.table)
library(nhppp)
```

Setup `pop`, the population `data.table`. The person specific values T_{k0}$ and $T_{k1}$ are variables in `pop`.
  
```{r population}
pop <- setDT(
  list(
    id = 1:K,
    T0 = rep(20, K),
    T1 = stats::runif(n = K, min = 60, max = 100)
  )
)
setindex(pop, id)
pop
```

## The Gompertz Process functions  

The Gompertz intensity function is 

$\lambda_k(t) = a_k b_k \ e^{b_k t}$, 

where $t$ is age in years. The parameters $a_k, \b_k$ are random over the individuals in the population with 
$a_k \sim U(0.0045, 0.0055)$, and 
$b_k \sim U(0.085, 0.095)$.

Add these values as person-level parameters in the dataset: 

``` {r} 
pop[, `:=`(
  a_gompertz = stats::runif(n = K, min = 0.0045, max = 0.0055),
  b_gompertz = stats::runif(n = K, min = 0.085, max = 0.095)
  )]
```


A cumulative intensity function is 

$\Lambda_k(t) = a_k (e^{b_k t} - 1)$, 

with the integration constant was arbitrarily set so that $\Lambda(0) = 0$. The inverse cumulative intensity function is 

$\Lambda^{-1}(z) = b_k^{-1} \ \log \big (\frac{z}{a_k} -1 \big) $.

#### Vectorized specification of the Gompertz $\lambda()$, $\Lambda()$, and $\Lambda^{-1}()$

Define vectorized forms of the Gompertz $\lambda()$, $\Lambda()$, and $\Lambda^{-1}()$ functions that take as 
default the values in `pop`.

```{r function-lambda-gompertz} 
l_gompertz <- function(t, a = pop$a_gompertz, b = pop$b_gompertz, ...) {
  a * b * exp(b * t)
}

L_gompertz <- function(t, a = pop$a_gompertz, b = pop$b_gompertz, ...) {
  a * exp(b * t) - a
}

Li_gompertz <- function(z, a = pop$a_gompertz, b = pop$b_gompertz, ...) {
  log(z/a +1) / b
}
 
```


### Method 1: Vectorized sampling using only $\lambda()$ 

When you only know the intensity function $\lambda$, `nhppp` employs a thinning 
algorithm. 

One of the items needed for the thinning algorithm is a piecewise constant
majorizer function $\lambda_*$ such that: $\lambda_*(t) >= \lambda(t)$ for 
all $t$ of interest.
 
The `nhppp::vdraw_intensity` function assumes that 
you will provide the majorizer function as a matrix (`lambda_maj_matrix`). 
To create this matrix, split the simulation time (here, from age 40 to age 100) 
in $M$ equal-length intervals. 
For person $k$ and interval $m$, the element `lambda_maj_matrix[k, m]`
records a supremum of $\lambda_k$ over the $m$-th interval. 
Any supremum will do -- but the algorithm is most 
efficient when you give it the least upper bound  -- practically, the maximum 
of $\lambda(t)$ over all $t$ in the interval. For monotone intensity functions, 
such as the Gompertz, the maximum is at one of the interval's bounds. 
It will be at the left bound, if $\lambda$ is decreasing, and at the right bound, 
if $\lambda$ is increasing. 

There is a helper function in `nhppp` that generates the majorizer matrix 
automatically for monotone (and possibly discontinuous) functions and 
for nonmonotone continuous Lipschitz functions (functions whose maximum slope is 
bounded). Even if your case is more complex, you should be able to find a 
supremum that works. 

This code samples in a vectorized fashion when you know only $\lambda()$. It creates
a majorizer matrix over $M=5$ intervals (we chose this arbitrarily -- not trying to be fast). To let the software know which times 
correspond to each of the $M$ intervals it suffices to specify a start and stop 
time for each row of the majorizer matrix with the `rate_matrix_t_min` and `rate_matrix_t_max`
options. The sampling intervals $[T_{k0}, T_{k1})$for each simulated 
person are a subset of the interval for which the majorizer matrix is defined, 
and are specified with the `t_min` and `t_max` options. 
(The `atmostB` option can be useful to speed up the sampling and minimize memory 
needs when one is interested in the first event only. The smaller the value, the faster 
the algorithm but you have to check that you have not specified it to be too small.
In this example, `atmostB = 5` is fine -- it returns exact solutions; but we 
have checked it [not shown].  If you do not want to mess with it, 
do not use the option. The function may be already fast enough for your needs). 

```{r method-1}
tictoc::tic("Method 1 (vectorized, thinning)")
M <- 5
break_points <- seq.int(from = 20, to = 100, length.out = M + 1)
breaks_mat <- matrix(rep(break_points, each = K), nrow = K)

lmaj_mat <- nhppp::get_step_majorizer(
  fun = l_gompertz,
  breaks = breaks_mat,
  is_monotone = TRUE
)

pop[
  ,
  t_thinning := nhppp::vdraw_intensity(
    lambda = l_gompertz,
    lambda_maj_matrix = lmaj_mat,
    rate_matrix_t_min = 20,
    rate_matrix_t_max = 100,
    t_min = pop$T0,
    t_max = pop$T1,
    atmost1 = TRUE,
    atmostB = 10
  )
]
tictoc::toc(log = TRUE) # timer end
```

### Method 2: Vectorized sampling using $\Lambda()$ and $\Lambda^{-1}()$

The most efficient sampling is possible when one knows $\Lambda()$ and $\Lambda^{-1}()$. 
The `nhppp` package can sample in this case using the `vdraw_cumulative_intensity()`
function. Here `range_t` is a matrix with information on each person's $[T_{k0}, T_{k1})$. 

```{r method-2}
tictoc::tic("Method 2 (vectorized, inversion)")
pop[
  ,
  t_inversion := nhppp::vdraw_cumulative_intensity(
    Lambda = L_gompertz,
    Lambda_inv = Li_gompertz,
    t_min = pop$T0,
    t_max = pop$T1,
    atmost1 = TRUE
  )
]
tictoc::toc(log = TRUE) # timer end
```

### Comparisons 

#### Simulation time-costs
The simulation time-costs that you see in this document depend on the machine 
that rendered it. If you read this online, this machine is probably some 
virtual server with minimal resources. If you installed the package locally, 
it is probably the machine you are using to run `R`. 

1. `r tictoc::tic.log()[[1]]`. This is the slower approach -- but still not
 bad for $`r K`$ samples! It uses the thinning algorithm which is very flexible -- it can accommodate very complex time varying intensity 
functions. You almost always know $\lambda$ and can get its majorizer $\lambda_*$ 
easily and fast.  

2. `r tictoc::tic.log()[[2]]`. This approach is many times faster that the previous one, 
but requires implementations for $\Lambda$ and $\Lambda^{-1}$.

### Simulated times 

Both methods sample correctly from the specified Gompertz process. There 
is no approximation at play.

The QQ plots compare the simulated times with the two methods. The agreement 
is excellent over this population of size $K = `r K`$. As $K$ increases 
the agreement remains excellent (not shown here - try it for yourself). 
The paper in the bibliography includes in-depth comparisons. 
A set of QQ plots should suffice here.

```{r qq-plot, fig.alt="QQ plot comparing simulated times with the two methods. The QQ plot indicates excellent agreement."}
qqplot(pop$t_thinning, pop$t_inversion)
```

## Simulating all event trajectories

In the above, we simulated the time to first event. Let's simulate all times that realize from that process in the interval of interest. You can use either method; we demonstrate with the faster one. 

``` {r all_events}
Z <- nhppp::vdraw_cumulative_intensity(
    Lambda = L_gompertz,
    Lambda_inv = Li_gompertz,
    t_min = pop$T0,
    t_max = pop$T1,
    atmost1 = FALSE
  )

dim(Z) # a lot of events!
```

There is at least one person among the `r K` simulated persons with `r dim(Z)[2]` total events in the interval of interest. 


## Demonstrating that we simulate from the correct intensity function 

Let's fix the parameter values for all $K$ people and do a histogram of the simulated times. They match the shape of the intensity function over the interval $[T_0, T_1)$, scaled to unit area, i.e. $\frac{\lambda(t)}{\big(\Lambda(T_1)-\Lambda(T_0) \big)}$. Run more samples to convince yourself -- or also calculate the Wasserstein distance of the empirical and theoretical distribution, as described in the numerical analyses in the `nhppp` paper. 

```{r histogram}
if(nchar(system.file(package = 'ggplot2'))>0) {
  pop2 <- setDT(list(id = 1:1000)) 
  pop2[, `:=`(
    T0 = 20, 
    T1 = 100, 
    a = 0.005, 
    b = 0.090
  )]
  setindex(pop2, id)
  pop2
  
  ## re-define functions to use `pop2` by default 
  ##    clunckiness of `nhppp` -- help us improve it!
  l  <- function(t, a = pop2$a, b = pop2$b, ...) a * b * exp(b * t)
  L  <- function(t, a = pop2$a, b = pop2$b, ...) a * exp(b * t) - a
  Li <- function(z, a = pop2$a, b = pop2$b, ...) log(z/a +1) / b
  
  ## get all times (full trajectories)
  Z <- nhppp::vdraw_cumulative_intensity(
      Lambda = L,
      Lambda_inv = Li,
      t_min = pop2$T0,
      t_max = pop2$T1,
      atmost1 = FALSE
    )
  
  ## ignore non-realized times 
  Z <- as.vector(Z)
  Z <- Z[!is.na(Z)]
  
  x <- seq(from = 20, to = 100, length.out = 100)
  ## normalize the hazard rate to have area 1 between 20 and 100
  y <- l(x, a = pop2$a[1], b = pop2$b[1]) / (
        L(100, a = pop2$a[1], b = pop2$b[1]) - 
        L(20, a = pop2$a[1], b = pop2$b[1])
       )
  
  
  library(ggplot2)
  ggplot(xlim = c(0, 110)) +
    geom_histogram(
      aes(x = Z, y = after_stat(density)), 
      bins = 100, 
      fill = "gray", 
      color = "white") +
    geom_line(
      aes(x = x, y = y), 
      color = "red"
    ) +
    labs(title = "Histogram of all simulated Gompertz times", x = "Time", y = "Empirical and theoretical density")
}
```



## Acknowledgments 

Thanks to Fernando Alarid-Escudero for suggesting this vignette. 


## Bibliography

Trikalinos TA, Sereda Y. _nhppp: Simulating Nonhomogeneous Poisson Point Processes in R_. arXiv preprint arXiv:2402.00358. 2024 Feb 1.

Trikalinos TA, Sereda Y. _The nhppp package for simulating non-homogeneous Poisson point processes in R_. PLoS One. 2024 Nov 21;19(11):e0311311.

Marshall AW, Olkin I. "The Gompertz distribution" (p 364) in _Life distributions_  Springer, New York; 2007.

Since the publication of the paper, the syntax and options of the `nhppp` package have
evolved. To reproduce the code in the paper, you have to install the version 
of `nhppp` used in the paper. Alternatively, take a look at the vignettes, 
which are written to work with the current package.  

---
title: "Sampling log-linear times"
author: "TA Trikalinos" 
date: "2024-10-17"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Sampling log-linear times}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
  % \VignetteDepends{tictoc, truncnorm}
---

```{r, include = FALSE}
set.seed(20241017)
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
K <- 10^5
```


## Simulation description  

Assume a population of $K = `r K`$ individuals indexed by $k \in [K] := \{ 1, \dots, K\}$. 
For person $k$, we want to simulate the first occurrence of an event (e.g., emergence of a tumor) 
over the age (time) interval $[T_{k0}, T_{k1})$. For example, $T_{k0}$ may be the age in years when person $k$ enters the simulation, and $T_{k1}$ the age in years when that person dies from non-cancer causes. (In practice, $T_{k1}$ would be obtained by separate point process.)  Only some people will develop clinical cancer over their simulated lifetime. 

To fix a simulation scenario, let $a_k = 40$ for all $k$ and $b_k \sim U(50 , 100)$, where $U()$ is the uniform distribution.  


We use the log-linear intensity function 

$\lambda_k(t) = e^{\alpha_k + \beta_k t}$, 

where $t$ is age in years. The parameters $\alpha_k, \beta_k$ are random over the individuals in the population with 
$\alpha_k \sim N(-4, 0.5)$, and 
$\beta_k \sim N_{0+}(0.03, 0.003)$, 
where $N()$ is a normal distribution and $N_{0+}()$ is a truncated normal distribution with support $[0, \infty)$.


## Overview of sampling methods described here
We will use three methods: 

1. Simulate one person at a time looping over persons. The fastest way to do 
this is to use a special case function in the `nhppp` package that 
samples from log-linear intensity functions in a non-vectorized fashion. Yet 
this will be slowest approach. 

2. Simulate all persons in a vectorized way using only the intensity 
function $\lambda(t)$. When you only know $\lambda(t)$ `nhppp` uses the thinning 
algorithm. It is a very flexible approach because it does not 
require a lot of information -- just $\lambda(t)$. 
This will by much faster than the first option, but not as fast as the third option.  

3. Simulate all persons in a vectorized way using the cumulative 
intensity function $\Lambda(t)$ and its inverse $\Lambda^{-1}(z)$, 
which are defined below. When you have analytic expressions for these objects,
you get the fastest sampling in the `nhppp` package. Then the package can 
use the `inversion` or the `order statistics` 
algorithms which are more efficient. 


## Setup 
We will use `data.table`s -- but the same analysis should be obvious in base `R`. 
We will use functions from three more packages, without loading them:
the `truncnorm` package is only needed for the truncated normal distribution. 
The `tictoc()` package will be used for a simple time comparison between
the different ways one can simulate this problem. The `stats` package is used
to generate normally and uniformly distributed samples. 

```{r setup}
library(data.table)
library(nhppp)
```

Setup `pop`, the population `data.table`. The person specific parameters $\alpha_k, \beta_k, T_{k0}$ and $T_{k1}$ are variables in `pop`.
  
```{r population}
pop <- setDT(
  list(
    id = 1:K,
    alpha = stats::rnorm(n = K, mean = -4, sd = 0.5),
    beta = truncnorm::rtruncnorm(n = K, mean = 0.03, sd = 0.003, a = 0, b = Inf),
    T0 = rep(40, K),
    T1 = stats::runif(n = K, min = 50, max = 100)
  )
)
setindex(pop, id)
pop
```

### Intensity, cumulative intensity, and inverse cumulative intensity functions 

Define some bespoke functions for the different simulation approaches below.
The trick is to define functions so that they work in a vectorized form. 
For an example,  take a look at the intensity function below. 
The other functions have the same bahavior. 

#### Intensity function $\lambda()$

Define a vectorized form of the intensity function. 
```{r function-lambda-0} 
l <- function(t, alpha = pop$alpha, beta = pop$beta, ...) exp(alpha + beta * t)
```

Arguments `alpha` and `beta` can be scalars or vectors (or column matrices). 
Argument `t` can be a scalar, a vector of $K$ ages (times), or a $K \times s$ matrix. See some examples of the behavior of the vectorized function. The ellipses (`...`) allow `l()` to ignore extra arguments without breaking the execution of the script. While we won't do that on purpose, some sampling functions in the `nhppp` package that take function arguments need the ability to pass optional extraneous arguments. 

Scalar arguments, evaluated at `t = 45`
```{r} 
l(t = 45, alpha = -4, beta = 0.03)
```


Scalar arguments, evaluated at `t = 45:50` (vector) or as a column matrix 
The function returns results in the format of the `t` argument. 
``` {r}  
l(t = 45:50, alpha = -4, beta = 0.03)

l(t = matrix(45:50, ncol = 1), alpha = -4, beta = 0.03)
```


Vector arguments, using the first 5 people and evaluating all people at `t = 45`
```{r}
l(t = 45, alpha = pop$alpha[1:5], beta = pop$beta[1:5])
```

Matrix arguments are convenient: 
the rows are people and the columns are the ages (times) -- and they can differ across persons. 
For the first 5 people, we evaluate a different age for each person. 
The results are returned in a matrix format. 
```{r}
l(t = matrix(c(45, 50, 45, 47.4, 30), ncol = 1), alpha = pop$alpha[1:5], beta = pop$beta[1:5])
``` 

For the first 5 people, we evaluate three different ages for each person. We arrange a `t_mat`
matrix of ages (times) with people in the rows and ages (times) for each person in the columns. 
The results are returned in a matrix format. 

```{r}
t_mat <- matrix(c(
  45, 50, 45, 47.4, 30,
  45.1, 50.1, 45.5, 47.8, 38,
  48, 52.7, 60.1, 70.1, 99.9
), ncol = 3, byrow = TRUE)
t_mat

l(t = t_mat, alpha = pop$alpha[1:5], beta = pop$beta[1:5])
```


The defaults for the `alpha` and `beta` arguments of the `l()` functions
are the respective columns of the whole population. We can then evaluate the 
intensity function $\lambda()$ for the whole population passing only the `t`
argument.  

```{r}
t_mat <- matrix(rep(c(45, 50, 55), each = K), ncol = 3, byrow = TRUE)
l(t = t_mat) |> head()
```


#### Cumulative intensity function $\Lambda(t)$

The cumulative intensity function is $\Lambda(a, t) = \int_a^t \lambda(s) \ \textrm{d}s$. 
It does not matter what we choose for the lower limit of integration in the definition of $\Lambda$. The lower limit cancels out in the mathematics of the sampling algorithms.  Thus we are free to 
use any lower limit (say 0) for any antiderivative of $\lambda()$ that is convenient (for any integration constant).

With a slight abuse of notation, define 0 as the lower integration limit and write ${\Lambda(t) := \Lambda(0, t) = \int_0^t \lambda(s) \ \textrm{d}s}$.
For the log-linear intensity in this example, 
$\Lambda(t) = \frac{1}{\beta} (e^{\alpha + \beta t} - e^\alpha)$. 

The vectorized version of $\Lambda$ for our example is 
``` {r Lambda}
L <- function(t, alpha = pop$alpha, beta = pop$beta, ...) {
  (exp(alpha + beta * t) - exp(alpha)) / beta
}
```


#### Inverse cumulative intensity function $\Lambda^{-1}(z)$

By its construction, $\Lambda$ is a strictly positive monotone function in $t$, and thus invertible.
The inverse of the cumulative intensity function $\Lambda^{-1}$ is defined as the function that 
recovers the $t$ when you pass it the $\Lambda(t)$. The definition is that
it satisfies ${\Lambda^{-1} \big ( \Lambda(t) \big ) = t}$. In our example, $\Lambda^{-1}(z) = \big(\log(\beta z + e^\alpha) - \alpha\big)/\beta$, which is easily derived from the formula of $\Lambda(t)$.  

The vectorized implementation of $\Lambda^{-1}$ for our example is 
``` {r Lambda-inv}
Li <- function(z, alpha = pop$alpha, beta = pop$beta, ...) {
  (log(beta * z + exp(alpha)) - alpha) / beta
}
```


## Method 1: non-vectorized sampling with `nhppp::draw_sc_loglinear()` 

The `nhppp` package function `draw_sc_loglinear()` draws times from log-linear 
densities for each person at a time. This is slower than the other methods, 
but can be practical even for sizeable simulations that will be run once (e.g., 
for statistical simulation analyses) or when you develop code.  

It's arguments `intercept`, `slope` have the same name as the parameters 
in our log-linear intensity function $\lambda$. The `t_min`, `t_max` arguments ask for
the bounds of the interval $[T_{k0}, T_{k1})$. The argument `atmost1` asks only 
for the first time. This special case function uses a bespoke inversion algorithm; 
it's as fast as we can do without vectorization. 

```{r method-1}
tictoc::tic("Method 1 (nonvectorized)")
t_nonvec_special_case <- rep(NA, K)
for (k in 1:K) {
  t1 <- nhppp::draw_sc_loglinear(
    intercept = pop$alpha[k],
    slope = pop$beta[k],
    t_min = pop$T0[k],
    t_max = pop$T1[k],
    atmost1 = TRUE
  )
  if (length(t1) != 0) {
    t_nonvec_special_case[k] <- t1
  }
}
tictoc::toc(log = TRUE)
pop[, t_nonvec_special_case := t_nonvec_special_case]
```


## Method 2: Vectorized sampling using only $\lambda()$ 

When you only know the intensity function $\lambda$, `nhppp` employs a thinning 
algorithm. 

One of the items needed for the thinning algorithm is a piecewise constant
majorizer function $\lambda_*$ such that: $\lambda_*(t) >= \lambda(t)$ for 
all $t$ of interest.
 
The `nhppp::vdraw_intensity` function assumes that 
you will provide the majorizer function as a matrix (`lambda_maj_matrix`). 
To create this matrix, split the simulation time (here, from age 40 to age 100) 
in $M$ equal-length intervals. 
For person $k$ and interval $m$, the element `lambda_maj_matrix[k, m]`
records a supremum of $\lambda_k$ over the $m$-th interval. 
Any supremum will do -- but the algorithm is most 
efficient when you give it the least upper bound  -- practically, the maximum 
of $\lambda(t)$ over all $t$ in the interval. For monotone intensity functions, 
such as the function in the example, the maximum is at one of the interval's bounds. 
It will be at the left bound, if $\lambda$ is decreasing, and at the right bound, 
if $\lambda$ is increasing. 

There is a helper function in `nhppp` that generates the majorizer matrix 
automatically for monotone (and possibly discontinuous) functions and 
for nonmonotone continuous Lipschitz functions (functions whose maximum slope is 
bounded). Even if your case is more complex, you should be able to find a 
supremum that works. 

This code samples in a vectorized fashion when you know only $\lambda()$. It creates
a majorizer matrix over $M=5$ intervals. To let the software know which times 
correspond to each of the $M$ intervals it suffices to specify a start and stop 
time for each row of the majorizer matrix with the `rate_matrix_t_min` and `rate_matrix_t_max`
options. The sampling intervals $[T_{k0}, T_{k1})$for each simulated 
person are a subset of the interval for which the majorizer matrix is defined, 
and are specified with the `t_min` and `t_max` options. 
(The `atmostB` option can be useful to speed up the sampling and minimize memory 
needs when one is interested in the first event only. The smaller the value, the faster 
the algorithm but you have to check that you have not specified it to be too small.
In this example, `atmostB = 5` is fine -- it returns exact solutions; but we 
have checked it [not shown].  If you do not want to mess with it, 
do not use the option. The function may be already fast enough for your needs).  

```{r method-2}
tictoc::tic("Method 2 (vectorized, thinning)")
M <- 5
break_points <- seq.int(from = 40, to = 100, length.out = M + 1)
breaks_mat <- matrix(rep(break_points, each = K), nrow = K)

lmaj_mat <- nhppp::get_step_majorizer(
  fun = l,
  breaks = breaks_mat,
  is_monotone = TRUE
)

pop[
  ,
  t_thinning := nhppp::vdraw_intensity(
    lambda = l,
    lambda_maj_matrix = lmaj_mat,
    rate_matrix_t_min = 40,
    rate_matrix_t_max = 100,
    t_min = pop$T0,
    t_max = pop$T1,
    atmost1 = TRUE,
    atmostB = 5
  )
]
tictoc::toc(log = TRUE) # timer end
```


## Method 3: Vectorized sampling using $\Lambda()$ and $\Lambda^{-1}()$

The most efficient sampling is possible when one knows $\Lambda()$ and $\Lambda^{-1}()$. 
The `nhppp` package can sample in this case using the `vdraw_cumulative_intensity()`
function. Here `range_t` is a matrix with information on each person's $[T_{k0}, T_{k1})$. 

```{r method-3}
tictoc::tic("Method 3 (inversion)")
pop[
  ,
  t_inversion := nhppp::vdraw_cumulative_intensity(
    Lambda = L,
    Lambda_inv = Li,
    t_min = pop$T0,
    t_max = pop$T1,
    atmost1 = TRUE
  )
]
tictoc::toc(log = TRUE) # timer end
```

## Comparisons 

### Simulation time-costs
The simulation time-costs that you see in this document depend on the machine 
that rendered it. If you read this online, this machine is probably some 
virtual server with minimal resources. If you installed the package locally, 
it is probably the machine you are using to run `R`. 

1. `r tictoc::tic.log()[[1]]`. This is the slowest approach -- but still not
 bad for $`r K`$ samples!

2. `r tictoc::tic.log()[[2]]`. This approach is many times faster that then first approach. 
It is very flexible -- it can accommodate very complex time varying intensity 
functions. You almost always know $\lambda$ and can get its majorizer $\lambda_*$ 
easily and fast.  

3. `r tictoc::tic.log()[[3]]`. This approach is many times faster that the second one, 
but requires implementations for $\Lambda$ and $\Lambda^{-1}$.
 


### Simulated times 

All three methods sample correctly from the specified log-linear process. There 
is no approximation at play.

The QQ plots compare the simulated times with the three methods. The agreement 
is excellent over this population of size $K = `r K`$. As $K$ increases 
the agreement remains excellent (not shown here - try it for yourself). 
The paper in the bibliography includes in-depth comparisons. 
A set of QQ plots should suffice here.

```{r qq-plots, fig.alt="QQ plots comparing simulated times with the three methods. The QQ plots indicate excellent agreement."}
qqplot(pop$t_nonvec_special_case, pop$t_thinning)
qqplot(pop$t_nonvec_special_case, pop$t_inversion)
```

## Acknowledgments 

Thanks to Carolyn Rutter and Hui Hsuan Chan for providing the numerical example in this vignette. 


## Bibliography

Trikalinos TA, Sereda Y. _nhppp: Simulating Nonhomogeneous Poisson Point Processes in R_. arXiv preprint arXiv:2402.00358. 2024 Feb 1.

Since the publication of the paper, the syntax and options of the `nhppp` package have
evolved. To reproduce the code in the paper, you have to install the version 
of `nhppp` used in the paper. Alternatively, take a look at the vignettes, 
which are written to work with the current package.  

---
title: "A simple discrete event simulation model of a cancer's natural history"
author: "TA Trikalinos, Y Sereda, S Chrysanthopoulou, F Alarid-Escudero" 
date: "2024-10-17"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{A simple discrete event simulation model of a cancer's natural history}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
  % \VignetteDepends{tictoc}
---

```{r, include = FALSE}
set.seed(20241018)
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
K <- 10^5
```


## The simulation world 

We will model a cancer's natural history in a population. We index people by $k \in [K] := \{1, \dots, K \}$. The following assumptions fix a simulation world. 

1. All types of events can be modeleld with nonhomogeneous Poisson point processes (NHPPPs).  

1. Persons are alive and cancer free at 40 years of age. No person will live past 110 years. 
All people can die from causes other than the cancer of interest (hereafter, _death from other causes_). Write $\rho_k(t)$ for the corresponding intensity function. 

1. Some people may be exposed to an environmental toxin, with exposure that 
varies over time. Write $\xi_k(t) \ge 0$ for the exposure function. 
Positive values mean that a person is exposed at that particular instance. 
(The never exposed have zeros throughout their life.) We will assume that 
the exposure to said toxin is a risk factor only for cancer emergence, and 
that the toxin has no cumulative effects -- only the instantaneous exposure 
levels matter. Write $\delta_k$ for the effect of the toxin
instantaneous exposure on developing cancer.

1. Some persons will develop a preclinical cancer with a time-varying intensity function $\lambda_{k}(t) = \underbrace{\lambda_{k0}(t)}_{\textrm{non-risk factor part}} + \underbrace{\delta_k \ \xi_k(t)}_{\textrm{risk factor part}}$. 

1. Some preclinical cancers may progress to clinical cancer with an intensity function $\mu_k(t)$, at which point they are considered diagnosed. 

1. Once people develop preclinical cancer they can die from cancer with intensity 
function $\nu_k(t)$. The cancer death rate does not explicitly 
depend on whether the cancer has been diagnosed or not.  Thus, we have two 
competing causes of death: death due to cancer and due to other causes. 


## Setup 

Load `nhppp` and `data.table`. (If you prefer to not use `data.table`, you 
should be able to implement this example in base `R` with little trouble.)

```{r setup}
library(data.table)
library(nhppp)
```


## Simulation model 

We now fix the mathematical description of the model. We will simulate $K = `r K`$ 
males and females (with equal probability) from the 2015 birth cohort. 

```{r pop}
pop <- data.table(
  id = 1:K,
  birth_cohort = 2015,
  spawn_age = 40,
  max_simulation_age = 110,
  sex = sample(c("male", "female"), K, replace = TRUE)
)

## It would make sense to execute the commented-out code now.
## It generates model parameters used in later stages.
## For expository clarity, we generate each parameter when it is introduced
# pop[, `:=`(
#    param_cancer_emergence_shape = runif(.N, 7, 9),
#    param_cancer_emergence_scale = rnorm(.N, 150, 20),
#    param_toxin_exposure_diff = pmax(0.005, rnorm(.N, 0.01, 0.005)),
#    param_cancer_death_intercept := rnorm(.N, -2, 0.5),
#    param_cancer_death_slope := runif(n= .N, min = 0, max = 0.003),
#    param_clinical_cancer_dx_rate := runif(.N, 0.20, 0.27)
#  )]
```

### Death from other causes 
The death from other causes $\rho_k(t)$
depends on the age ($t$, measured in years), sex (male or female), and birth 
year of person $k$. Function $\rho$ is is a piecewise constant over each year 
of age. It is a `regular' step function (all steps have the same length of one 
year). 

If the cancer is not a major cause of death, then the intensity function for all
cause deaths is a good approximation for the intensity function for death from other causes. The internal dataset `annual_mortality_rates_2015` has all cause mortality data for the 2015 birth cohort. It has the values of the piecewise constant $\rho$ per birth cohort, sex, and age. 
Here is a peek at some columns.

```{r rho}
annual_mortality_rates_2015[
  sex %in% c("male", "female"),
  c(1:5, 111:113)
]
```

When we have a _step_ (piecewise constant) intensity function over _regular_ time intervals 
(here, all one year long), we can use `nhppp`'s `vdraw_sc_step_regular()` function.
We need to specify the following: 

1. Pass the intensities as a matrix (argument `lambda_matrix`); the number of columns 
in the matrix are the number of time intervals in the step function.  

```{r death_other_causes_intensity_matrix}
rhos <- annual_mortality_rates_2015[
  pop,
  on = c("birth_cohort", "sex")
]
setindex(rhos, "id")
rho_matrix <- as.matrix(rhos[, c(paste0("age_", 0:109), "age_110+"),
  with = FALSE
])

rm(list = "rhos") # cleanup
```


2. Give information about how long each time step is, by specifying the 
age bounds `rate_matrix_t_min` and `rate_matrix_t_max` over which the intensity matrix applies; 

2. Optionally, if we want to sample times in a sub-interval of `(rate_matrix_t_min, rate_matrix_t_max]`, we can specify even narrower bounds, `t_min` and `t_max`. If you omit 
`t_min` or `t_max`, the software uses `rate_matrix_t_min` or `rate_matrix_t_max`,
respectively, to specify the sampling interval. 

2. Because no person lives beyond the maximum simulation age of 110 years, 
we need to force the simulation of at least one death event over the 
simulation interval. This means that we are sampling from a zero-truncated NHPPP.
Setting the option `atleast1` to `TRUE` achieves this.

2. We only need to sample the earliest event from this NHPPP. So we set the `atmost1`
option to `TRUE`. 

```{r age_dead_other_causes}
pop[
  ,
  age_dead_from_other_causes :=
    nhppp::vdraw_sc_step_regular(
      lambda_matrix = rho_matrix,
      rate_matrix_t_min = 0,
      rate_matrix_t_max = 110,
      t_min = pop$spawn_age, # 40
      t_max = pop$max_simulation_age, # 110
      atmost1 = TRUE,
      atleast1 = TRUE
    )
]
```

### Environmental toxin exposure histories 

We will generate environmental exposure histories with a phenomenological 
model. We will assume that 

1. People may be exposed to the environmental toxin with probability $p_{start} = 0.20$. 

1. For those who will be exposed, the start age of exposure is $t_{k0} \sim U(12, 35)$, provided that they are still alive. 

1. Among those who are exposed, the probability that their exposure will eventually 
stop is $p_{stop} = 0.60$. 

1. In the pertinent subgroup of persons, the duration of the exposure is $d_k \sim U(1, 35)$, 
if they are still alive.

1. For people with at least some exposure to the toxin, for all times in the exposure window $(t_{k0}, t_{k1}]$ the exposure levels are $\xi_k(t) = \Xi_k \cdot \Big(\frac{1}{2} +\frac{1}{4}\big(\cos(0.5 t) + \cos(0.45 t) \big) \Big)$, where $t$ is a person's age and the amplitude (maximum exposure) $\Xi_k$ has model $\Xi_k \sim U(0.2, 1)$. Otherwise, $\xi_k(t) = 0$.

We now add the per-person parameters for exposure histories in the population data.table. 
For people who will never be exposed we set $\Xi_k$ to zero, 
and their exposure start and stop ages at the `max_simuation_age`. This avoids 
`if ... else` statements, and is still pretty fast. If you run a massive model, 
though, you may want to be smarter about it.  

``` {r xi_params}
pop[, `:=`(
  exposure_start_age = max_simulation_age,
  exposure_stop_age = max_simulation_age,
  maximum_exposure = 0
)][
  ,
  will_start_exposure := runif(.N) < 0.20
][
  will_start_exposure == TRUE,
  will_stop_exposure := runif(.N) < 0.60
][
  will_start_exposure == TRUE,
  exposure_start_age := pmin(runif(.N, 12, 35), age_dead_from_other_causes)
][
  will_stop_exposure == TRUE,
  exposure_stop_age := pmin(
    exposure_start_age + runif(.N, 1, 35),
    age_dead_from_other_causes
  )
][
  will_start_exposure == TRUE,
  maximum_exposure := runif(.N, 1 / 5, 1)
]

# cleanup
pop[, will_start_exposure := NULL][, will_stop_exposure := NULL]
```



We implement $\xi_k(t)$ as a function that is vectorized 
over all its arguments. The arguments `start_age`, `stop_age`, `max_exposure` correspond 
to the variables $t_{k0}, t_{k1}, \Xi$ in the equation above.

```{r xi}
xi <- toxin_exposure <- function(t, max_exposure, start_age, stop_age) {
  (start_age <= t) * (stop_age >= 1) * max_exposure * (1 / 2 + (cos(t / 2) + cos(0.9 * t / 2)) / 4)
}
```

### Emergence of pre-clinical cancer in unexposed and exposed intervals

Assume that the intensity function for cancer emergence in the absence of 
toxin exposure is

 $\lambda_{k0}(t) = \frac{shape_k}{scale_k} \Big(\frac{t}{scale_k}\Big)^{shape_k -1}$, 
 
where $t$ is age in years.

This intensity function generates a Weibull point process (a special case of an NHPPP). The parameters $shape_k$ and $scale_k$ are assumed to vary across people according to the models $shape_k \sim U(7, 9)$ and $scale_k \sim N(150, 20)$, where $U(\cdot)$ and $N(\cdot)$ stand for uniform and normal distributions.  We sample these values
for each person in the population. 

```{r params_cancer_generation}
pop[, `:=`(
  param_cancer_emergence_shape = runif(.N, 7, 9),
  param_cancer_emergence_scale = rnorm(.N, 150, 20)
)]
```


Generating a Weibull point process is easy in `R` using the 
`stats::rweibull()` function. (This would take care of the cancer emergence times 
for people without toxin exposure, but not for people with toxin exposures.)
Accounting for toxin exposure histories, the intensity function for cancer 
emergence is $\lambda_{k}(t) = l_{k0}(t) + \delta_k \xi_k(t)$. 

We will assume that the toxin exposure effect $\delta_k$ is distributed as
$\delta_k \sim N_+\big(2, 0.5\big)$, where $N_+(\cdot)$ is a slab and smear normal 
distribution.

```{r delta_k}
pop[, param_toxin_exposure_diff := pmax(0, rnorm(.N, 0.01, 0.005))]
```

We need to sample from an NHPPP with an intensity function that varies over time as per $\lambda_k(t)$. We will use `nhppp`'s `vdraw_intensity()` function, which needs

1. The intensity function (argument `lambda`) in a vectorized form, so that 
age $t$ is the only needed argument and all other arguments are set by default.

2. A majorizer piecewise constant function, which will be specified as a matrix `lambda_maj_matrix`. 

3. The `rate_matrix_t_min`, `rate_matrix_t_max` arguments that specify the time bounds for the matrix `lambda_maj_matrix`. 

4. `t_min`, and `t_max` arguments, for the subinterval over whchi we will sample 
times. Here, `t_min = 40` -- the spawn age in the simulation, and `t_max` is the 
`age_dead_from_other_causes`. 

Let's implement the above. 

The _intensity function_ is specified as follows. Observe that it is in vectorized form. 

```{r cancer_generation}
lambda <- function(t, P = pop, ...) {
  # non-risk factor part: shape / scale * (t/scale)^(shape - 1)
  (P$param_cancer_emergence_shape / P$param_cancer_emergence_scale) *
    (t / P$param_cancer_emergence_scale)^(P$param_cancer_emergence_shape - 1) +
    # risk factor (toxin exposure) part: delta_k * xi(t)
    P$param_toxin_exposure_diff *
      xi(
        t = t,
        max_exposure = P$maximum_exposure,
        start_age = P$exposure_start_age,
        stop_age = P$exposure_stop_age
      )
}
```

We also need a piecewise constant _majorizer_ function $\lambda_*(t)$. We say that $\lambda_*(t)$ majorizes $\lambda(t)$ if $\lambda_*(t) \ge \lambda(t)$ for all $t$ of interest. The function expects a _regular step_ majorizer function. 
We will create such a function with $M = 10$ equally-spaced intervals
over the whole simulation window. First, generate the endpoints of the 
$M$ intervals. This will be a matrix where the rows correspond to persons
and the columns to the $M + 1 =11$ interval bounds. 
```{r time breaks}
# define interval bounds for the step function, one row per person
M <- 10
time_breaks <- matrix(
  data = rep(x = seq(from = 40, to = 110, length.out = M + 1), each = K),
  byrow = FALSE,
  nrow = K
)

time_breaks[1:3, ]
```
... and now generate the majorizer matrix using `nhppp`'s `get_step_majorizer()` 
function. (The paper in the Bibliography explains how this function works.)  
``` {r lambda star}
lambda_star <- nhppp::get_step_majorizer(
  fun = lambda,
  breaks = time_breaks,
  is_monotone = FALSE,
  K = 1.9 / 4 # This is the maximum slope of xi() -- which you get with some calculus
)

lambda_star[1:3, ]
```

And now, we can sample the cancer generation times, and create a variable 
to identify patients with cancer.  

```{r lambda}
pop[
  ,
  age_cancer_emergence := nhppp::vdraw_intensity(
    lambda = lambda,
    lambda_maj_matrix = lambda_star,
    rate_matrix_t_min = 40,
    rate_matrix_t_max = 110,
    t_min = pop$spawn_age,
    t_max = pmin(pop$age_dead_from_other_causes, 110, na.rm = TRUE),
    atmost1 = TRUE
  )
][
  ,
  with_cancer := !is.na(age_cancer_emergence),
]
```


### Dying from cancer 

People with preclinical cancer may die from cancer causes. We will assume that 
the intensity from cancer deaths is loglinear in time, that is 

$\nu_k = e^{\alpha_k + \beta_k t}$, 

with parameters $\alpha_k \sim N(-3, 0.2)$ and $\beta_k ~ U(0, 0.003)$. 
```{r params_cancer_death}
pop[, param_cancer_death_intercept := rnorm(.N, -3, 0.2)]
pop[, param_cancer_death_slope := runif(.N, 0, 0.003)]
```

We could use the `vdraw_intensity()` function again, since we already know 
the intensity $\nu_k$ and we can easily create a majorizer function for it, as 
we did when we generated cancer emergence times. This would be plenty fast for
our small simulation with $K = `r K`$ and
requires no additional mathematics. 

We can sample even faster if we can analytically obtain the cumulative intensity 
function $N_k(t) = \int_0^t{\nu_k(s) \ \textrm{d}s}$, and its inverse $N_k^{-1}(z)$. 
(The inverse function recovers $t$ from the value of $N_k(t)$: $t = N_k^{-1}\big( N_k(t)\big)$). This sampling is done with `nhppp`'s `vdraw_cumulative_intensity()`
function. 

A bit of calculus can yield $N_k(t) = \frac{1}{\beta_k} (e^{\alpha_k + \beta_k t} - e^{\alpha_k})$, which we can implement in vectorized form and with default parameters
already set: 

```{r Nu}
Nu <- function(t, Lambda_args = list(population), ...) {
  P <- Lambda_args$population
  (
    exp(P$param_cancer_death_intercept + P$param_cancer_death_slope * t) -
      exp(P$param_cancer_death_intercept)
  ) / P$param_cancer_death_slope
}
```
The inverse $N_k^{-1}(\cdot)$ is 

$N_k^{-1}(z) = \big( \log( \beta_k z + e^\alpha_k ) - \alpha_k \big)/\beta_k $

```{r Nu_inv}
Nu_inv <- function(z, Lambda_inv_args = list(population), ...) {
  P <- Lambda_inv_args$population
  (
    log(P$param_cancer_death_slope * z +
      exp(P$param_cancer_death_intercept)) -
      P$param_cancer_death_intercept
  ) / P$param_cancer_death_slope
}
```

Then, we can sample the times to cancer death.  

```{r age_dead_cancer_causes}
args_list <- list(population = pop[!is.na(age_cancer_emergence), ])
pop[
  !is.na(age_cancer_emergence),
  age_dead_from_cancer_causes := nhppp::vdraw_cumulative_intensity(
    Lambda = Nu,
    Lambda_args = args_list,
    Lambda_inv = Nu_inv,
    Lambda_inv_args = args_list,
    t_min = pop[!is.na(age_cancer_emergence), age_cancer_emergence],
    t_max = pop[!is.na(age_cancer_emergence), age_dead_from_other_causes],
    atmost1 = TRUE
  )
]

rm(list = "args_list") # cleanup
```


### Dying from all causes 

The age of death from all causes is the minimum of the ages across both causes of death.
```{r age_dead}
pop[
  ,
  age_dead := pmin(age_dead_from_other_causes,
    age_dead_from_cancer_causes,
    na.rm = TRUE
  )
]
```

### Clinical cancer diagnosis 

Cancers first emerge in a pre-clinical stage. Some will be diagnosed as  
clinical cancers with intensity function $\mu_k$. We will assume that 
clinical diagnosis has constant rate which is distributed according to the model

$\mu_k(t) := \mu_k \sim U(0.20, 0.27)$, 

where $k$ indexes over people with cancer. 

``` {r param_cancer_dx_rates}
pop[
  !is.na(age_cancer_emergence),
  param_clinical_cancer_dx_rate := runif(.N, 0.20, 0.27)
]
```

Constant rates result in exponential
times, which we can easily sample with the `stats::rexp()` function, as per 
the commented out code below.  
``` {r rexp}
### Using rexp()
tictoc::tic()
pop[
  !is.na(age_cancer_emergence),
  age_clinical_cancer_dx :=
    age_cancer_emergence +
    rexp(.N, rate = param_clinical_cancer_dx_rate)
]
pop[
  age_clinical_cancer_dx >= age_dead,
  age_clinical_cancer_dx := NA
]
tictoc::toc()
```

With `nhppp`, we can use the `vdraw_sc_step_regular()` function that samples from 
piecewise constant intensities. (A constant function over an interval is still 
a piecewise constant function -- with a single piece.) The `nhppp` implementation 
will be only a bit slower -- but it is worth showing. 

``` {r cancer_dx_rates}
tictoc::tic()
mu_mat <- as.matrix(pop[
  !is.na(age_cancer_emergence),
  param_clinical_cancer_dx_rate
])

pop[
  !is.na(age_cancer_emergence),
  age_clinical_cancer_dx := nhppp::vdraw_sc_step_regular(
    lambda_matrix = mu_mat,
    rate_matrix_t_min = pop[!is.na(age_cancer_emergence), age_cancer_emergence],
    rate_matrix_t_max = pop[!is.na(age_cancer_emergence), age_dead],
    atmost1 = TRUE
  )
]
tictoc::toc()
```


## Some descriptives
```{r descriptives}
# pop$age_cancer_emergence |> summary()
summary(pop)
```


## Bibliography

Trikalinos TA, Sereda Y. _nhppp: Simulating Nonhomogeneous Poisson Point Processes in R_. arXiv preprint arXiv:2402.00358. 2024 Feb 1.

Since the publication of the paper, the syntax and options of the `nhppp` package have
evolved. To reproduce the code in the paper, you have to install the version 
of `nhppp` used in the paper. Alternatively, take a look at the vignettes, 
which are written to work with the current package.  

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ztppp.R
\name{ztppp}
\alias{ztppp}
\title{Simulate a zero-truncated homogeneous Poisson Point Process over (t_min, t_max]}
\usage{
ztppp(rate, t_min, t_max, atmost1 = FALSE)
}
\arguments{
\item{rate}{(scalar, double) constant instantaneous rate}

\item{t_min}{(scalar, double) lower bound of the time interval}

\item{t_max}{(scalar, double) upper bound of the time interval}

\item{atmost1}{boolean, draw at most 1 event time}
}
\value{
a vector of event times of size \code{size}
}
\description{
Simulate a zero-truncated homogeneous Poisson Point Process over (t_min, t_max]
}
\examples{
x <- ztppp(t_min = 0, t_max = 10, rate = 0.001)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ztdraw_sc_linear.R
\name{ztdraw_sc_linear}
\alias{ztdraw_sc_linear}
\title{Simulate \code{size} samples from a zero-truncated non homogeneous Poisson Point Process (zt-NHPPP) from
(t_min, t_max) with linear intensity function}
\usage{
ztdraw_sc_linear(intercept, slope, t_min, t_max, atmost1 = FALSE)
}
\arguments{
\item{intercept}{(double) the intercept}

\item{slope}{(double) the slope}

\item{t_min}{(double) the lower bound of the time interval}

\item{t_max}{(double) the upper bound of the time interval}

\item{atmost1}{(boolean) draw 1 event time}
}
\value{
a vector of at least 1 event times
}
\description{
Sample zero-truncated NHPPP times from a linear intensity function
using the inversion method, optionally using an \code{rstream}
generator
}
\examples{
x <- ztdraw_sc_linear(intercept = 0, slope = 0.2, t_min = 0, t_max = 10)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-helper.R
\name{expect_no_error}
\alias{expect_no_error}
\title{Helper functions}
\usage{
expect_no_error(expr)
}
\arguments{
\item{expr}{Expression.}
}
\description{
helper function that augments
\code{test_that::expect_no_error()} to expect no error.
Copied from the \code{R6} source code.
}
\details{
Small utility functions. Not to be exported to the user.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ppp_next_n.R
\name{ppp_next_n}
\alias{ppp_next_n}
\title{Simulate n events from a homogeneous Poisson Point Process.}
\usage{
ppp_next_n(n = 1, rate = 1, t_min = 0, rng_stream = deprecated())
}
\arguments{
\item{n}{scalar number of samples}

\item{rate}{scalar instantaneous rate}

\item{t_min}{scalar for the starting time value}

\item{rng_stream}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} an \code{rstream} object}
}
\value{
a vector with event times t (starting from t_min)
}
\description{
Simulate n events from a homogeneous Poisson Point Process.
}
\examples{
x <- ppp_next_n(n = 10, rate = 1, t_min = 0)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ppp_n.R
\name{ppp_n}
\alias{ppp_n}
\title{Simulate specific number of points from a homogeneous Poisson Point Process over (t_min, t_max]}
\usage{
ppp_n(size, range_t = c(0, 10), rng_stream = NULL)
}
\arguments{
\item{size}{(int) the number of points to be simulated}

\item{range_t}{(vector, double) min and max of the time interval}

\item{rng_stream}{an \code{rstream} object}
}
\value{
a vector of event times of size \code{size}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}
Use \code{ppp_exactly_n} instead.
}
\examples{
x <- ppp_n(size = 10, range_t = c(0, 10))
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-invert_bijection.R
\name{inverse_with_uniroot_sorted}
\alias{inverse_with_uniroot_sorted}
\title{Numerically evaluate the inverse of a monotonically increasing continuous
function from R to R at specific points.}
\usage{
inverse_with_uniroot_sorted(
  f,
  y,
  range_x = c(0, 10),
  range_y = c(f(range_x[1]), f(range_x[2]))
)
}
\arguments{
\item{f}{(function) the function to be inverted; must be continuous
and increasing}

\item{y}{(vector, double) the f(x)=y values in which to evaluate the inverse;
must be in ascending order}

\item{range_x}{(vector, double) the min and max of the domain of f()}

\item{range_y}{(vector, double) the min and max in the range of f()}
}
\value{
(vector, double) vector of x=f^(-1)(y): the inverted values
}
\description{
Numerically evaluate the inverse of a monotonically increasing continuous
function from R to R at specific points.
}
\examples{
inverse_with_uniroot_sorted(f = function(x) {
  2 * x
}, y = c(0, 0.5))
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/draw_sc_step_regular.R
\name{draw_sc_step_regular}
\alias{draw_sc_step_regular}
\title{Sampling from NHPPPs with piecewise constant intensities
with same interval lengths (non-vectorized)}
\usage{
draw_sc_step_regular(
  Lambda_vector = NULL,
  lambda_vector = NULL,
  t_min = NULL,
  t_max = NULL,
  atmost1 = FALSE,
  atleast1 = FALSE
)
}
\arguments{
\item{Lambda_vector}{(scalar, double) \code{K} integrated intensity rates at the end of each interval}

\item{lambda_vector}{(scalar, double) \code{K} constant intensity rates, one per interval}

\item{t_min}{(scalar, double) lower bound of the time interval}

\item{t_max}{(scalar, double) upper bound of the time interval}

\item{atmost1}{boolean, draw at most 1 event time}

\item{atleast1}{boolean, draw at least 1 event time}
}
\value{
a vector of event times t
if no events realize, it will have 0 length
}
\description{
Sampling from NHPPPs with piecewise constant intensities
with same interval lengths (non-vectorized)
}
\examples{
x <- draw_sc_step_regular(Lambda_vector = 1:5, t_min = 0, t_max = 5)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ppp.R
\name{ppp}
\alias{ppp}
\title{Simulate a homogeneous Poisson Point Process in (t_min, t_max]}
\usage{
ppp(rate, t_min, t_max, atmost1 = FALSE, tol = 10^-6)
}
\arguments{
\item{rate}{(scalar, double) constant instantaneous rate}

\item{t_min}{(scalar, double) the lower bound of the time interval}

\item{t_max}{(scalar, double) the upper bound of the time interval}

\item{atmost1}{boolean, draw at most 1 event time}

\item{tol}{the probability that we will have more than
the drawn events in (t_min, t_max]}
}
\value{
a vector of event times t
if no events realize, it will have 0 length
}
\description{
Simulate a homogeneous Poisson Point Process in (t_min, t_max]
}
\examples{
x <- ppp(rate = 1, t_min = 0, t_max = 10, tol = 10^-6)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_step_majorizer.R
\name{get_step_majorizer}
\alias{get_step_majorizer}
\title{Piecewise constant (step) majorizer for K-Lipschitz functions over an interval
(vectorized over the \code{breaks} argument).}
\usage{
get_step_majorizer(fun, breaks, is_monotone = TRUE, K = 0)
}
\arguments{
\item{fun}{A function object with a single argument \code{x}.
If \code{x} is a matrix, \code{fun} should be vectorized over it.}

\item{breaks}{(vector or matrix) The set of \code{M+1} boundaries for the \code{M} subintervals in \code{x}.
If breaks is a matrix, each row is treated as a separate set of breaks.}

\item{is_monotone}{(boolean) Is the function monotone? (Default is \code{TRUE}.)}

\item{K}{(double) A non-negative number for the Lipschitz cone. (Default is 0.)}
}
\value{
A vector of length \code{M} with the values of the piecewise constant majorizer
}
\description{
Return a piecewise constant (step) majorizer for K-Lipschitz functions
over an interval. The function is vectorized over the \code{breaks} argument.
The returned object has the same dimensions as \code{breaks}.
}
\examples{
get_step_majorizer(fun = abs, breaks = -5:5, is_monotone = FALSE, K = 1)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ztdraw_sc_loglinear.R
\name{ztdraw_sc_loglinear}
\alias{ztdraw_sc_loglinear}
\title{Simulate from a zero-truncated non homogeneous Poisson Point Process (zt-NHPPP) from
(t_min, t_max) with a log-linear intensity function}
\usage{
ztdraw_sc_loglinear(intercept, slope, t_min, t_max, atmost1 = FALSE)
}
\arguments{
\item{intercept}{(double) the intercept in the exponent}

\item{slope}{(double) the slope in the exponent}

\item{t_min}{(double) the lower bound of the time interval}

\item{t_max}{(double) the upper bound of the time interval}

\item{atmost1}{boolean, 1 event time}
}
\value{
a vector of at least 1 event times
}
\description{
Sample zt-NHPPP times from an log-linear intensity function
}
\examples{
x <- ztdraw_sc_loglinear(intercept = 0, slope = 0.2, t_min = 0, t_max = 10)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vdraw_sc_step_regular_cpp.R
\name{vdraw_sc_step_regular_cpp}
\alias{vdraw_sc_step_regular_cpp}
\title{Vectorized sampling from NHPPPs with piecewise constant intensities
with same interval lengths (C++)}
\usage{
vdraw_sc_step_regular_cpp(
  lambda_matrix = NULL,
  Lambda_matrix = NULL,
  rate_matrix_t_min = NULL,
  rate_matrix_t_max = NULL,
  t_min = NULL,
  t_max = NULL,
  tol = 10^-6,
  atmost1 = FALSE,
  atmostB = NULL
)
}
\arguments{
\item{lambda_matrix}{(matrix) intensity rates, one per interval}

\item{Lambda_matrix}{(matrix) integrated intensity rates at the end of each interval}

\item{rate_matrix_t_min}{(scalar | vector | column matrix) is the lower bound
of the time interval for each row of (Lambda|lambda)_maj_matrix.
The length of this argument is the number of point processes that should be drawn.}

\item{rate_matrix_t_max}{(scalar | vector | column matrix) the upper bound
of the time interval for each row of (Lambda|lambda)_maj_matrix.
The length of this argument is the number of point processes that should be drawn.}

\item{t_min}{(scalar | vector | column matrix) is the lower bound
of a subinterval of (rate_matrix_t_min, rate_matrix_t_max]. If set,
times are sampled from the subinterval.
If omitted, it is equivalent to \code{rate_matrix_t_min}.}

\item{t_max}{(scalar | vector | column matrix) is the upper bound
of a subinterval of (rate_matrix_t_min, rate_matrix_t_max]. If set,
times are sampled from the subinterval.
If omitted, it is equivalent to \code{rate_matrix_t_max}.}

\item{tol}{(scalar, double) tolerance for the number of events}

\item{atmost1}{boolean, draw at most 1 event time}

\item{atmostB}{If not NULL, draw at most B (B>0) event times. NULL means ignore.}
}
\value{
a matrix of event times t, with rows corresponding to the sampled point processes.
}
\description{
Simulate a piecewise constant-rate Poisson Point Process over \verb{(t_min, t_max]} (inversion method)
where the intervals have the same length (are "regular").
}
\examples{
x <- vdraw_sc_step_regular_cpp(
  Lambda_matrix = matrix(1:5, nrow = 1),
  rate_matrix_t_min = 100,
  rate_matrix_t_max = 110,
  atmost1 = FALSE
)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/matrix_diff_columns.R
\name{matrix_diff_columns}
\alias{matrix_diff_columns}
\title{Usage: matrix_diff_columns( X )}
\usage{
matrix_diff_columns(X)
}
\description{
Usage: matrix_diff_columns( X )
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-vectorized.R
\name{make_cumulative_Lambda_matrix}
\alias{make_cumulative_Lambda_matrix}
\title{Helper function for the vectorized versions of sampling functions.
Takes the usual ways that \code{lambda_mat} and \code{Lambda_mat} are specified
and returns \code{Lambda_mat}.}
\usage{
make_cumulative_Lambda_matrix(
  Lambda_mat = NULL,
  lambda_mat = NULL,
  interval_duration = NULL
)
}
\arguments{
\item{Lambda_mat}{a matrix of cumulative intensities or missing}

\item{lambda_mat}{a matrix of intensities or missing}

\item{interval_duration}{a vector with the same number of elements as the rows of \code{Lambda_mat}}
}
\value{
A matrix (r x 2), row-expanded if needed
}
\description{
Helper function for the vectorized versions of sampling functions.
Takes the usual ways that \code{lambda_mat} and \code{Lambda_mat} are specified
and returns \code{Lambda_mat}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ppp_orderstat.R
\name{ppp_orderstat}
\alias{ppp_orderstat}
\title{Simulate a homogeneous Poisson Point Process over (t_min, t_max] (order statistics method)}
\usage{
ppp_orderstat(range_t = c(0, 10), rate = 1, rng_stream = NULL, atmost1 = FALSE)
}
\arguments{
\item{range_t}{(vector, double) min and max of the time interval}

\item{rate}{(scalar, double) constant instantaneous rate}

\item{rng_stream}{an \code{rstream} object}

\item{atmost1}{boolean, draw at most 1 event time}
}
\value{
a vector of event times t
if no events realize, it will have 0 length
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}
Use \code{ppp2} instead.
}
\examples{
x <- ppp_orderstat(range_t = c(0, 10), rate = 1)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ztdraw_cumulative_intensity.R
\name{ztdraw_cumulative_intensity}
\alias{ztdraw_cumulative_intensity}
\title{Simulate from a zero-truncated non homogeneous Poisson Point Process (zt-NHPPP) from
(t_min, t_max) (order statistics method)}
\usage{
ztdraw_cumulative_intensity(Lambda, Lambda_inv, t_min, t_max, atmost1 = FALSE)
}
\arguments{
\item{Lambda}{(function, double vector) a continuous increasing R to R map
which is the integrated rate of the NHPPP}

\item{Lambda_inv}{(function, double vector) the inverse of \code{Lambda()}}

\item{t_min}{(double) the lower bound of the time interval}

\item{t_max}{(double) the upper bound of the time interval}

\item{atmost1}{(boolean) draw at most 1 event time}
}
\value{
a vector of at least 1 event times
}
\description{
Sample zero-truncated NHPPP times using the order statistics method,
optionally using an \code{rstream} generator
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/draw_sc_step.R
\name{draw_sc_step}
\alias{draw_sc_step}
\title{Simulate a piecewise constant-rate Poisson Point Process over \verb{(t_min, t_max]} (inversion method)
The intervals need not have the same length.}
\usage{
draw_sc_step(lambda_vector, time_breaks, atmost1 = FALSE, atleast1 = FALSE)
}
\arguments{
\item{lambda_vector}{(scalar, double) \code{K} constant rates, one per interval}

\item{time_breaks}{(vector, double) \code{K+1} time points defining \code{K} intervals
of constant rates:
\verb{[t_1 = range_t[1],       t_2)}: the first interval
\verb{[t_k,                t_\{k+1\})}: the \code{k}-th interval
\verb{[t_\{K\}, t_\{K+1\} = range_t[2])}: the \code{K}-th (last) interval}

\item{atmost1}{boolean, draw at most 1 event time}

\item{atleast1}{boolean, draw at least 1 event time}
}
\value{
a vector of event times t
if no events realize, it will have 0 length
}
\description{
Simulate a piecewise constant-rate Poisson Point Process over \verb{(t_min, t_max]} (inversion method)
The intervals need not have the same length.
}
\examples{
x <- draw_sc_step(lambda_vector = rep(1, 5), time_breaks = c(0:5))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/matrix_cumsum_columns_inplace.R
\name{matrix_cumsum_columns_inplace}
\alias{matrix_cumsum_columns_inplace}
\title{Usage: matrix_cumsum_columns_inplace( X )}
\usage{
matrix_cumsum_columns_inplace(X)
}
\description{
Usage: matrix_cumsum_columns_inplace( X )
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-rngstream.R
\name{rztpois}
\alias{rztpois}
\title{Zero-truncated Poisson random samples (basic R)}
\usage{
rztpois(n, lambda)
}
\arguments{
\item{n}{Integer, number of samples}

\item{lambda}{Positive number, the mean of the original
(untruncated) Poisson distribution}
}
\value{
a vector of non zero counts of size \code{n}
}
\description{
Sample zero-truncated Poisson random samples (basic R)
}
\examples{
rztpois(10, 1)
rztpois(10, 1:10)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-rngstream.R
\name{rng_stream_rztpois}
\alias{rng_stream_rztpois}
\title{Zero-truncated Poisson random samples from \code{rstream} objects}
\usage{
rng_stream_rztpois(size = 1, lambda = 1, rng_stream = NULL)
}
\arguments{
\item{size}{Integer, number of samples}

\item{lambda}{Positive number, the mean of the original
(untruncated) Poisson distribution}

\item{rng_stream}{(\code{rstream}) an \code{rstream} object or \code{NULL}}
}
\value{
a vector of non zero counts of size \code{size}
}
\description{
Sample from \code{rstream} objects
}
\examples{
rng_stream_rztpois(10)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vdraw_cumulative_intensity.R
\name{vdraw_cumulative_intensity}
\alias{vdraw_cumulative_intensity}
\title{Vectorized simulation from a non homogeneous Poisson Point Process (NHPPP) from
(t_min, t_max) given the cumulative intensity function and its inverse}
\usage{
vdraw_cumulative_intensity(
  Lambda,
  Lambda_inv,
  t_min,
  t_max,
  Lambda_args = NULL,
  Lambda_inv_args = NULL,
  tol = 10^-6,
  atmost1 = FALSE,
  atleast1 = FALSE
)
}
\arguments{
\item{Lambda}{(function, double vector) an increasing function
which is the integrated rate of the NHPPP.
It should take a vectorized argument t for times and an optional arguments list.}

\item{Lambda_inv}{(function, double vector) the inverse of \code{Lambda()}, also in vectorized form
It should take a vectorized argument z and an optional arguments list.}

\item{t_min}{(scalar | vector | column matrix) the lower bound of the interval for each sampled point process
The length of this argument is the number of point processes that should be drawn.}

\item{t_max}{(scalar | vector | column matrix) the upper bound of the interval for each sampled point process
The length of this argument is the number of point processes that should be drawn.}

\item{Lambda_args}{(list) optional arguments to pass to Lambda.}

\item{Lambda_inv_args}{(list) optional arguments to pass to Lambda_inv().}

\item{tol}{the tolerange for the calulations.}

\item{atmost1}{boolean, draw at most 1 event time per sampled point process.}

\item{atleast1}{boolean, draw at least 1 event time}
}
\value{
a matrix of event times with one row per sampled point process.
}
\description{
Sample NHPPP times using the cumulative intensity function and its inverse.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/draw_intensity_line.R
\name{draw_intensity_line}
\alias{draw_intensity_line}
\title{Simulate from a non homogeneous Poisson Point Process (NHPPP) from
(t0, t_max) (thinning method)}
\usage{
draw_intensity_line(
  lambda,
  majorizer_intercept,
  majorizer_slope,
  t_min,
  t_max,
  majorizer_is_loglinear = FALSE,
  atmost1 = FALSE
)
}
\arguments{
\item{lambda}{(function) the instantaneous rate of the NHPPP.}

\item{majorizer_intercept}{(double) the intercept (\code{alpha}) of the \link{log}linear majorizer function.}

\item{majorizer_slope}{(double) the slope (`beta') of the \link{log}linear majorizer function.}

\item{t_min}{(double) the lower bound of the time interval.}

\item{t_max}{(double) the upper bound of the time interval.}

\item{majorizer_is_loglinear}{(boolean) if \code{TRUE} the majorizer is loglinear \code{exp(alpha + beta * t)}}

\item{atmost1}{boolean, draw at most 1 event time}
}
\value{
a vector of event times (t_); if no events realize,
a vector of length 0
}
\description{
Sample NHPPP times using the thinning method
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-helper.R
\name{compare_ppp_vectors}
\alias{compare_ppp_vectors}
\title{Check that two ppp vectors Q-Q agree}
\usage{
compare_ppp_vectors(ppp1, ppp2, threshold = 0.15, showQQ = TRUE)
}
\arguments{
\item{ppp1}{(vector, double) the first vector}

\item{ppp2}{(vector, double) the second vector}

\item{threshold}{(double) optional: the cutoff for a large absolute threshold}

\item{showQQ}{(boolean) optional: show the QQ plot if the absolute value of the
Difference vs Average ratio in any decile is bigger than the \code{threshold}}
}
\value{
None
}
\description{
Compare that the deciles of two vectors have absolute difference
over average ratios less than \code{threshold}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-linear.R
\name{Lambda_linear_form}
\alias{Lambda_linear_form}
\title{Definite integral of \code{l = intercept + slope*t} at time \code{t}
with \code{L(t0) = 0}}
\usage{
Lambda_linear_form(t, intercept, slope, t0)
}
\arguments{
\item{t}{(double) the time point}

\item{intercept}{(double) the intercept}

\item{slope}{(double) the slope}

\item{t0}{(double) the starting time}
}
\description{
Definite integral of \code{l = intercept + slope*t} starting at
\code{t0} -- only for \verb{l+}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ppp_sequential.R
\name{ppp_sequential}
\alias{ppp_sequential}
\title{Simulate a homogeneous Poisson Point Process over (t_min, t_max]}
\usage{
ppp_sequential(
  range_t = c(0, 10),
  rate = 1,
  tol = 10^-6,
  rng_stream = NULL,
  atmost1 = FALSE
)
}
\arguments{
\item{range_t}{(vector, double) min and max of the time interval}

\item{rate}{(scalar, double) constant instantaneous rate}

\item{tol}{the probability that we will have more than
the drawn events in (t_min, t_max]}

\item{rng_stream}{an \code{rstream} object}

\item{atmost1}{boolean, draw at most 1 event time}
}
\value{
a vector of event times t
if no events realize, it will have 0 length
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}
Use \code{ppp} instead.
}
\examples{
x <- ppp_sequential(range_t = c(0, 10), rate = 1, tol = 10^-6)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nhppp-package.R
\docType{package}
\name{nhppp-package}
\alias{nhppp}
\alias{nhppp-package}
\title{nhppp: Simulating Nonhomogeneous Poisson Point Processes}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

Simulates events from one dimensional nonhomogeneous Poisson point processes (NHPPPs) as per Trikalinos and Sereda (2024, \doi{10.48550/arXiv.2402.00358} and 2024, \doi{10.1371/journal.pone.0311311}). Functions are based on three algorithms that provably sample from a target NHPPP: the time-transformation of a homogeneous Poisson process (of intensity one) via the inverse of the integrated intensity function (Cinlar E, "Theory of stochastic processes" (1975, ISBN:0486497996)); the generation of a Poisson number of order statistics from a fixed density function; and the thinning of a majorizing NHPPP via an acceptance-rejection scheme (Lewis PAW, Shedler, GS (1979) \doi{10.1002/nav.3800260304}).
}
\seealso{
Useful links:
\itemize{
  \item \url{https://bladder-ca.github.io/nhppp/}
  \item \url{https://github.com/bladder-ca/nhppp}
  \item Report bugs at \url{https://github.com/bladder-ca/nhppp/issues}
}

}
\author{
\strong{Maintainer}: Thomas Trikalinos \email{thomas_trikalinos@brown.edu} (\href{https://orcid.org/0000-0002-3990-1848}{ORCID}) [copyright holder]

Authors:
\itemize{
  \item Yuliia Sereda \email{sereda_yuliia@brown.edu} (\href{https://orcid.org/0000-0002-4017-4561}{ORCID})
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vztdraw_intensity.R
\name{vztdraw_intensity}
\alias{vztdraw_intensity}
\title{Vectorized sampling from a zero-truncated non homogeneous Poisson Point Process (NHPPP) from
an interval (thinning method) with piecewise constant_majorizers}
\usage{
vztdraw_intensity(
  lambda = NULL,
  lambda_args = NULL,
  Lambda_maj_matrix = NULL,
  lambda_maj_matrix = NULL,
  range_t = NULL,
  tol = 10^-6,
  atmost1 = FALSE,
  ...
)
}
\arguments{
\item{lambda}{(function) a vectorized intensity function, with one or two arguments.
The first is time. The optional second is a named list with additional arguments.}

\item{lambda_args}{(list) optional list of named arguments for \code{lambda()}}

\item{Lambda_maj_matrix}{(matrix) for the majorizeintegrated intensity rates at the end of each interval}

\item{lambda_maj_matrix}{(matrix) intensity rates, one per interval}

\item{range_t}{(vector, or matrix) \code{t_min} and \code{t_max}, possibly vectorized}

\item{tol}{(scalar, double) tolerance for the number of events}

\item{atmost1}{boolean, draw at most 1 event time}

\item{...}{(any) other arguments (ignored  -- used for flexibility in calling from other functions)}
}
\value{
a matrix of event times (columns) per draw (rows)
NAs are structural empty spots
}
\description{
Vectorized sampling from a zero-truncated non homogeneous Poisson Point Process (NHPPP) from
an interval (thinning method) with piecewise constant_majorizers.
The majorizers are step functions over equal-length time intevals.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/draw_cumulative_intensity_orderstats.R
\name{draw_cumulative_intensity_orderstats}
\alias{draw_cumulative_intensity_orderstats}
\title{Simulate from a non homogeneous Poisson Point Process (NHPPP) from
(t_min, t_max) (order statistics method)}
\usage{
draw_cumulative_intensity_orderstats(
  Lambda,
  Lambda_inv,
  t_min,
  t_max,
  atmost1 = FALSE
)
}
\arguments{
\item{Lambda}{(function, double vector) a continuous increasing R to R map
which is the integrated rate of the NHPPP}

\item{Lambda_inv}{(function, double vector) the inverse of \code{Lambda()}}

\item{t_min}{(double) the lower bound of the time interval}

\item{t_max}{(double) the upper bound of the time interval}

\item{atmost1}{boolean, draw at most 1 event time}
}
\value{
a vector of event times (t_); if no events realize,
a vector of length 0
}
\description{
Sample NHPPP times using the order statistics method,
optionally using an \code{rstream} generator object.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-rngstream.R
\name{rng_stream_rexp}
\alias{rng_stream_rexp}
\title{Exponential random samples from \code{rstream} objects}
\usage{
rng_stream_rexp(size = 1, rate = 1, rng_stream = NULL)
}
\arguments{
\item{size}{Integer, number of samples}

\item{rate}{Positive number, the rate (i.e., 1/mean)}

\item{rng_stream}{(\code{rstream}) an \code{rstream} object or \code{NULL}}
}
\value{
a vector of exponential variates of size \code{size}
}
\description{
Sample from \code{rstream} objects
}
\examples{
rng_stream_rexp(10)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-helper.R
\name{check_ppp_vector_validity}
\alias{check_ppp_vector_validity}
\title{Check the validity of a ppp vector.}
\usage{
check_ppp_vector_validity(
  times,
  t_min,
  t_max = NULL,
  size = NULL,
  atmost1 = FALSE,
  atleast1 = FALSE
)
}
\arguments{
\item{times}{(vector, double) the times to be checked}

\item{t_min}{(double) the start of the time nterval}

\item{t_max}{(double) optional: the end of the time interval}

\item{size}{(double) optional: the size of the vector}

\item{atmost1}{(boolean) optional: at most one sample returned}

\item{atleast1}{(boolean) optional: at least one sample returned}
}
\value{
None
}
\description{
Standard checks for a vector of ordered times. Check
that the \code{times} vector is sorted, has unique values, has all values
in \verb{[t_min, t_max]}, and has length \code{size} (if applicable).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/draw_intensity_step.R
\name{draw_intensity_step}
\alias{draw_intensity_step}
\title{Simulate from a non homogeneous Poisson Point Process (NHPPP) from
(t0, t_max) (thinning method) with piecewise constant_majorizer}
\usage{
draw_intensity_step(lambda, majorizer_vector, time_breaks, atmost1 = FALSE)
}
\arguments{
\item{lambda}{(function) the instantaneous rate of the NHPPP.
A continuous function of time.}

\item{majorizer_vector}{(scalar, double) \code{K} constant majorizing rates, one per interval}

\item{time_breaks}{(vector, double) \code{K+1} time points defining \code{K} intervals
of constant rates:
\verb{[t_1 = range_t[1],       t_2)}: the first interval
\verb{[t_k,                t_\{k+1\})}: the \code{k}-th interval
\verb{[t_\{K\}, t_\{K+1\} = range_t[2])}: the \code{K}-th (last) interval}

\item{atmost1}{boolean, draw at most 1 event time}
}
\value{
a vector of event times (t_); if no events realize,
a vector of length 0
}
\description{
Sample NHPPP times using the thinning method
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-vectorized.R
\name{mat_cumsum_columns}
\alias{mat_cumsum_columns}
\title{Return matrix with column-wise cumulative sum
No checks for arguments is done.}
\usage{
mat_cumsum_columns(X)
}
\arguments{
\item{X}{(matrix)}
}
\value{
matrix
}
\description{
Return matrix with column-wise cumulative sum
No checks for arguments is done.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-calculus.R
\name{simpson_num_integr}
\alias{simpson_num_integr}
\title{Simpson's method to integrate a univariate function.}
\usage{
simpson_num_integr(f, a, b, n)
}
\arguments{
\item{f}{function that takes a single argument}

\item{a}{the lower limit of integration}

\item{b}{the upper limit of integration}

\item{n}{integer, number of integration points with a and b}
}
\value{
numeric, the integration value
examples
#expect 1
simpson_num_integr(sin, 0, pi/2, 100)
#max error for simpson_num_integr(sin, 0, pi/2, 100) is 5.312842e-10
1 * (pi/2 - 0)^5/(180 * 100^4)
}
\description{
Simpson's method to integrate a univariate continuous function.
Faster that R's \code{integrate()} and precise enough, but does not do any checks.
The error is at most \verb{M (b-a)^5/(180 n^4)} where \code{M} is the maximum of
the fourth derivative of the integrand in the interval \verb{[a, b]}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{annual_mortality_rates_2015}
\alias{annual_mortality_rates_2015}
\title{Human mortality database age and sex specific rates for all cause deaths}
\format{
\subsection{\code{annual_mortality_rates_2015}}{

a data.table with 3 rows and 113 columns.
\describe{
\item{birth_cohort}{Birth cohort as a calendar year}
\item{sex}{sex: \code{female}, \code{male}, \code{total}.}
\item{age_0, ..., age 110+}{age-specific death rates}
}
}
}
\source{
\url{https://www.mortality.org}
}
\usage{
annual_mortality_rates_2015
}
\description{
This is the 2015 annual death rates from the 2023 version of the
Human Mortality Database for the USA.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ppp2.R
\name{ppp2}
\alias{ppp2}
\title{Simulate a homogeneous Poisson Point Process over (t_min, t_max] (order statistics method)}
\usage{
ppp2(rate, t_min, t_max, atmost1 = FALSE)
}
\arguments{
\item{rate}{(scalar, double) constant instantaneous rate}

\item{t_min}{(scalar, double) the lower bound of the time interval}

\item{t_max}{(scalar, double) the upper bound of the time interval}

\item{atmost1}{boolean, draw at most 1 event time}
}
\value{
a vector of event times t
if no events realize, it will have 0 length
}
\description{
Internal function -- not to be exported.
Same as \code{ppp} but uses the Order Statistics algorithm.
}
\examples{
x <- ppp(rate = 1, t_min = 0, t_max = 10, tol = 10^-6)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-vectorized.R
\name{make_range_t_matrix}
\alias{make_range_t_matrix}
\title{Helper function for the vectorized versions of sampling functions.
Takes the usual ways that \code{range_t} is specified
(a 2-vector, a 1 x 2 or an r x 2 matrix) and
returns a r x 2 matrix.}
\usage{
make_range_t_matrix(range_t, n_rows)
}
\arguments{
\item{range_t}{a 2-vector, a 1 x 2 or an r x 2 matrix}

\item{n_rows}{the number of rows in the fully expanded matrix (\code{r})}
}
\value{
A matrix (r x 2), row-expanded if needed
}
\description{
Helper function for the vectorized versions of sampling functions.
Takes the usual ways that \code{range_t} is specified
(a 2-vector, a 1 x 2 or an r x 2 matrix) and
returns a r x 2 matrix.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ppp_exactly_n.R
\name{ppp_exactly_n}
\alias{ppp_exactly_n}
\title{Simulate exactly \code{n} points from a homogeneous Poisson Point Process over (t_min, t_max]}
\usage{
ppp_exactly_n(n, t_min, t_max)
}
\arguments{
\item{n}{(int) the number of points to be simulated}

\item{t_min}{(double) the lower bound of the time interval}

\item{t_max}{(double) the upper bound of the time interval}
}
\value{
a vector of event times of size \code{n}
}
\description{
Simulate exactly \code{n} points from a homogeneous Poisson Point Process over (t_min, t_max]
}
\examples{
x <- ppp_exactly_n(n = 10, t_min = 0, t_max = 10)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/draw.R
\name{draw}
\alias{draw}
\title{Generic function for simulating from NHPPPs given the intensity function
or the cumulative intensity function.}
\usage{
draw(
  Lambda = NULL,
  Lambda_inv = NULL,
  lambda = NULL,
  line_majorizer_intercept = NULL,
  line_majorizer_slope = NULL,
  line_majorizer_is_loglinear = FALSE,
  step_majorizer_vector = NULL,
  t_min = NULL,
  t_max = NULL,
  atmost1 = FALSE,
  atleast1 = FALSE
)
}
\arguments{
\item{Lambda}{(function, double vector) the integrated (cumulative) rate of the NHPPP}

\item{Lambda_inv}{(function, double vector) the inverse of `Lambda()'}

\item{lambda}{(function) the instantaneous rate}

\item{line_majorizer_intercept}{The intercept \code{alpha} of the \link{log}linear majorizer function: \code{alpha + beta * t} or \code{exp(alpha + beta * t)}}

\item{line_majorizer_slope}{The slope \code{beta} of the \link{log}linear majorizer function: \code{alpha + beta * t} or \code{exp(alpha + beta * t)}}

\item{line_majorizer_is_loglinear}{(boolean) if \code{TRUE} the majorizer is loglinear \code{exp(alpha + beta * t)}; if \code{FALSE} it is a linear function}

\item{step_majorizer_vector}{(vector, double) \code{K} constant majorizing rates, one per interval; all intervals are of equal length (regular)}

\item{t_min}{(double) the lower bound of the interval}

\item{t_max}{(double) the upper bound of the interval}

\item{atmost1}{boolean, draw at most 1 event time}

\item{atleast1}{boolean, draw at least 1 event time in interval}
}
\value{
a vector of event times
}
\description{
This is a wrapper to the package's specific functions, and thus somewhat slower.
For time-intensive simulations prefer one of the specific functions.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-invert_bijection.R
\name{inverse_with_uniroot}
\alias{inverse_with_uniroot}
\title{Numerically evaluate the inverse of a function at a specific point}
\usage{
inverse_with_uniroot(
  f = f,
  y,
  min_x = 0,
  max_x = 1,
  min_y = f(min_x),
  max_y = f(max_x)
)
}
\arguments{
\item{f}{(function) the function to be inverted; must be continuous
and increasing}

\item{y}{(scalar, double) the f(x)=y value in which to evaluate the inverse}

\item{min_x}{(scalar, double) the min of the domain of f()}

\item{max_x}{(scalar, double) the max of the domain of f()}

\item{min_y}{(scalar, double) the min in the range of f()}

\item{max_y}{(scalar, double) the max in the range of f()}
}
\value{
(scalar, double) vector of x=f^(-1)(y): the inverted value
}
\description{
Numerically evaluate the inverse of a function at a specific point
}
\examples{
inverse_with_uniroot(f = function(x) {
  2 * x
}, y = 0.5)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-vectorized.R
\name{mat_cumsum_columns_with_vector_ceiling}
\alias{mat_cumsum_columns_with_vector_ceiling}
\title{Return matrix with column-wise cumulative sum
replacing cells larger than \code{ceil} with \code{NA}.
No checks for arguments is done.}
\usage{
mat_cumsum_columns_with_vector_ceiling(X, ceil = Inf)
}
\arguments{
\item{X}{(matrix)}

\item{ceil}{(vector or Inf) the set of ceilings to be applied, per row of \code{X}}
}
\value{
matrix
}
\description{
Return matrix with column-wise cumulative sum
replacing cells larger than \code{ceil} with \code{NA}.
No checks for arguments is done.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-linear.R
\name{Lambda_inv_exp_form}
\alias{Lambda_inv_exp_form}
\title{Inverse of the definite integral of \code{l = exp(intercept + slope*t)} at time \code{t}}
\usage{
Lambda_inv_exp_form(z, intercept, slope, t0)
}
\arguments{
\item{z}{(double) the value of integrated rate for which you want to find the time}

\item{intercept}{(double) the intercept}

\item{slope}{(double) the slope}

\item{t0}{(double) the starting time}
}
\description{
Inverse of the definite integral of \code{l = exp(intercept + slope*t)} only for \verb{l+}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/draw_sc_loglinear.R
\name{draw_sc_loglinear}
\alias{draw_sc_loglinear}
\title{Special case: Simulate from a non homogeneous Poisson Point Process (NHPPP) from
(t_min, t_max) with log-linear intensity function (inversion method)}
\usage{
draw_sc_loglinear(intercept, slope, t_min, t_max, atmost1 = FALSE)
}
\arguments{
\item{intercept}{(double) the intercept in the exponent}

\item{slope}{(double) the slope in the exponent}

\item{t_min}{(double) lower bound of the time interval}

\item{t_max}{(double) upper bound of the time interval}

\item{atmost1}{boolean, draw at most 1 event time}
}
\value{
a vector of event times (t_); if no events realize,
a vector of length 0
}
\description{
Sample NHPPP times from an log linear intensity function
using the inversion method, optionally using an \code{rstream}
generator
}
\examples{
x <- draw_sc_loglinear(intercept = 0, slope = 0.2, t_min = 0, t_max = 10)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-helper.R
\name{read_code}
\alias{read_code}
\title{Read code from text file as string}
\usage{
read_code(codeFile)
}
\arguments{
\item{codeFile}{Path to file}
}
\value{
\code{codeFile} contents as a character string
}
\description{
Read code from text file as string
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vdraw_intensity_step_regular_cpp.R
\name{vdraw_intensity_step_regular_cpp}
\alias{vdraw_intensity_step_regular_cpp}
\title{Vectorized sampling from a non homogeneous Poisson Point Process (NHPPP) given
the intensity function with piecewise constant_majorizers (C++)}
\usage{
vdraw_intensity_step_regular_cpp(
  lambda = NULL,
  lambda_args = NULL,
  Lambda_maj_matrix = NULL,
  lambda_maj_matrix = NULL,
  rate_matrix_t_min = NULL,
  rate_matrix_t_max = NULL,
  t_min = NULL,
  t_max = NULL,
  tol = 10^-6,
  atmost1 = FALSE,
  atmostB = NULL
)
}
\arguments{
\item{lambda}{(function) intensity function, vectorized}

\item{lambda_args}{(list) optional arguments to pass to \code{lambda}}

\item{Lambda_maj_matrix}{(matrix) integrated intensity rates at the end of each interval}

\item{lambda_maj_matrix}{(matrix) intensity rates, one per interval}

\item{rate_matrix_t_min}{(scalar | vector | column matrix) is the lower bound
of the time interval for each row of (Lambda|lambda)_maj_matrix.
The length of this argument is the number of point processes that should be drawn.}

\item{rate_matrix_t_max}{(scalar | vector | column matrix) the upper bound
of the time interval for each row of (Lambda|lambda)_maj_matrix.
The length of this argument is the number of point processes that should be drawn.}

\item{t_min}{(scalar | vector | column matrix) is the lower bound
of a subinterval of (rate_matrix_t_min, rate_matrix_t_max]. If set,
times are sampled from the subinterval.
If omitted, it is equivalent to \code{rate_matrix_t_min}.}

\item{t_max}{(scalar | vector | column matrix) is the upper bound
of a subinterval of (rate_matrix_t_min, rate_matrix_t_max]. If set,
times are sampled from the subinterval.
If omitted, it is equivalent to \code{rate_matrix_t_max}.}

\item{tol}{(scalar, double) tolerance for the number of events}

\item{atmost1}{boolean, draw at most 1 event time}

\item{atmostB}{If not NULL, draw at most B (B>0) event times. NULL means ignore.}
}
\value{
a matrix of event times (columns) per draw (rows)
NAs are structural empty spots
}
\description{
Vectorized sampling from a non homogeneous Poisson Point Process (NHPPP) using the intensity
function and piecewise constant_majorizers.
The majorizers are step functions over equal-length time intevals.
}
\keyword{internal}
